/**
 * C-code modifications Copyright (C) 2009 Alfred Steffens Jr.
 * based on TekTracker, Copyright (C) 2001 Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif // HAVE_CONFIG_H
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <stdarg.h>
#include <defs.h>
#include <utils.h>
#include <logprintf.h>
#include <errorprintf.h>
#include <msgprintf.h>
#include <thread.h>
#include <record.h>
#include <xmltools.h>
#include <song.h>
#include <screen.h>
#include <midicontrol.h>

#ifdef SYNCRONIZE_EXIT
    static unsigned long break_count = 0;
    int exit_count = 0;
#endif

//
//
//
char version_number[64];
//
//
//
static char dev_names[MAX_DEVICES][MAX_PATH_LENGTH];
static int dev_count = 0;
static char song_name[MAX_PATH_LENGTH];
static int alsa_flag = 0;
static int intclock = M_TRUE;
static int beatpulse = M_FALSE;
static int realtime = M_FALSE;
static int rtclkfreq = 1024;
struct Thread *pth = NULL;
extern int g_error_display;
extern int g_message_display;


//
//  Global struct
//  The Screen struct for the application
//
struct Screen *p_Screen = NULL;
//
//  Global struct
//  The global Song struct
//
struct Song *p_Song = NULL;
//
//  Global struct
//  Midi parameters
//
struct MidiControl *p_MidiControl = NULL;
extern struct RecordEvent *p_REvents, *p_CurrREvents;
struct RecordEvent *p_Rec, *p_CurrRec;

//
//  MIDI Bank Settings
//
struct Bank *p_Banks = NULL;
struct Bank **pp_BankAry = NULL;
int g_Max_Bank = -1;

//
//  Global quit flag shows that the thread loop
//  has successfully broken out
//
int g_Thread_Loop_Done = M_FALSE;
int g_Record_Session = M_FALSE;
int g_Playback_Session = M_FALSE;
char g_Playback_Filename[MAX_PATH_LENGTH];

int read_settings(void);


//  -------------------------------------------------------------------------
//  Function    --  main
//
//  Description
//
//      The entry point into the program.
//
//  Arguments
//
//      argc
//
//      argv
//
//  Return
//
//      Integer: 0 on success, nonzero on failure.
//
//  -------------------------------------------------------------------------
int main( int argc, char **argv )
{
    int i, optch;
    extern char *optarg;
    extern int optind, opterr, optopt;


    LogFile_Create();

    //
	// Announcements.
    //
	Error_Printf("attrk version %s\n", VERSION );
	Error_Printf("attrk: Copyright 2009 afred steffens jr.\n" );
	Error_Printf("based on TekTracker, Copyright 2001 billy biggs\n" );
	Error_Printf("This is free software under the GNU GPL, "
			"for details see http://www.gnu.org.\n" );


    //
    //  Command-line supercedes the config file
    //
    song_name[0] = 0x00;
    for (i=0; i < MAX_DEVICES; i++)
        dev_names[i][0] = 0x00;

    //
	// Reads in the settings from the .attrkrc file.
    //
	if (read_settings() < 0)
    {
        Error_Printf("ERROR: .attrkrc file not found\n");
        LogFile_Exit();
        return 1;
    }


    //
    //  Read the command-line
    //
    optch = 0;
    while (optch != -1)
    {
        optch = getopt(argc, argv, "abc:d:ehij:s:r");
        switch (optch)
        {
            case 'a':
                //
                //  Command-line ALSA flag
                //      --alsa
                //      or
                //      -a
                //
                LogFile_Printf("command-line : ALSA flag\n");
                alsa_flag = M_TRUE;
            break;

            case 'd':
                if (dev_count < MAX_DEVICES)
                {
                    //
                    //  Command-line MIDI devices
                    //
                    //      --device='/dev/midi00'
                    //      or
                    //      -d '/dev/midi00'
                    //
                    LogFile_Printf("command-line : dev %d, %s\n",
                                   dev_count, optarg);
                    if (optarg)
                        strcpy(dev_names[dev_count++], optarg);
                }
            break;

            case 'i':
                //
                //  Command-line internal clock
                //
                //      --intclock
                //      or
                //      -i
                //
                LogFile_Printf("command-line : INTERNAL clock\n");
                intclock = M_TRUE;
            break;

            case 'e':
                //
                //  Command-line clock flag
                //
                //      --extclock
                //      or
                //      -e
                //
                LogFile_Printf("command-line : EXTERNAL clock\n");
                intclock = M_FALSE;
            break;

            case 'b':
                //
                //  Command-line beatpulse flag
                //      --beatpulse
                //      or
                //      -b
                //
                LogFile_Printf("command-line : beat pulse\n");
                beatpulse = M_TRUE;
            break;

            case 'r':
                //
                //  Command-line REAL TIME flag
                //      --realtime
                //      or
                //      -r
                //
                LogFile_Printf("command-line : REAL TIME\n");
                realtime = M_TRUE;
            break;

            case 'c':
                //
                //  Command-line frequency
                //
                //      --rtclkfreq=1024
                //      or
                //      -c 1024
                //
                if (optarg)
                {
                    rtclkfreq = atoi(optarg);
                    LogFile_Printf("command-line : real time"
                                   " clock FREQUENCY %d\n", rtclkfreq);
                }
            break;

            case 's':
                //
                //  Command-line initial song
                //
                //      --song='/home/mysong.ttrk'
                //      or
                //      -s '/home/mysong.ttrk'
                //
                if (optarg)
                {
                    strncpy(song_name, optarg, MAX_PATH_LENGTH-1);
                    song_name[MAX_PATH_LENGTH-1] = 0x00;
                    LogFile_Printf("command-line : song name"
                                   " %s\n", song_name);
                }
            break;

            case 'h':
                //
                //  Command-line -- record the session
                //      -h
                //
                LogFile_Printf("command-line : RECORD SESSION\n");
                g_Record_Session = M_TRUE;
            break;

            case 'j':
                //
                //  Command-line -- playback a session
                //      -j filename
                //
                LogFile_Printf("command-line : PLAYBACK SESSION %s\n",
                                optarg);
                g_Playback_Session = M_TRUE;
                strcpy(g_Playback_Filename, optarg);
            break;

        }
    }
    // while -- optarg

    //
    //  Create and open the Midi Machinery
    //
    p_MidiControl = Create_MidiControl(intclock, realtime, beatpulse, 0,
                               rtclkfreq, 0, 0);
    if (p_MidiControl == NULL)
        goto Error;

    if (dev_count == 0)
    {
        //
        //  No command-line arguments, so using only 1 MIDI device
        //
        if (mctrl_openMidiDevice(p_MidiControl,  "/dev/midi00"))
        {
            Error_Printf("Cannot open device /dev/midi00\n");
            goto Error;
        }
    }
    else
    {
        //
        //  use as many MIDI devices as were given on the command line
        //
        for (i=0; i < dev_count; i++)
        {
            if (mctrl_openMidiDevice(p_MidiControl,  dev_names[i]))
            {
                Error_Printf("Cannot open device (%s)\n", dev_names[i]);
                goto Error;
            }
        }
    }

    //
    //  We need at least one Song struct defined in order to proceed
    //
    p_Song = Create_Song();
    if (p_Song == NULL)
        goto Error;

    //
    //  Are we recording?
    //
    if (g_Record_Session)
    {
        p_Song->re_record = 1;
    }

    //
    //  Are we doing playback?
    //
    if (g_Playback_Session)
    {
        struct RecordEvent *prec;

        prec = Read_Events_File(g_Playback_Filename);
        if (prec)
        {
            p_REvents = p_CurrREvents = prec;
            p_Song->re_play = 1;
        }
        else
            Error_Printf("ERROR: main: cannot open session (%s)\n",
                         g_Playback_Filename);
    }

    //
    // Load a file if one is specified.
    //
    if (strlen(song_name) > 0)
    {
        if (song_loadSong(p_Song, song_name))
            goto Error;
    }

    //
    // Start up the MIDI playing thread.
    //
    pth = Create_Thread();
    if (pth == NULL)
    {
        Error_Printf("ERROR: main: NULL from Create_Thread\n");
        goto Error;
    }
    
	if (thread_run(pth) == M_FALSE)
    {
        Error_Printf("ERROR: main: thread_run FALSE\n");
        goto Error;
    }

    //
	//  User interface
    //
    g_error_display = ERROR_DISPLAY_SLANG;
    g_message_display = MESSAGE_DISPLAY_SLANG;
    p_Screen = Create_Screen(p_Song);
    if (p_Screen == NULL)
    {
        Error_Printf("ERROR: main: cannot create Screen\n");
        goto Error;
    }

    //
	// Forground loop -- listen for the key strokes
    //
    for(;;)
    {
        //
        //  Listen for keyboard (PC) entry
        //
        if (screen_pollScreenRec(p_Screen))
        {
            //
            //  Quit -- tell the midi thread to stop and break out
            //
            LogFile_Printf("I999: main: break from pollScreen\n"); 
            thread_stop(pth);
            usleep(100000);
            break;
        }
        //
        //  List for recorded key strokes
        //
        (void)screen_pollRecord(p_Screen);
    }


#ifdef SYNCRONIZE_EXIT
    //
    //  syncronization-on-exit loop
    //
    while (g_Thread_Loop_Done == M_FALSE && break_count < MAX_BREAK_COUNT)
    {
        usleep(10000);
        break_count++;
        if ((break_count % 50) == 0)
        {
            exit_count++;
            screen_exitWaitmsg(p_Screen, (exit_count % 80));
        }
    }
    if (break_count >= MAX_BREAK_COUNT)
        LogFile_Printf("WARNING: breaking out of sync loop waiting for"
                       " g_Thread_Loop_Done\n");
#endif // SYNCRONIZE_EXIT


Main_Done:
    //
    //  Done -- clean up
    //
    Delete_Thread(pth);
    Delete_Bank(p_Banks);
    if (pp_BankAry)
        free(pp_BankAry);
    Delete_RecordEvent(p_REvents);
    Delete_RecordEvent(p_Rec);
    Delete_Screen(p_Screen);
    Delete_Song(p_Song);
    Delete_MidiControl(p_MidiControl);
    LogFile_Exit();
    return 0;

Error:

    Delete_Thread(pth);
    Delete_Bank(p_Banks);
    if (pp_BankAry)
        free(pp_BankAry);
    Delete_RecordEvent(p_REvents);
    Delete_RecordEvent(p_Rec);
    Delete_Song(p_Song);
    Delete_MidiControl(p_MidiControl);
    LogFile_Exit();
    return 1;
}
//  function -- main
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  read_settings
//
//  Description
//
//      Read the settings in the .attrkrc file in the current directory.
//
//  Arguments
//
//      Void
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
int read_settings(void)
{
    FILE *file=NULL;
    char *pch, *plf;
    static char Line[MAX_LINE_LENGTH];
    char seps[5] = {0x09, 0x0A, 0x0D, 0x20, 0x00};
    struct Bank *curr_bk = NULL;
    int bank_count = 0;

    file = fopen(".attrkrc", "r");
    if (file)
    {
        //
        //  Read each line of the file
        //
        while (feof(file)==0)
        {
            char *ptok;

            pch = fgets(Line, MAX_LINE_LENGTH-1, file);
            if (pch == NULL)
                break;

            Line[MAX_LINE_LENGTH-1] = 0x00;

            if (strlen(Line) == 0)
                continue;

            //
            //  Parse the line
            //
            ptok = strtok(Line, seps);
            if (ptok)
            {
                char *parg;

                //
                //  skip commented lines
                //
                if (strchr(ptok, '#') == ptok)
                    continue;

                plf = strchr(ptok, 0x0D);
                if (plf)
                    *plf = 0x00;
                plf = strchr(ptok, 0x0A);
                if (plf)
                    *plf = 0x00;

                if (strcmp(ptok, "device") == 0)
                {
                    //
                    //  A MIDI output device
                    //

                    parg = strtok(NULL, seps);
                    if (parg)
                    {
                        if (dev_count < MAX_DEVICES)
                        {
                            char *peol;
                            peol = strchr(parg, 0x0D);
                            if (peol)
                                *peol = 0x00;
                            peol = strchr(parg, 0x0A);
                            if (peol)
                                *peol = 0x00;
                            strcpy(dev_names[dev_count++], parg);
                            LogFile_Printf(".attrkrc : dev %d, %s\n",
                                   dev_count, parg);
                        }
                    }
                }
                else if (strcmp(ptok, "g_Record_Session") == 0)
                {
                    //
                    //  Flag that turns on the session record feature
                    //

                    g_Record_Session = M_TRUE;
                    LogFile_Printf(".attrkrc : RECORD SESSION\n");
                }
                else if (strcmp(ptok, "g_Playback_Session") == 0)
                {
                    //
                    //  Flag that turns on the session playback feature
                    //

                    parg = strtok(NULL, seps);
                    if (parg)
                    {
                        g_Playback_Session = M_TRUE;
                        strcpy(g_Playback_Filename, parg);
                        LogFile_Printf(".attrkrc : PLAYBACK SESSION %s\n",
                                parg);
                    }
                }
                else if (strcmp(ptok, "beatpulse") == 0)
                {
                    //
                    //  Flag that turns on the beat pulse feature
                    //

                    beatpulse = M_TRUE;
                    LogFile_Printf(".attrkrc : BEAT PULSE on\n");
                }
                else if (strcmp(ptok, "rt_clock_frequency") == 0)
                {
                    //
                    //  Real time clock frequency setting
                    //

                    parg = strtok(NULL, seps);
                    if (parg)
                    {
                        rtclkfreq = atoi(parg);
                        LogFile_Printf(".attrkrc : RT CLOCK FREQUENCY %d\n",
                                atoi(parg));
                    }
                }
                else if (strcmp(ptok, "external_clock") == 0)
                {
                    //
                    //  Flag that turns on the external-sync clock
                    //

                    intclock = M_FALSE;
                    LogFile_Printf(".attrkrc : EXTERNAL CLOCK on\n");
                }
                else if (strcmp(ptok, "realtime") == 0)
                {
                    //
                    //  Flag that turns on REAL TIME
                    //

                    realtime = M_TRUE;
                    LogFile_Printf(".attrkrc : REAL TIME on\n");
                }
                else if (strcmp(ptok, "bank") == 0)
                {
                    int argcnt, msg, msg_value;
                    struct BankSel *bs_head = NULL,
                                   *curr_bs = NULL;
                    //
                    //  A bank setting
                    //

                    argcnt = 0;
                    parg = strtok(NULL, seps);
                    while (parg)
                    {

                        if (strlen(parg) == 0)
                        {
                            parg = strtok(NULL, seps);
                            continue;
                        }
                        if (strchr(parg, '#') == parg)
                        {
                            break;
                        }

                        if (argcnt % 2 == 0)
                        {
                            msg = atoi(parg);
                        }
                        else
                        {
                            msg_value = atoi(parg);
                            if (bs_head == NULL)
                            {
                                bs_head = Create_BankSel(msg,msg_value);
                                curr_bs = bs_head;
                            }
                            else
                            {
                                curr_bs->next = Create_BankSel(msg,msg_value);
                                curr_bs = curr_bs->next;
                            }
                        }

                        argcnt++;

                        //
                        //  get next Msb
                        //
                        parg = strtok(NULL, seps);

                    } // while -- parg

                    //
                    //  Add a Bank preset to the list
                    //
                    if (bs_head)
                    {
                        struct BankSel *tmpbs;

                        if (p_Banks == NULL)
                        {
                            p_Banks = Create_Bank(bs_head);
                            curr_bk = p_Banks;
                        }
                        else
                        {
                            curr_bk->next = Create_Bank(bs_head);
                            curr_bk = curr_bk->next;
                        }

                        LogFile_Printf(".attrkrc : bank %d :", bank_count+1);
                        tmpbs = curr_bk->msg;
                        while (tmpbs)
                        {
                            LogFile_Printf(" %d %d",
                                           tmpbs->number, tmpbs->value);
                            tmpbs = tmpbs->next;
                        }
                        LogFile_Printf("\n");
                        bank_count++;
                        g_Max_Bank++;
                    }
                }

            }
        }

        //
        //  Create an array of pointers to the Bank Presets (if any)
        //
        if (p_Banks)
        {
            int bj;

            pp_BankAry = (struct Bank **)malloc(
                          sizeof(struct Bank *) * bank_count
                          );
            if (pp_BankAry == NULL)
            {
                LogFile_Printf(".attrkrc : mem alloc on Bank settings\n");
                goto Error;
            }

            //
            //  Assign each Bank pointer to an element in the array
            //
            bj = 0;
            curr_bk = p_Banks;
            while (curr_bk)
            {
                pp_BankAry[bj] = curr_bk;
                bj++;
                curr_bk = curr_bk->next;
            }
        }

        //
        //  clean up
        //
        fclose(file);
    }
    else
        return -1;

    return 0; // success

Error:

    fclose(file);
    Delete_Bank(p_Banks);
    return -1;
}
//  function -- read_settings
//  -------------------------------------------------------------------------

