/**
 * C code modifications Copyright (C) 2009 Alfred Steffens Jr.
 * based on TekTracker, Copyright (C) 2001 Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#ifdef ALSA_DRIVER
#include <alsa/asoundlib.h>
#endif
#include <defs.h>
#include <xmltools.h>
#include <song.h>
#include <config.h>
#include <defs.h>
#include <logprintf.h>
#include <errorprintf.h>
#include <msgprintf.h>
#include <utils.h>
#include <thread.h>
#include <rtctimer.h>
#include <mididev.h>
#include <rawmidi.h>
#include <midicontrol.h>
#include <song.h>

static void diffTimes( struct timeval *result, struct timeval *big,
	struct timeval *little);
static char devtester[10] = "/dev/midi";

extern struct Song *p_Song;
extern int g_ThreadOk;
extern int g_Thread_Loop_Done;

//
//  MIDI Bank Settings
//
extern struct Bank *p_Banks;
struct Bank **pp_BankAry;
extern int g_Max_Bank;




//  -------------------------------------------------------------------------
//  Function    --  Create_MidiControl
//
//  Description
//
//      Constructor for the MidiControl struct instances.
//
//  Arguments
//
//      Initializaers for the MidiControl struct.  Legacy design from
//      original Tektracker.
//
//  Return
//
//      Returns a new instance of the MidiControl struct.
//
//  -------------------------------------------------------------------------
struct MidiControl *Create_MidiControl(int useintclock, int isrt, int bp,
                                       int bcount, int rtcfreq,
                                       int midiclockcounter, int wasplaying)
{
struct MidiControl *mctrl;

    mctrl = (struct MidiControl *)malloc( sizeof(struct MidiControl) );
    if (mctrl == NULL)
    {
        Error_Printf("E100: Create_MidiControl: mem alloc\n");
        return NULL;
    }

    mctrl->mididev = NULL;
    mctrl->useintclock = useintclock;
    mctrl->isrt = isrt;
    mctrl->bp = bp;
    mctrl->bcount = bcount;
    mctrl->rtcfreq = rtcfreq;       // 1024 or up to 8192, 64 if not root
    mctrl->last_sec = 0;
    mctrl->last_usec = 0;
    mctrl->last_diff = 0;
    mctrl->rtc = NULL;
    mctrl->midiclockcounter = midiclockcounter;
    mctrl->wasplaying = wasplaying;
    mctrl->alsa = 0;
    mctrl->next = NULL;

    return mctrl;
}
//  function -- Create_MidiControl
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_Close
//
//  Description
//
//      Stops the instruments and flushes the messages queues.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_Close(struct MidiControl *mctrl)
{
    struct RawMidi *mididev;

    ERROR_NULL_VOID(mctrl);

    mctrl_shutUp(mctrl);
    mididev = mctrl->mididev;
    mdev_flush(mididev);
}
//  function -- mctrl_Close
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_MidiControl
//
//  Description
//
//      destructor for the MidiControl struct instance.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void Delete_MidiControl(struct MidiControl *mctrl)
{
    struct MidiControl *pN, *pM;

    if (mctrl == NULL)
        return;

    pM = mctrl;
    while(pM)
    {
        pN = pM;
        pM = pM->next;
        mctrl_Close(pN);
        Delete_RawMidi(pN->mididev);
        free(pN);
    }
}
//  function -- Delete_MidiControl
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_openMidiDevice
//
//  Description
//
//      Opens a MIDI device.  If this is not the first MIDI device opened
//      it will add this device to the end of a linked list of devices.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      filename
//
//          Pointer to string containing the device name of the MIDI
//          device.
//
//  Return
//
//      Zero on success, nonzero on failure.
//
//  -------------------------------------------------------------------------
int mctrl_openMidiDevice(struct MidiControl *mctrl,  const char *filename)
{
    struct RawMidi *last_mididev = NULL, *this_mididev = NULL;
    int last_port = 0, this_port = 0;
    static char lfilename[MAX_PATH_LENGTH];
    int alsa = 0;

    ERROR_NULL(mctrl,-1);

    //
    //  Get a pointer to the last device created
    //
    last_mididev = Find_Last_RawMidi(mctrl->mididev);
    if (last_mididev)
    {
        last_port = last_mididev->port;
        this_port = last_port + 1;
    }

    //
	// Now open the device.  I'm scared this might block sometimes, so we
	// had better announce what's going on to give a clue..
    //
	if( filename ) {
        //
        //  We try to open the specified device filename
        //
        strncpy(lfilename, filename, MAX_PATH_LENGTH-1);
        lfilename[MAX_PATH_LENGTH-1] = 0x00;
		Msg_Printf("I001: Using raw MIDI device %s\n", lfilename );
	} else {
        //
        //  Or we default to open the next sequential device
        //
        sprintf(lfilename, "/dev/midi%02d", this_port);
		Msg_Printf("I002: Using raw MIDI device %s.\n", lfilename);
	}

    #ifdef ALSA_DRIVER
    //
    //  Determine if an ALSA driver should be opened, based on the type
    //  of name of the device.
    //
    if (strstr(filename, devtester) == filename)
    {
        //
        //  Found an OSS-style device name
        //
        alsa = 0;
    }
    else
    {
        //
        //  Anuthing else is interpreted as ALSA
        //
        alsa = 1;
    }
    #endif

    //
    //  Force ALSA mode depending on the type of device name
    //
    mctrl->alsa = alsa;

    //
    //  Open the next device
    //
    if (last_mididev)
    {
        last_mididev->next = Create_RawMidi(lfilename, MDEV_READWRITEMODE,
                                            mctrl->alsa);
        this_mididev = last_mididev->next;
    }
    else
    {
        this_mididev = Create_RawMidi(lfilename, MDEV_READWRITEMODE,
                                      mctrl->alsa);
        mctrl->mididev = this_mididev;
    }

    if (this_mididev == NULL)
    {
        return -2;
    }

    this_mididev->port = this_port;
    return 0;
}
//  function -- mctrl_openMidiDevice
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doNote
//
//  Description
//
//      High-level function to play a note, given the SongChannel,
//      SongPattern and beat number.
//
//      Called by function mctrl_doCurrentNotes, which is called on every
//      MIDI thread loop.
//      Calls mdev_noteOn.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      chan
//
//          The SongChannel (column on the screen) where the pattern is
//          located.
//
//      pat
//
//          The SongPattern in which the current note is located.
//
//      beat
//
//          The current beat number in this SongPattern.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doNote(struct MidiControl *mctrl,  struct SongChannel *chan,
                         struct SongPattern *pat, unsigned int beat)
{
    struct RawMidi *mididev = NULL;
	int data, last_note, mchan, mport, note, slide, mute;

    ERROR_NULL_VOID(mctrl);
    ERROR_NULL_VOID(chan);
    ERROR_NULL_VOID(pat);
    mididev = mctrl->mididev;

    //
    //  Initialize locals
    //
    data = spat_getVolume(pat, beat);

    last_note = schan_getLastNote(chan);
    mchan = schan_getMidiChannel(chan);     // midi channel num, [0..n-1]
    mport = schan_getMidiPort(chan);
    note = spat_getNote(pat, beat);
    slide = spat_getSlide(pat, beat);
    mute = schan_isMuted(chan);

    //
	// Shut up if we're in mute mode or a NoteOff is reached.
    //
    if (mute || note == SONG_NOTEOFF) {
        if (last_note)
            mdev_noteOff(mididev, mchan, mport, last_note, data);
		schan_setLastNote(chan, 0);
		return;
	}

    //
	// If we have no note to play, we're done here.
    //
	if( note < 0 )
        return;

    if (slide)
    {
        //
        // Play legato.
        //
        mdev_noteOn(mididev, mchan, mport, note, data);
        if (last_note)
            mdev_noteOff(mididev, mchan, mport, last_note, data);
    }
    else
    {
        //
		// No legato.
        //
        if (last_note)
            mdev_noteOff(mididev, mchan, mport, last_note, data);
        mdev_noteOn(mididev, mchan, mport, note, data);
	}

    //
	// Remember to set the last note so we can shut it up later.
    //
	schan_setLastNote(chan, note);
}
//  function -- mctrl_doNote
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doProgramChange
//
//  Description
//
//      Executes a program change MIDI message.
//
//      Called by function mctrl_doCurrentNotes, which is called on every
//      MIDI thread loop.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      chan
//
//          The SongChannel (column on the screen) where the pattern is
//          located.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doProgramChange(struct MidiControl *mctrl, struct SongChannel *chan)
{
    struct RawMidi *mididev = NULL;
    int mchan, mport;

    ERROR_NULL_VOID(mctrl);
    ERROR_NULL_VOID(chan);
    mididev = mctrl->mididev;

    //
    // If we have no program change
    //
    if(chan->prchg_ready == 0)
        return;

    //
    //  Initialize locals
    //
    mchan = schan_getMidiChannel(chan);
    mport = schan_getMidiPort(chan);

    //
    //  Bank switch on the Yamaha Motif series on instruments
    //
    if( chan->bnchg_ready )
    {
        int banknum;

        banknum = schan_getBankNumber(chan)-1;
        if (banknum >= 0)
        {
            //
            //  First, try a list of Bank presets
            //
            if (pp_BankAry && banknum <= g_Max_Bank)
            {
                struct BankSel *curr_bs;
                //
                //  Loop through the bank select messages
                //
                curr_bs = pp_BankAry[banknum]->msg;
                while (curr_bs)
                {
                    mdev_controlChange(mididev, mchan, mport,
                                       curr_bs->number, curr_bs->value);
                    curr_bs = curr_bs->next;
                }
            }
            else
            {
                //
                //  Otherwise, use the Yamaha ES as the default
                //
                mdev_controlChange(mididev, mchan, mport, 0, 63);
                mdev_controlChange(mididev, mchan, mport, 32, banknum);
            }
        }
        chan->bnchg_ready = 0;
    }
    mdev_programChange(mididev, mchan, mport, schan_getProgramNumber(chan)-1);
    chan->prchg_ready = 0;
}
//  function -- mctrl_doProgramChange
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doVolumeChange
//
//  Description
//
//      Executes a volume change MIDI message.
//
//      Called by function mctrl_doCurrentNotes, which is called on every
//      MIDI thread loop.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      chan
//
//          The SongChannel (column on the screen) where the pattern is
//          located.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doVolumeChange(struct MidiControl *mctrl, struct SongChannel *chan)
{
    struct RawMidi *mididev = NULL;
    int mchan, mport;

    ERROR_NULL_VOID(mctrl);
    ERROR_NULL_VOID(chan);
    mididev = mctrl->mididev;

    //
    // If we have no volume change
    //
    if(chan->vlchg_ready == 0)
        return;

    //
    //  Initialize locals
    //
    mchan = schan_getMidiChannel(chan);
    mport = schan_getMidiPort(chan);

    mdev_controlChange(mididev, mchan, mport, 7, chan->volume);
    chan->vlchg_ready = 0;
}
//  function -- mctrl_doVolumeChange
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doControlChange
//
//  Description
//
//      Executes a control change MIDI message.
//
//      Called by function mctrl_doCurrentNotes, which is called on every
//      MIDI thread loop.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      ichan
//
//          The number of the SongChannel to send a MIDI control message on.
//
//      ctrlnumber
//
//          The MIDI control number (volume = 7).
//
//      value
//
//          The value of the MIDI control message (0--127).
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doControlChange(struct MidiControl *mctrl, int ichan,
                           int ctrlnumber, int value)
{
    struct RawMidi *mididev = NULL;
    int mchan, mport;
    struct SongChannel *chan;

    ERROR_NULL_VOID(mctrl);
    mididev = mctrl->mididev;

    chan = song_getChannel(p_Song, ichan);
    if (chan == NULL)
    {
        Error_Printf("E104: mctrl_doControlChange: NULL channel\n");
        return;
    }
    

    //
    //  Initialize locals
    //
    mchan = schan_getMidiChannel(chan);
    mport = schan_getMidiPort(chan);

    mdev_controlChange(mididev, mchan, mport, ctrlnumber, value);
}
//  function -- mctrl_doControlChange
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doControlSetupChange
//
//  Description
//
//      Called by mctrl_doSetupMessages, called by song_armMidiSetupMessages,
//      called by song_refreshMidiSetup, executed by the ":m r" command.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      chan
//
//          The SongChannel (column on the screen) where the pattern is
//          located.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doControlSetupChange(struct MidiControl *mctrl,
                                struct SongChannel *chan)
{
    struct RawMidi *mididev = NULL;
    struct SChanMidiCtrl *curr_sctrl;
    int mchan, mport;

    ERROR_NULL_VOID(mctrl);
    ERROR_NULL_VOID(chan);
    mididev = mctrl->mididev;

    if (chan->ctrl_msg == NULL)
        return;

    //
    //  Initialize locals
    //
    mchan = schan_getMidiChannel(chan);
    mport = schan_getMidiPort(chan);

    curr_sctrl = chan->ctrl_msg;
    while (curr_sctrl)
    {
        if (curr_sctrl->chg_ready)
        {
            mdev_controlChange(mididev, mchan, mport,
                           curr_sctrl->control_number,
                           curr_sctrl->control_value);
            mdev_flush(mididev);
            usleep(MIDI_WRITE_PAUSE);
            curr_sctrl->chg_ready = 0;
        }
        curr_sctrl = curr_sctrl->next;
    }
}
//  function -- doControlSetupChange
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_shutUp
//
//  Description
//
//      Turn off all MIDI notes.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_shutUp(struct MidiControl *mctrl)
{
    ERROR_NULL_VOID(mctrl);
    mdev_allNotesOff_all(mctrl->mididev);
}
//  function -- mctrl_shutUp
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  diffTimes
//
//  Description
//
//      MIDI synchronization.  Legacy code from the original Tektracker.
//
//  Arguments
//
//      result
//
//          struct timeval pointer.
//          Result of time diff calculation.
//
//      big
//
//          struct timeval pointer.
//
//      little
//
//          struct timeval pointer.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
static void diffTimes( struct timeval *result, struct timeval *big,
	struct timeval *little )
{
    ERROR_NULL_VOID(result);
    ERROR_NULL_VOID(big);
    ERROR_NULL_VOID(little);

	result->tv_sec = big->tv_sec - little->tv_sec;
	if( big->tv_usec - little->tv_usec < 0 ) {
		result->tv_sec -= 1;
		result->tv_usec = ( 1000 * 1000 ) + big->tv_usec - little->tv_usec;
	} else {
		result->tv_usec = big->tv_usec - little->tv_usec;
	}
}
//  function -- diffTimes
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_rtcTimeCheck
//
//  Description
//
//      Real-time clock "time check".  Called by the MIDI thread loop.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_rtcTimeCheck(struct MidiControl *mctrl)
{
    struct RealTimeClock *rtc = NULL;
	struct timeval tv;
	struct timeval tvdiff;
	struct timeval result;
	int diff;
	int tt;

    ERROR_NULL_VOID(mctrl);
    ERROR_NULL_VOID(p_Song);

    if (g_ThreadOk == 0)
        return;

    //
    //  Initialize locals
    //
    tt = song_getUsecpb(p_Song) / 3;
    rtc = mctrl->rtc;

    //
	// Throttle the box.
    //
    rtc_nextTick(rtc);

    gettimeofday( &tv, 0 );
    diff = (tv.tv_sec - mctrl->last_sec) * 1000 * 1000 +
            tv.tv_usec - mctrl->last_usec;

    if (diff > tt)
    {
        while (diff > tt)
        {
            diff -= tt;
            ++mctrl->midiclockcounter;
            mctrl->midiclockcounter %= 3;
            mdev_syncTick(mctrl->mididev);
            mctrl->bcount++;
            mctrl->bcount %= 24;
            if (mctrl->bcount == 0)
                mctrl->bp = M_TRUE;
            else
                mctrl->bp = M_FALSE;

            if (mctrl->midiclockcounter == 0 && song_isPlaying(p_Song))
            {
                if (mctrl->wasplaying == M_FALSE)
                    mdev_syncStart(mctrl->mididev);
                mctrl->wasplaying = M_TRUE;
                mctrl_doCurrentNotes(mctrl);
            }
            else if (mctrl->wasplaying == M_TRUE && !song_isPlaying(p_Song))
            {
                mctrl->wasplaying = M_FALSE;
                mdev_syncStop(mctrl->mididev);
                mdev_shutUp(mctrl->mididev);
            }

            if (g_ThreadOk == 0)
                return;
        }

        // Now write all the MIDI data.
        mdev_flush(mctrl->mididev);

        tvdiff.tv_sec = 0;
        tvdiff.tv_usec = diff;
        diffTimes(&result, &tv, &tvdiff);

        mctrl->last_sec = result.tv_sec;
        mctrl->last_usec = result.tv_usec;
    }
}
//  function -- mctrl_rtcTimeCheck
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_midiTimeCheck
//
//  Description
//
//      MIDI "time check".  Called by the MIDI thread loop.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_midiTimeCheck(struct MidiControl *mctrl)
{
    int curmessage;

    ERROR_NULL_VOID(mctrl);
    ERROR_NULL_VOID(p_Song);

    if (g_ThreadOk == 0)
        return;

    //
    //  Fixme
    //  This code is for syncing to an external source.
    //  The midi channel is needed in order to know which port to listen on.
    //  For now, the default channel from which to sync is the first channel
    //
    curmessage = mdev_readMessage(mctrl->mididev, 0);

    switch (curmessage)
    {
        case MDEV_NOMESSAGE: return;

        case MDEV_MIDICLOCKTICK:
            mdev_syncTick(mctrl->mididev);

            ++mctrl->midiclockcounter;
            mctrl->midiclockcounter %= 3;
            mctrl->bcount++;
            mctrl->bcount %= 24;
            if (mctrl->bcount == 0)
                mctrl->bp = M_TRUE;
            else
                mctrl->bp = M_FALSE;

            if (mctrl->midiclockcounter == 0 && song_isPlaying(p_Song))
                mctrl_doCurrentNotes(mctrl);
            break;

        case MDEV_MIDISTART:
            mdev_syncStart(mctrl->mididev);

			// Reset everything to the start
			mctrl->midiclockcounter = 0;
            song_rewind(p_Song);
            song_start(p_Song);

            // Better send out the current notes now.
            mctrl_doCurrentNotes(mctrl);
            break;

        case MDEV_MIDISTOP:
            mdev_syncStop(mctrl->mididev);
            song_stop(p_Song);
            mdev_shutUp(mctrl->mididev);
            break;

        case MDEV_MIDICONTINUE:
            mdev_syncContinue(mctrl->mididev);
            song_start(p_Song);
            break;
    }

    mdev_flush(mctrl->mididev);
}
//  function -- mctrl_midiTimeCheck
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_isMidiDeviceOpen
//
//  Description
//
//      TBD.
//      Checks if the specific MIDI device has been opened.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      port
//
//          Integer number of the port (MIDI device).
//  Return
//
//      Integer (boolean) indicating open or closed.
//
//  -------------------------------------------------------------------------
int mctrl_isMidiDeviceOpen(struct MidiControl *mctrl, int port)
{
    ERROR_NULL(mctrl,M_FALSE);

    if (Find_Nth_RawMidi(mctrl->mididev, 0))
        return M_TRUE;

    return M_FALSE;
}
//  function -- mctrl_isMidiDeviceOpen
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_isRealtime
//
//  Description
//
//      Query if the current. isrt flag is set.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Integer (boolean) indicating the state.
//
//  -------------------------------------------------------------------------
int mctrl_isRealtime(struct MidiControl *mctrl)
{
    ERROR_NULL(mctrl,M_FALSE);

    return mctrl->isrt;
}
//  function -- mctrl_isRealtime
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_useInternalClock
//
//  Description
//
//      Sets the state flag whether to use the real-time clock.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_useInternalClock(struct MidiControl *mctrl)
{
    ERROR_NULL_VOID(mctrl);
    mctrl->useintclock = M_TRUE;
}
//  function -- mctrl_useInternalClock
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_useExternalClock
//
//  Description
//
//      Sets the state flag whether to use the external clock.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_useExternalClock(struct MidiControl *mctrl)
{
    ERROR_NULL_VOID(mctrl);
    mctrl->useintclock = M_FALSE;
}
//  function -- mctrl_useExternalClock
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_usingInternalClock
//
//  Description
//
//      Query the state flag whether using the internal clock.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//  -------------------------------------------------------------------------
int mctrl_usingInternalClock(struct MidiControl *mctrl)
{
    ERROR_NULL(mctrl,M_FALSE);
    return mctrl->useintclock;
}
//  function -- mctrl_usingInternalClock
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_beatPulse
//
//  Description
//
//      Gets the current "beat pulse" flag.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      The beat pulse.
//
//  -------------------------------------------------------------------------
int mctrl_beatPulse(struct MidiControl *mctrl)
{
    ERROR_NULL(mctrl,0);
    return mctrl->bp;
}
//  function -- mctrl_beatPulse
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_setRTCFrequency
//
//  Description
//
//      Sets the real-time clock frequency.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//      newfreq
//
//          The new frequency.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_setRTCFrequency(struct MidiControl *mctrl, int newfreq)
{
    ERROR_NULL_VOID(mctrl);
    mctrl->rtcfreq = newfreq;
}
//  function -- mctrl_setRTCFrequency
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doCurrentNotes
//
//  Description
//
//      Executes the MIDI note messages from the top-level.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doCurrentNotes(struct MidiControl *mctrl)
{
    int i;

    ERROR_NULL_VOID(mctrl);

    //
    // Play notes.
    //
    for(i = 0; i < SONGMAXCHAN; i++ ) {
        struct SongChannel *curchan;
        struct SongPattern *curpat;

        curchan = song_getChannel(p_Song, i);
        if (curchan == NULL)
        {
            Error_Printf("E104: mctrl_doCurrentNotes: NULL channel\n");
            return;
        }

        curpat = schan_getSongPattern(curchan, song_getCurPat(p_Song));
        if (curpat)
        {
            mctrl_doNote(mctrl, curchan, curpat, song_getCurBeat(p_Song));
            mctrl_doProgramChange(mctrl, curchan);
            mctrl_doVolumeChange(mctrl, curchan);
        }
    }

    // Increment the beat counter.
    song_incrBeat(p_Song);
}
//  function -- mctrl_doCurrentNotes
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doCurrentMessages
//
//  Description
//
//      Executes the MIDI messages from the top-level.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doCurrentMessages(struct MidiControl *mctrl)
{
    int i;

    ERROR_NULL_VOID(mctrl);

    //
    // Update all the Midi (control and program) messages
    //
    for(i = 0; i < SONGMAXCHAN; i++ ) {
        struct SongChannel *curchan;
        struct SongPattern *curpat;

        curchan = song_getChannel(p_Song, i);
        if (curchan == NULL)
        {
            Error_Printf("E104: mctrl_doCurrentMessages: NULL channel\n");
            return;
        }

        curpat = schan_getSongPattern(curchan, song_getCurPat(p_Song));
        if (curpat)
        {
            mctrl_doVolumeChange(mctrl, curchan);
            mdev_flush(mctrl->mididev);
            usleep(10000L);

            mctrl_doProgramChange(mctrl, curchan);
            mdev_flush(mctrl->mididev);
            usleep(10000L);
        }
    }
}
//  function -- mctrl_doCurrentMessages
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mctrl_doSetupMessages
//
//  Description
//
//      Send the MIDI setup messages for this SongChannel.
//
//  Arguments
//
//      mctrl
//
//          The MidiControl struct for the current song.
//
//      schan
//
//          The SongChannel pointer for the current SongChannel.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void mctrl_doSetupMessages(struct MidiControl *mctrl, struct SongChannel *schan)
{
    ERROR_NULL_VOID(mctrl);

    mctrl_doVolumeChange(mctrl, schan);
    mdev_flush(mctrl->mididev);
    usleep(MIDI_WRITE_PAUSE);

    mctrl_doProgramChange(mctrl, schan);
    mdev_flush(mctrl->mididev);
    usleep(MIDI_WRITE_PAUSE);

    //
    //  MIDI setup messages for the linked list of control messages
    //
    if (schan->ctrl_msg)
    {
        mctrl_doControlSetupChange(mctrl, schan);
    }
}
//  function -- mctrl_doSetupMessages
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  thread_main
//
//  Description
//
//      The top-level of the MIDI event thread.
//
//  Arguments
//
//      mctrl
//
//          Pointer to the current MidiControl instance.
//
//  Return
//
//      Integer error code.  Returns zero on success.
//
//  -------------------------------------------------------------------------
int thread_main(struct MidiControl *mctrl)
{
    struct timeval tv;
    int rt = -1;
    struct RealTimeClock *rtc = NULL;

    ERROR_NULL(mctrl,-1);

    //
    // If we're root, get realtime priority.
    //
    if (mctrl->isrt == M_TRUE)
    {
        LogFile_Printf("I900: requesting REAL TIME thread priority\n");
        rt = set_realtime_priority();
        if (rt < 0)
            LogFile_Printf("I901: failure to set REAL TIME priority"
                           " (must be ROOT for that)\n");
        
    }
    else
        LogFile_Printf("I902: requesting normal thread priority\n");

    //
    // Initialize for internalTimeCheck().
    //
    gettimeofday( &tv, 0 );
    mctrl->last_sec = tv.tv_sec;
    mctrl->last_usec = tv.tv_usec;

    //
    // Initialize for rtcTimeCheck();
    //
    rtc = Create_RealTimeClock();
    if (rtc == NULL)
    {
        Error_Printf("E105: thread_main: NULL pointer from"
                     " Create_RealTimeClock\n");
        return -1;
    }
    mctrl->rtc = rtc;
    if( rt < 0 ) {
        //
        // We're not root.  Setup for shitty rtc resolution.
        //
        rtc_setInterval(mctrl->rtc, (int)DEFAULT_RTCLOCK);
        mctrl->isrt = M_FALSE;
        LogFile_Printf("I903: thread_main: setting LOW frequency %d\n",
                       (int)DEFAULT_RTCLOCK);
    }
    else
    {
        rtc_setInterval(mctrl->rtc, mctrl->rtcfreq);
        mctrl->isrt = M_TRUE;
        LogFile_Printf("I904: thread_main: setting REAL TIME frequency"
                        " %d\n", mctrl->rtcfreq);
    }
    rtc_startClock(mctrl->rtc);
    mctrl->last_diff = 0;

    for(;;) {

        usleep(MIDI_THREAD_BREATH);

        thread_testcancel();

        //
        //  First try gentle fall-through to stop thread
        //
        if (g_ThreadOk == 0)
        {
            LogFile_Printf("I905: thread_main: g_ThreadOk: returning"
                           " from thread\n");

#ifdef SYNCRONIZE_EXIT
            g_Thread_Loop_Done = M_TRUE;
#endif // SYNCRONIZE_EXIT

            return 0;
        }

        if( mctrl->useintclock )
        {
            mctrl_rtcTimeCheck(mctrl);
        }
        else
        {
            mctrl_midiTimeCheck(mctrl);
        }
    }

    return 0;
}
//  function -- thread_main
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_BankSel
//
//  Description
//
//      Constructor for the BankSel struct.
//
//  Arguments
//
//      number
//
//          Integer in range 0-127.  The control number.
//
//      value
//
//          Integer in range 0-127.  The control value.
//
//  Return
//
//      Pointer to an allocated BankSel struct.
//
//  -------------------------------------------------------------------------
struct BankSel *Create_BankSel(int number, int value)
{
    struct BankSel *bs;
    int ctrl_number, ctrl_value;

    ctrl_number = number;
    ctrl_value = value;
    if (ctrl_number < 0)
        ctrl_number = 0;
    if (ctrl_number > 127)
        ctrl_number = 127;
    if (ctrl_value < 0)
        ctrl_value = 0;
    if (ctrl_value > 127)
        ctrl_value = 127;
        
    bs = (struct BankSel *)malloc( sizeof(struct BankSel) );
    if (bs == NULL)
    {
        Error_Printf("E106: Create_BankSel: mem alloc\n");
        return NULL;
    }

    bs->number = ctrl_number;
    bs->value = ctrl_value;
    bs->next = NULL;

    return bs;
}
//  function -- Create_BankSel
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_BankSel
//
//  Description
//
//      Deletes a linked list of BankSel structs.
//
//  Arguments
//
//      Pointer to the head of the linked list.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void Delete_BankSel(struct BankSel *bs)
{
    struct BankSel *pB, *pN;

    if (bs == NULL)
        return;

    pB = bs;
    while (pB)
    {
        pN = pB;
        pB = pB->next;
        free(pN);
    }
}
//  function -- Delete_BankSel
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_Bank
//
//  Description
//
//      Constructor for the Bank struct. (see mctrl_doProgramChange).
//
//  Arguments
//
//      Pointer to the head of a linked list of BankSel structs.
//
//  Return
//
//      Pointer to the allocated Bank struct.
//
//  -------------------------------------------------------------------------
struct Bank *Create_Bank(struct BankSel *bs)
{
    struct Bank *bk;

    bk = (struct Bank *)malloc( sizeof(struct Bank) );
    if (bk == NULL)
    {
        Error_Printf("E107: Create_Bank: mem alloc\n");
        return NULL;
    }

    bk->msg = bs;
    bk->next = NULL;

    return bk;
}
//  function -- Create_Bank
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_Bank
//
//  Description
//
//      Deletes a linked list of Bank structs.
//
//  Arguments
//
//      Pointer to the head of the linked list.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void Delete_Bank(struct Bank *bk)
{
    struct Bank *pB, *pN;

    if (bk == NULL)
        return;

    pB = bk;
    while (pB)
    {
        pN = pB;
        pB = pB->next;
        Delete_BankSel(pN->msg);
        free(pN);
    }
}
//  function -- Delete_Bank
//  -------------------------------------------------------------------------


