/**
 * C-code modifications Copyright (C) 2009 Alfred Steffens Jr.
 * based on TekTracker, Copyright (C) 2001 Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif // HAVE_CONFIG_H
#include <slang.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdarg.h>
#include <defs.h>
#include <logprintf.h>
#include <errorprintf.h>
#include <xmltools.h>
#include <song.h>
#include <patbar.h>
#include <screen.h>

extern struct Song *p_Song;


//  -------------------------------------------------------------------------
//  Function    --  Init_PatBar
//
//  Description
//
//      Initialization of a PatBar instance.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      initchan
//
//          It's pattern column on the current screen (0--7)
//
//      startrow
//
//          The row on the text screen where this PatBar starts drawing.
//
//      startcol
//
//          The column on the text screen where this PatBar starts drawing.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void Init_PatBar(struct PatBar *pbar, int initchan, int startrow, int startcol)
{
    ERROR_NULL_VOID(pbar);

    pbar->chan = initchan;
	pbar->pos = 0;
	pbar->row = startrow;
	pbar->col = startcol;
	pbar->cursor = -1;
	pbar->scrolltop = 0;
}
//  function -- Init_PatBar
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_PatBar
//
//  Description
//
//      Constructor for the PatBar struct instance.
//
//  Arguments
//
//      initchan
//
//          It's pattern column on the current screen (0--7)
//
//      startrow
//
//          The row on the text screen where this PatBar starts drawing.
//
//      startcol
//
//          The column on the text screen where this PatBar starts drawing.
//
//  Return
//
//      Pointer to new PatBar struct instance.
//
//  -------------------------------------------------------------------------
struct PatBar *Create_PatBar(int initchan, int startrow, int startcol)
{
    struct PatBar *pbar;

    pbar = (struct PatBar *)malloc(sizeof(struct PatBar));
    if (pbar == NULL)
    {
        Error_Printf("E110: Create_PatBar: mem alloc\n");
        return NULL;
    }

    Init_PatBar(pbar, initchan, startrow, startcol);
    return pbar;
}
//  function -- Create_PatBar
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_setPosition
//
//  Description
//
//      Called by screen_setPosition().
//      Sets the cursor position in the new row due to user typing arrow
//      keys.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      newpos
//
//          New row position on the screen for the cursor.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_setPosition(struct PatBar *pbar,  int newpos)
{
    if (pbar == NULL)
        return;

    if (newpos < 0)
        newpos = 0;
    if (newpos > SONGLEN)
        newpos = SONGLEN;
    pbar->pos = newpos;
}
//  function -- pbar_setPosition
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_setNote
//
//  Description
//
//      Called by screen_setNote(), and calls spat_setNote().
//      The user is setting a MIDI note number at the current note
//      position (row) in the current pattern in the current SongChannel.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      note
//
//          The MIDI note number (0--127).
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_setNote(struct PatBar *pbar, int note)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E111: pbar_setNote: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat == NULL)
    {
        int i;

        for (i = 0; i < SONGMAXPAT; i++ )
        {
            pat = schan_getSongPattern(schan, i);
            if (spat_isBlank(pat))
            {
                schan_setSongPattern(schan, pbar->pos, i + 1);
                pat = schan_getSongPattern(schan, pbar->pos);
                break;
            }
        }
    }

	assert(pbar->cursor != -1);
	spat_setNote(pat, pbar->cursor, note);
	pbar_printPatLine(pbar, pbar->cursor);
}
//  function -- pbar_setNote
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_refresh
//
//  Description
//
//      Redraw the current PatBar on the screen after changes were made.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_refresh(struct PatBar *pbar)
{
    struct SongChannel *schan;
    struct SongPattern *pat;
	int i, muted, scrsize;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E112: pbar_refresh: NULL SongChannel\n");
        return;
    }

    muted = schan_isMuted(schan);
    pat = schan_getSongPattern(schan, pbar->pos);

    SLsmg_gotorc(pbar->row, pbar->col);
    SLsmg_printf( "%-8s", schan_getName(schan));

    SLsmg_gotorc(pbar->row + 1, pbar->col);
    SLsmg_printf( "%02x ", pbar->chan);
    if (pat)
    {
        SLsmg_printf( "%02x ", spat_getIndex(pat) );
    }
    else
    {
        SLsmg_printf( "-- " );
    }
    if (muted)
    {
        SLsmg_printf( "M " );
    }
    else
    {
        SLsmg_printf( "  " );
    }

    SLsmg_gotorc(pbar->row + 2, pbar->col);
    SLsmg_printf( "Vol: %02d ", schan_getVolume(schan));

    SLsmg_gotorc(pbar->row + 3, pbar->col);
    #ifdef OLD_TEKTRAKER
    SLsmg_printf( "Ch : %02d ", schan_getMidiChannel(schan) + 1);
    #endif
    SLsmg_printf( "Ch%1d: %02d ", schan_getMidiPort(schan)+1,
                                  schan_getMidiChannel(schan)+1);

    SLsmg_gotorc(pbar->row + 4, pbar->col);
    if (pat)
    {
        SLsmg_printf( "%-8s", spat_getName(pat) );
    }
    else
    {
        SLsmg_printf( "        " );
    }

    SLsmg_gotorc(pbar->row + 5, pbar->col);
    SLsmg_printf( "========" );

    for (i = 0; i < SONGPATLEN; i++ )
    {
        pbar_printPatLine(pbar, i);
    }
    scrsize = SLtt_Screen_Rows - pbar->row - HEADERSIZE - BOTTOMSIZE;
    if (scrsize > SONGPATLEN)
    {
        int j;

        for(j = SONGPATLEN; j < scrsize; ++j)
        {
            SLsmg_gotorc(pbar->row + HEADERSIZE + j, 0);
            SLsmg_erase_eol();
        }
    }
}
//  function -- pbar_refresh
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_noteToString
//
//  Description
//
//      Converts a MIDI note number (0--127) to a musical notation string.
//
//  Arguments
//
//      str
//
//          String pointer to memory where result will be written.
//
//      note
//
//          The input string number.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_noteToString(char *str, int note)
{
    ERROR_NULL_VOID(str);

    switch (note)
    {
        case SONG_NOTEEMPTY:
            str[0] = '-';
            str[1] = '-';
            str[2] = '-';
            str[3] = 0x00;
        break;

        case SONG_NOTEOFF:
            str[0] = '^';
            str[1] = '^';
            str[2] = '^';
            str[3] = 0x00;
        break;

        default:
            switch (note % 12)
            {
                case 0:
                    str[0] = 'C';
                    str[1] = '-';
                break;
                case 1:
                    str[0] = 'C';
                    str[1] = '#';
                break;
                case 2:
                    str[0] = 'D';
                    str[1] = '-';
                break;
                case 3:
                    str[0] = 'D';
                    str[1] = '#';
                break;
                case 4:
                    str[0] = 'E';
                    str[1] = '-';
                break;
                case 5:
                    str[0] = 'F';
                    str[1] = '-';
                break;
                case 6:
                    str[0] = 'F';
                    str[1] = '#';
                break;
                case 7:
                    str[0] = 'G';
                    str[1] = '-';
                break;
                case 8:
                    str[0] = 'G';
                    str[1] = '#';
                break;
                case 9:
                    str[0] = 'A';
                    str[1] = '-';
                break;
                case 10:
                    str[0] = 'A';
                    str[1] = '#';
                break;
                case 11:
                    str[0] = 'B';
                    str[1] = '-';
                break;
            }
            str[2] = ((char)((float)note / 12.0) - 1) + '0';
            str[3] = 0x00;
            break; // default
    }
}
//  function -- pbar_noteToString
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_printPatLine
//
//  Description
//
//      Prints a single line of the pattern bar on the screen.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      i
//
//          The note position in the current pattern.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_printPatLine(struct PatBar *pbar, int i)
{
	char note[ 4 ] = {'-', '-', '-', 0x00};
	char slide[2] = {'-', 0x00};
	int vol = 64;
	int printrow;
    struct SongChannel *schan;
    struct SongPattern *pat;
	int scrsize;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E113: pbar_printPatLine: NULL SongChannel\n");
        return;
    }

    scrsize = SLtt_Screen_Rows - pbar->row - HEADERSIZE - BOTTOMSIZE;
    if (i < pbar->scrolltop || i >= (pbar->scrolltop + scrsize))
        return;
    printrow = pbar->row + HEADERSIZE + i - pbar->scrolltop;

    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
    {
        pbar_noteToString(note, spat_getNote(pat, i));
        if (i == spat_getEndNote(pat))
        {
            if (spat_getSlide(pat, i) )
                slide[0] = '$';
            else
                slide[0] = '|';
        }
        else
        {
            if (spat_getSlide(pat, i))
                slide[0] = 'S';
        }
        vol = spat_getVolume(pat, i);
    }

    SLsmg_gotorc(printrow, pbar->col);
    if (i == pbar->cursor && i % 4 == 0)
        SLsmg_set_color(HILIGHTCURSOR_COLOR);
    else if (i == pbar->cursor)
        SLsmg_set_color(CURSOR_COLOR);
    else if (i % 4 == 0)
        SLsmg_set_color(HILIGHT_COLOR);
    else if (schan_isMuted(schan))
        SLsmg_set_color(MUTEDCHANNEL_COLOR);
    SLsmg_printf( "%s %02x %s", note, vol, slide);
    SLsmg_set_color(DEFAULT_COLOR);
}
//  function -- pbar_printPatLine
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  pbar_setCursor
//
//  Description
//
//      Sets the screen "cursor" to the new position.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      newpos
//
//          The new cursor position.
//
//  Return
//
//      Returns the cursor position corresponding to the top of the
//      screen.
//
//  -------------------------------------------------------------------------
int pbar_setCursor(struct PatBar *pbar, int newpos)
{
	int scrsize;
    int new_scrolltop;  // the new scrolltop
    int curr_scrolltop, curr_scrollbottom;
    int maxindex = SONGPATLEN-1;
    int maxscrolltop;

    ERROR_NULL(pbar,0);

    pbar->cursor = newpos;
    scrsize = SLtt_Screen_Rows - pbar->row - HEADERSIZE - BOTTOMSIZE;
    curr_scrolltop = pbar->scrolltop;   // top-most visible index
    curr_scrollbottom = curr_scrolltop + scrsize - 1; // bottom most index
    maxscrolltop = maxindex-scrsize+1;

    if (scrsize < SONGPATLEN)
    {
        if (newpos < 0)
        {
            return 0;
        }
        else if (newpos > maxindex)
        {
            return maxscrolltop;
        }
        else
        {
            if (newpos > curr_scrollbottom)
            {
                new_scrolltop = newpos - scrsize + 1;
                if (new_scrolltop < 0)
                    new_scrolltop = 0;
                if (new_scrolltop > maxscrolltop)
                    new_scrolltop =  maxscrolltop;
            }
            else if (newpos < curr_scrolltop)
            {
                new_scrolltop = newpos;
                if (new_scrolltop < 0)
                    new_scrolltop = 0;
                if (new_scrolltop > maxscrolltop)
                    new_scrolltop =  maxscrolltop;
            }
            else
                new_scrolltop = curr_scrolltop;

            return new_scrolltop;
		}
    }
    else
    {
        return 0;
    }
}
//  function -- pbar_setCursor
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_editCursorVolume
//
//  Description
//
//      Modifies the MIDI velocity of the note under the cursor.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      mod
//
//          Amount by which to modify the velocity.
//
//  Return
//
//      Void.
//
//  -------------------------------------------------------------------------
void pbar_editCursorVolume(struct PatBar *pbar,  int mod)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E114: pbar_editCursorVolume: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
        spat_setVolume(pat, pbar->cursor,
                       spat_getVolume(pat, pbar->cursor) + mod);
    pbar_refresh(pbar);
}
//  function -- pbar_editCursorVolume
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_setCursorVolume
//
//  Description
//
//      Sets the MIDI velocity of the note under the cursor.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      new_velocity
//
//          Amount by which to modify the velocity.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_setCursorVolume(struct PatBar *pbar,  int new_velocity)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E114: pbar_setCursorVolume: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
        spat_setVolume(pat, pbar->cursor, new_velocity);
    pbar_refresh(pbar);
}
//  function -- pbar_setCursorVolume
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_toggleCursorSlide
//
//  Description
//
//      Toggles on or off the "slide" mode for the note under the cursor.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_toggleCursorSlide(struct PatBar *pbar)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E115: pbar_toggleCursorSlide: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
        spat_toggleSlide(pat, pbar->cursor);
}
//  function -- pbar_toggleCursorSlide
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  pbar_editPattern
//
//  Description
//
//      Changes the song pattern to a different pattern number.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      mod
//
//          Amount by which to modify the pattern number.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_editPattern(struct PatBar *pbar, int mod)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E116: pbar_editPattern: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
    {
        int val = spat_getIndex(pat);
		schan_setSongPattern(schan, pbar->pos, val + mod);
    }
    else if (mod == 1)
    {
        schan_setSongPattern(schan, pbar->pos, 1);
    }
	pbar_refresh(pbar);
}
//  function -- pbar_editPattern
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_patternSet
//
//  Description
//
//      Queries whether the pattern has been set.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//  Return
//
//      Boolean value.
//
//  -------------------------------------------------------------------------
int pbar_patternSet(struct PatBar *pbar)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL(pbar,M_FALSE);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E117: pbar_editPattern: NULL SongChannel\n");
        return M_FALSE;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
        return M_TRUE;
    else
       return M_FALSE;
}
//  function -- pbar_patternSet
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_transposeNotes
//
//  Description
//
//      Makes subsequent MIDI note entries occur at a higher or lower
//      octave.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      trans
//
//          Amount by which to transpose the notes.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_transposeNotes(struct PatBar *pbar, int trans)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E118: pbar_transposeNotes: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
    {
        spat_transposeNotes(pat, trans);
        pbar_refresh(pbar);
    }
}
//  function -- pbar_transposeNotes
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  pbar_setPatternName
//
//  Description
//
//      Sets the name of the pattern.
//
//  Arguments
//
//      pbar
//
//          The pointer to the PatBar struct instance.
//
//      newname
//
//          string pointer to the new name
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void pbar_setPatternName(struct PatBar *pbar, const char *newname)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E119: pbar_setPatternName: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (!pat)
        return;
    spat_setName(pat, newname);
	pbar_refresh(pbar);
}
//  function -- pbar_setPatternName
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_toggleEndNote
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_toggleEndNote(struct PatBar *pbar)
{
    struct SongChannel *schan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E120: pbar_toggleEndNote: NULL SongChannel\n");
        return;
    }


    pat = schan_getSongPattern(schan, pbar->pos);
    if (pat)
    {
        spat_setEndNote(pat, pbar->cursor);
    }
}
//  function -- pbar_toggleEndNote
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_toggleChannelMute
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_toggleChannelMute(struct PatBar *pbar)
{
    struct SongChannel *schan;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E121: pbar_toggleChannelMute: NULL SongChannel\n");
        return;
    }

    schan_toggleMute(schan);
    pbar_refresh(pbar);
}
//  function -- pbar_toggleChannelMute
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_editMidiChannel
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_editMidiChannel(struct PatBar *pbar, int mod)
{
    struct SongChannel *schan;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E122: pbar_editMidiChannel: NULL SongChannel\n");
        return;
    }

    schan_setMidiChannel(schan, schan_getMidiChannel(schan)+mod);
    pbar_refresh(pbar);
}
//  function -- pbar_editMidiChannel
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_editMidiPort
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_editMidiPort(struct PatBar *pbar, int port)
{
    struct SongChannel *schan;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E123: pbar_editMidiPort: NULL SongChannel\n");
        return;
    }

    schan_setMidiPort(schan, port);
    pbar_refresh(pbar);
}
//  function -- pbar_editMidiPort
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_editChannelVolume
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_editChannelVolume(struct PatBar *pbar, int mod)
{
    struct SongChannel *schan;

    ERROR_NULL_VOID(pbar);

    //
    //  Get the current "song channel" struct pointer
    //
    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E124: pbar_editChannelVolume: NULL SongChannel\n");
        return;
    }

    schan_setVolume(schan, schan_getVolume(schan)+mod);
    pbar_refresh(pbar);
}
//  function -- pbar_editChannelVolume
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_getIndex
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int pbar_getIndex(struct PatBar *pbar)
{
    ERROR_NULL(pbar,0);
    return pbar->chan;
}
//  function -- pbar_getIndex
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_setChannel
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_setChannel(struct PatBar *pbar, int chan)
{
    ERROR_NULL_VOID(pbar);

    pbar->chan = chan;
    pbar_refresh(pbar);
}
//  function -- pbar_setChannel
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_setScrollTop
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_setScrollTop(struct PatBar *pbar, int newpos)
{
    ERROR_NULL_VOID(pbar);

    pbar->scrolltop = newpos;
    pbar_refresh(pbar);
}
//  function -- pbar_setScrollTop
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_setChannelName
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_setChannelName(struct PatBar *pbar, const char *newname)
{
    struct SongChannel *schan;

    ERROR_NULL_VOID(pbar);
    ERROR_NULL_VOID(newname);

    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E125: pbar_setChannelName: NULL SongChannel\n");
        return;
    }

    schan_setName(schan, newname);
    pbar_refresh(pbar);
}
//  function -- pbar_setChannelName
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  pbar_setChannelPatch
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void pbar_setChannelPatch(struct PatBar *pbar, int bank, int program)
{
    struct SongChannel *schan;

    ERROR_NULL_VOID(pbar);

    schan = song_getChannel(p_Song, pbar->chan);
    if (schan == NULL)
    {
        Error_Printf("E126: pbar_setChannelPatch: NULL SongChannel\n");
        return;
    }

    schan_setBankNumber(schan, bank);
    schan_setProgramNumber(schan, program);
}
//  function -- pbar_setChannelPatch
//  -------------------------------------------------------------------------

