/**
 * C code modifications Copyright (C) 2009 Alfred Steffens Jr.
 * based on TekTracker, Copyright (C) 2001 Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/poll.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <assert.h>
#ifdef ALSA_DRIVER
#include <alsa/asoundlib.h>
#endif
#include <defs.h>
#include <logprintf.h>
#include <errorprintf.h>
#include <mididev.h>
#include <rawmidi.h>


//  -------------------------------------------------------------------------
//  Function    --  Create_RawMidi
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct RawMidi *Create_RawMidi(const char *hwdev, int mode, int alsa)
{
    #ifdef ALSA_DRIVER
    snd_rawmidi_t *handle_out = NULL;
    snd_rawmidi_t *handle_in = NULL;
    #endif
    struct RawMidi *rmid = NULL;
	mode_t openmode = 0;
    int midi_fd = 0;

    if (alsa)
    {
        #ifdef ALSA_DRIVER

        int err;

        //
        //  Open the ALSA output port
        //
        err = snd_rawmidi_open(NULL, &handle_out, hwdev, 0);
        if (err) {
            Error_Printf("E151: Failed to open (ALSA write) %s: %s\n",
                         hwdev, err);
            return NULL;
        }

        err = snd_rawmidi_open(&handle_in, NULL, hwdev, SND_RAWMIDI_NONBLOCK);
        if (err) {
            Error_Printf("E151: Failed to open (ALSA read) %s: %s\n",
                         hwdev, err);
            return NULL;
        }

        //
        //  Set to non-block for the MIDI input
        //
        snd_rawmidi_nonblock(handle_in, 0);

        LogFile_Printf("I000: opening ALSA driver (%s)\n", hwdev);

        #else

        Error_Printf("E150: ALSA not compiled into program\n");
        return NULL;

        #endif // ALSA_DRIVER
    }
    else
    {

        // Le Kernel-Bug Workaround.
        switch( mode ) {

            case MDEV_READMODE:
                openmode = O_RDONLY | O_NONBLOCK;
            break;

            case MDEV_WRITEMODE:
                openmode = O_WRONLY | O_NONBLOCK;
            break;

            default:
            case MDEV_READWRITEMODE:
                openmode = O_RDWR | O_NONBLOCK;
            break;
        }
        if (( midi_fd = open( hwdev, openmode )) == -1)
        {
            Error_Printf("E151: Failed to open %s: %s (hint: you must be root).\n",
                    hwdev, strerror(errno));
            return NULL;
        }
        close(midi_fd);

        // We now know we can successfully open it in blocking mode.
        switch( mode ) {

            case MDEV_READMODE:
                openmode = O_RDONLY;
            break;

            case MDEV_WRITEMODE:
                openmode = O_WRONLY;
            break;

            default:
            case MDEV_READWRITEMODE:
                openmode = O_RDWR;
            break;
        }

        if( ( midi_fd = open( hwdev, openmode ) ) == -1 ) {
            Error_Printf("E152: Failed to re-open %s for blocking I/O: %s.\n",
                    hwdev, strerror( errno ) );
            return NULL;
        }

        LogFile_Printf("I000: opening OSS driver (%s)\n", hwdev);

    }  // else -- OSS driver

    rmid = (struct RawMidi *)malloc( sizeof(struct RawMidi) );
    if (rmid == NULL)
    {
        Error_Printf("E153: Create_RawMidi: mem alloc\n");
        return NULL;
    }

    #ifdef ALSA_DRIVER
    rmid->ahandle_o = handle_out;
    rmid->ahandle_i = handle_in;
    #endif

    rmid->midi_fd = midi_fd;
    rmid->buffpos = 0;
    rmid->buff[0] = 0x00;
    rmid->port = 0;
    rmid->alsa = alsa;
    rmid->next = NULL;

    return rmid;
}
//  function -- Create_RawMidi
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_RawMidi
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void Delete_RawMidi(struct RawMidi *rmid)
{
    struct RawMidi *pN, *pR;

    if (rmid == NULL)
        return;

    pR = rmid;
    while (pR)
    {
        pN = pR;
        pR = pR->next;
        #ifdef ALSA_DRIVER
        if (pN->ahandle_o)
        {
            snd_rawmidi_drain(pN->ahandle_o);
            snd_rawmidi_close(pN->ahandle_o);
        }
        if (pN->ahandle_i)
            snd_rawmidi_close(pN->ahandle_i);
        #endif
	    if( pN->midi_fd > 0 )
            close( pN->midi_fd );
        free(pN);
    }
}
//  funtion -- Delete_RawMidi
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Find_Last_RawMidi
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct RawMidi *Find_Last_RawMidi(struct RawMidi *mididev_head)
{
    struct RawMidi *curr_mididev;

    if (mididev_head == NULL)
        return NULL;

    curr_mididev = mididev_head;
    while (curr_mididev->next)
    {
        curr_mididev = curr_mididev->next;
    }
    return curr_mididev;
}
//  funtion -- Find_Last_RawMidi
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Find_Nth_RawMidi
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct RawMidi *Find_Nth_RawMidi(struct RawMidi *mididev_head, int number)
{
    struct RawMidi *curr_mididev;
    int count = 0;

    if (mididev_head == NULL)
        return NULL;

    curr_mididev = mididev_head;
    while (curr_mididev)
    {
        if (curr_mididev->port == number)
            return curr_mididev;
        curr_mididev = curr_mididev->next;
        count++;
    }
    return NULL;
}
//  funtion -- Find_Nth_RawMidi
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_noteOn
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_noteOn(struct RawMidi *mididev,  unsigned int chan,
                 unsigned int port, unsigned int note, unsigned int vel)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0x90 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = note;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = vel;
}
//  funtion -- mdev_noteOn
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_noteOff
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_noteOff(struct RawMidi *mididev,  unsigned int chan,
                  unsigned int port, unsigned int note, unsigned int vel)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0x80 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = note;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = vel;
}
//  funtion -- mdev_noteOff
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_polyphonicAftertouch
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_polyphonicAftertouch(struct RawMidi *mididev, unsigned int chan,
                               unsigned int port, unsigned int note, unsigned int amt)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xA0 | chan;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = note;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = amt;
}
//  funtion -- mdev_polyphonicAftertouch
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_programChange
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_programChange(struct RawMidi *mididev, unsigned int chan,
                        unsigned int port, unsigned int prog)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xC0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = prog;
}
//  funtion -- mdev_programChange
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_controlChange
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_controlChange(struct RawMidi *mididev, unsigned int chan,
                        unsigned int port, unsigned int num, unsigned int val)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xB0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = num;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = val;
}
//  funtion -- mdev_controlChange
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_channelAftertouch
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_channelAftertouch(struct RawMidi *mididev, unsigned int chan,
                            unsigned int port, unsigned int amt)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xD0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = amt;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0;
}
//  funtion -- mdev_channelAftertouch
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_pitchWheel
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_pitchWheel(struct RawMidi *mididev, unsigned int chan,
                     unsigned int port, unsigned int amt)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xE0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xF | amt;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = amt >> 8;
}
//  funtion -- mdev_pitchWheel
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_allSoundOff
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_allSoundOff(struct RawMidi *mididev, int chan, int port)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xB0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0x78;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0;
}
//  funtion -- mdev_allSoundOff
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_resetAllControllers
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_resetAllControllers(struct RawMidi *mididev, int chan, int port)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xB0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0x79;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0;
}
//  funtion -- mdev_resetAllControllers
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_setLocal
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_setLocal(struct RawMidi *mididev, int chan, int port, int on)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xB0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0x7A;

	if (on == M_TRUE)
    {
        curr_mididev->buff[ curr_mididev->buffpos++ ] = 127;
    }
    else
    {
        curr_mididev->buff[ curr_mididev->buffpos++ ] = 0;
    }
}
//  funtion -- mdev_setLocal
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_allNotesOff
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_allNotesOff(struct RawMidi *mididev, int chan, int port)
{
    struct RawMidi *curr_mididev;
    int channel;

    ERROR_NULL_VOID(mididev);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    channel = chan % 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
        return;

	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xB0 | channel;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0x7B;
	curr_mididev->buff[ curr_mididev->buffpos++ ] = 0;
}
//  funtion -- mdev_allNotesOff
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_allNotesOff_all
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_allNotesOff_all(struct RawMidi *mididev)
{
    struct RawMidi *curr_mididev;
    int i;

    ERROR_NULL_VOID(mididev);

    curr_mididev = mididev;
    while (curr_mididev)
    {
        for (i=0; i < 16; i++)
        {
            curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xB0 | i;
            curr_mididev->buff[ curr_mididev->buffpos++ ] = 0x7B;
            curr_mididev->buff[ curr_mididev->buffpos++ ] = 0;
        }

        curr_mididev = curr_mididev->next;
    }
}
//  funtion -- mdev_allNotesOff_all
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_shutUp
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_shutUp(struct RawMidi *mididev)
{
    struct RawMidi *curr_mididev;
    int i;

    ERROR_NULL_VOID(mididev);

    curr_mididev = mididev;
    while (curr_mididev)
    {
        for (i = 0; i < 16; i++)
            mdev_allNotesOff(mididev, i, curr_mididev->port);
        curr_mididev = curr_mididev->next;
    }
}
//  funtion -- mdev_shutUp
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  mdev_syncStart
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_syncStart(struct RawMidi *mididev)
{
    struct RawMidi *curr_mididev;

    ERROR_NULL_VOID(mididev);

    curr_mididev = mididev;
    while (curr_mididev)
    {
	    curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xFA;
        curr_mididev = curr_mididev->next;
    }
}
//  funtion -- mdev_syncStart
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_syncStop
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_syncStop(struct RawMidi *mididev)
{
    struct RawMidi *curr_mididev;

    ERROR_NULL_VOID(mididev);

    curr_mididev = mididev;
    while (curr_mididev)
    {
	    curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xFC;
        curr_mididev = curr_mididev->next;
    }
}
//  funtion -- mdev_syncStop
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_syncContinue
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_syncContinue(struct RawMidi *mididev)
{
    struct RawMidi *curr_mididev;

    ERROR_NULL_VOID(mididev);

    curr_mididev = mididev;
    while (curr_mididev)
    {
	    curr_mididev->buff[ curr_mididev->buffpos++ ] = 0xFB;
        curr_mididev = curr_mididev->next;
    }
}
//  funtion -- mdev_syncContinue
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_syncTick
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_syncTick(struct RawMidi *mididev)
{
    struct RawMidi *curr_mididev;

    ERROR_NULL_VOID(mididev);

    curr_mididev = mididev;
    while (curr_mididev)
    {
        curr_mididev->buff[ curr_mididev->buffpos ] = 0xF8;
        curr_mididev->buffpos++;
        if (curr_mididev->buffpos >= MIDI_BUFFSIZE)
            curr_mididev->buffpos = 0;
        curr_mididev = curr_mididev->next;
    }
}
//  funtion -- mdev_syncTick
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_readMessage
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int mdev_readMessage(struct RawMidi *mididev, int chan)
{
	unsigned char buff;
	int bytesread;
	struct pollfd pfd;
	int ret;
    struct RawMidi *curr_mididev;
    int port;

    ERROR_NULL(mididev,0);

    //
    //  "chan" is the midi channel (counting from 0)
    //
    port = chan / 16;
    curr_mididev = Find_Nth_RawMidi(mididev, port);
    if (curr_mididev == NULL)
    {
        Error_Printf("E158: mdev_readMessage: no MIDI device found\n");
        return 0;
    }

    if (curr_mididev->alsa)
    {
        #ifdef ALSA_DRIVER
        //
        //  TBD -- Alsa poll the MIDI device
        //

	    buff = 0;
	    bytesread = snd_rawmidi_read(curr_mididev->ahandle_i, &buff, 1);

        #endif
    }
    else
    {
        //
        //  OSS Read
        //
        pfd.fd = curr_mididev->midi_fd;
        pfd.events = POLLIN | POLLERR;

    again:
        ret = poll( &pfd, 1, 1000 );
        if( ret < 0 ) {
            if (errno == EINTR) {
                // This happens mostly when run under gdb, or when
                // exiting due to a signal.
                goto again;
            }
            return MDEV_NOMESSAGE;
        }

        if( ret == 0 ) return MDEV_NOMESSAGE;


	    buff = 0;
	    bytesread = read(curr_mididev->midi_fd, &buff, 1);

    } // else -- OSS Read

	if( bytesread < 1 ) {
		return MDEV_NOMESSAGE;
	}

	if( buff == 0xFA ) {
		return MDEV_MIDISTART;
	}

	if( buff == 0xFB ) {
		return MDEV_MIDICONTINUE;
	}

	if( buff == 0xFC ) {
		return MDEV_MIDISTOP;
	}

	if( buff == 0xF8 ) {
		return MDEV_MIDISTOP;
	}

	return MDEV_NOMESSAGE;
}
//  funtion -- mdev_readMessage
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_flush
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void mdev_flush(struct RawMidi *mididev)
{
    struct RawMidi *curr_mdev;

    ERROR_NULL_VOID(mididev);

    curr_mdev = mididev;
    while (curr_mdev)
    {
        if (curr_mdev->alsa)
        {
            #ifdef ALSA_DRIVER
            int err;

            err = snd_rawmidi_write(curr_mdev->ahandle_o, curr_mdev->buff,
                                  curr_mdev->buffpos);
            #else
            curr_mdev = curr_mdev->next;
            continue;
            #endif
        }
        else
        {
	        write(curr_mdev->midi_fd, curr_mdev->buff, curr_mdev->buffpos);
        }

	    curr_mdev->buffpos = 0;
        curr_mdev = curr_mdev->next;
    }
}
//  funtion -- mdev_flush
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  mdev_isOpen
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int mdev_isOpen(struct RawMidi *rmid)
{
    ERROR_NULL(rmid,0);

#ifdef ALSA_DRIVER
    return (rmid->ahandle_o != NULL);
#else
    return (rmid->midi_fd > 0);
#endif
}
//  funtion -- mdev_isOpen
//  -------------------------------------------------------------------------

