/**
 * Copyright (C) 2009 Alfred Steffens Jr.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif // HAVE_CONFIG_H
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdarg.h>
#include <defs.h>
#include <logprintf.h>
#include <errorprintf.h>
#include <record.h>

//
//  If REVENT events were read from a file, these pointers
//  will be used for playback
//
struct RecordEvent *p_REvents = NULL, *p_CurrREvents = NULL;
//
//  Pointers for recording
//
struct RecordEvent *p_Rec = NULL, *p_CurrRec = NULL;

//
//  Whitespace separators
//
static char whiteseps[5] = {0x09, 0x0A, 0x0D, 0x20, 0x00};

char *Rv_Names[] = {
    "none",                                 // 0
    "quit",                                 // 1
    "channel_mute_1",                       // 2    F1 key
    "channel_mute_2",                       // 3    F2 key
    "channel_mute_3",                       // 4    F3 key
    "channel_mute_4",                       // 5    F4 key
    "channel_mute_5",                       // 6    F5 key
    "channel_mute_6",                       // 7    F6 key
    "channel_mute_7",                       // 8    F7 key
    "channel_mute_8",                       // 9    F8 key
    "song_start",                           // 10   F9 key
    "song_rewind_start",                    // 11   F10 key
    "song_stop",                            // 12   F11 key
    "song_rewind",                          // 13   F12 key
    "key_down",                             // 14   down
    "key_up",                               // 15   up
    "key_left",                             // 16   left
    "key_right",                            // 17   right
    "shift_key_down",                       // 18   shift-down
    "shift_key_up",                         // 19   shift-up
    "shift_key_left",                       // 20   shift-left
    "shift_key_right",                      // 21   shift-right
    "midichannel_up",                       // 22   Ins key
    "midichannel_down",                     // 23   Del key
    "volume_up",                            // 24   '*' key
    "volume_down",                          // 25   '/' key
    "velocity_up",                          // 26   Home key
    "velocity_down",                        // 27   End key
    "pattern_up",                           // 28   ']' key
    "pattern_down",                         // 29   '[' key
    "octave_up",                            // 30   '+' key
    "octave_down",                          // 31   '-' key
    "semitone_up",                          // 32   (apostraphe) key
    "semitone_down",                        // 33   (backslash) key
    "songposition_up",                      // 34   PgUp key
    "songposition_down",                    // 35   PgDn key
    "toggle_slide",                         // 36   ` key
    "note_empty",                           // 37   <spacebar> key
    "note_c__lo",                           // 38   Z key
    "note_cs_lo",                           // 39   S key
    "note_d__lo",                           // 40   X key
    "note_ds_lo",                           // 41   D key
    "note_e__lo",                           // 42   C key
    "note_f__lo",                           // 43   V key
    "note_fs_lo",                           // 44   G key
    "note_g__lo",                           // 45   B key
    "note_gs_lo",                           // 46   H key
    "note_a__lo",                           // 47   N key
    "note_as_lo",                           // 48   J key
    "note_b__lo",                           // 49   M key
    "note_c__hi",                           // 50   Q key
    "note_cs_hi",                           // 51   2 key
    "note_d__hi",                           // 52   W key
    "note_ds_hi",                           // 53   3 key
    "note_e__hi",                           // 54   E key
    "note_f__hi",                           // 55   R key
    "note_fs_hi",                           // 56   5 key
    "note_g__hi",                           // 57   T key
    "note_gs_hi",                           // 58   6 key
    "note_a__hi",                           // 59   Y key
    "note_as_hi",                           // 60   7 key
    "note_b__hi",                           // 61   U key
    "note_c__hx",                           // 62   I key
    "note_cs_hx",                           // 63   9 key
    "note_d__hx",                           // 64   O key
    "note_ds_hx",                           // 65   0 key
    "note_e__hx",                           // 66   P key
    "tempo",                                // 67   ctrl-T
    "loop_on",                              // 68   ! key
    "loop_position",                        // 69   ctrl-L
    "loop_destination",                     // 70   ctrl-K
    "start_position",                       // 71   ctrl-F
    "save_song",                            // 72   ctrl-S
    "load_song",                            // 73   ctrl-A
    "toggle_sync",                          // 74   ctrl-E
    "save_copy",                            // 75   ctrl-C
    "paste_copy",                           // 76   ctrl-V
    "channel_name",                         // 77   ctrl-N
    "pattern_name",                         // 78   ctrl-P
    "note_off",                             // 79   backspace
    "toggle_endnote",                       // 80   ctrl-Z
    "midi_port",                            // 81   ctrl-B
    "midi_patch",                           // 82   ctrl-H
    "show_info",                            // 83   ctrl-G
    "switch_mode",                          // 84   tab
    "ext_loop_position",                    // 85   command, set loop pos
    "ext_loop_destination",                 // 86   command, set loop dest
    "ext_midi_flush",                       // 87   command, midi flush
    "ext_midi_control",                     // 88   command, midi control msg
    "ext_velocity_reset",                   // 89   command, velocity reset
    "ext_velocity_change",                  // 90   command, velocity change
    "ext_pattern_set"                       // 91   command, pattern set
};

char *convert_tolower(char *str);



//  -------------------------------------------------------------------------
//  Function    --  Create_RecordEvent
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct RecordEvent *Create_RecordEvent(unsigned long index, int type)
{
    int j;
    struct RecordEvent *revent;

    revent = (struct RecordEvent *)malloc( sizeof(struct RecordEvent) );
    if (revent == NULL)
    {
        Error_Printf("E400: Create_RecordEvent: mem alloc\n");
        return NULL;
    }
    revent->type = type;
    revent->has_iParams = 0;
    for (j=0; j < REVENT_MAX_IPARMS; j++)
        revent->iParams[j] = 0;
    revent->index = index;
    revent->next = NULL;
    return revent;
}
//  function -- Create_RecordEvent
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_RecordEvent
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void Delete_RecordEvent(struct RecordEvent *revent)
{
    struct RecordEvent *pN, *pR;

    if (revent == NULL)
        return;

    pR = revent;
    while (pR)
    {
        pN = pR;
        pR = pR->next;
        free(pN);
    }
}
//  function -- Delete_RecordEvent
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Revent_Array_From_Chain
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct RecordEvent *Revent_Array_From_Chain(struct RecordEvent *revent_head)
{
    struct RecordEvent *revent_array=NULL, *curr_revent;
    unsigned long count=0;

    if (revent_head == NULL)
        return NULL;

    //
    //  count the events
    //
    curr_revent = revent_head;
    while (curr_revent)
    {
        count++;
        curr_revent = curr_revent->next;
    }

    if (count > 0)
    {
        unsigned long j;

        revent_array = (struct RecordEvent *)malloc(
                        sizeof(struct RecordEvent) * count );
        if (revent_array == NULL)
        {
            Error_Printf("E401: Revent_Array_From_Chain: mem alloc\n");
            return NULL;
        }
        //
        //  Copy the chain into the array
        //
        curr_revent = revent_head;
        for (j=0; j < count; j++)
        {
            int k;

            revent_array[j].type = curr_revent->type;
            revent_array[j].index = curr_revent->index;
            revent_array[j].next = NULL;
            for (k=0; k<REVENT_MAX_IPARMS; k++)
                revent_array[j].iParams[k] = curr_revent->iParams[k];
            curr_revent = curr_revent->next;
        }
        return revent_array;    // success
    }
    else
        return NULL;
}
//  function -- Revent_Array_From_Chain
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Attach_Revent
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct RecordEvent *Attach_Revent(struct RecordEvent *curr_re,
                                  unsigned long beatcount, int type,
                                  int *iParams)
{
    struct RecordEvent *next_re = NULL;

    //
    //  Care must be taken that "curr_re" is the current pointer, not the
    //  head of the chain.  We DO NOT want to search for the last link.
    //
    next_re = Create_RecordEvent(beatcount, type);
    if (next_re == NULL)
        return NULL;

    if (iParams)
    {
        int i;
        for (i=0; i<REVENT_MAX_IPARMS; i++)
            next_re->iParams[i] = iParams[i];
    }

    //
    //  Attach it to the chain
    //
    if (curr_re == NULL)
    {
        //
        //  The chain is not created yet
        //  next_re will become the head of the chain
        //

        //
        //  This is global !
        //
        if (p_Rec == NULL)
            p_Rec = next_re;
        else
        {
            p_Rec = next_re;
            Error_Printf("E402: Re-creating head of record chain\n");
        }
    }
    else
    {
        //
        //  curr_re is the "current" pointer
        //
        curr_re->next = next_re;
    }

    //
    //  Return the new REVENT pointer, the calling function must use it
    //  as the new current pointer.
    //
    return next_re;
}
//  function -- Attach_Revent
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  Read_Events_File
//
//  Description
//
//      Reads an "events" file.  This is a session file or an edited session
//      file containing extended event commands.
//
//  Arguments
//
//      filename
//
//          The name of the events file.
//
//  Return
//
//      Returns a linked list of RecordEvent struct instances.  Returns NULL
//      on error.
//
//  -------------------------------------------------------------------------
struct RecordEvent *Read_Events_File(char *filename)
{
    FILE *file=NULL;
    struct RecordEvent *revent=NULL, *curr_revent=NULL;
    unsigned long count=0;
    static int iParams[REVENT_MAX_IPARMS];

    if (filename == NULL)
    {
        Error_Printf("E402: Read_Events_File: NULL filename\n");
        return NULL;
    }

    //
    //  Open the Events file
    //
    file = fopen(filename, "r");
    if (file == NULL)
    {
        LogFile_Printf("E403: Read_Events_File: cannot open events file (%s)\n",
                      filename);
        return NULL;
    }

    //
    //  scan the file
    //
    while (feof(file) == 0)
    {
        static char Line[128], RvStr[REVENT_MAX_NAME_LENGTH];
        char *pch, *ptok;
        unsigned long index;
        int i, type, valid_event, has_iParams;

        index = 0;
        type = 0;
        RvStr[0] = 0x00;
        valid_event = 0;
        has_iParams = 0;
        for (i=0; i<REVENT_MAX_IPARMS; i++)
            iParams[i] = 0;

        //
        //  Get the next line from the file
        //
        pch = fgets(Line, 127, file);
        if (pch==NULL)
            break;

        //
        //  Get the first token -- usually the Beat Index
        //
        ptok = strtok(Line, whiteseps);

        //
        //  Skip blank lines
        //
        if (ptok && (strlen(ptok) == 0))
        {
            //
            //  A blank line
            //  (I'm not sure if this condition ever occurs)
            //
            continue;
        }

        //
        //  Skip comment lines
        //
        if (ptok && (strchr(ptok, '#') == ptok))
        {
            //
            //  A comment line
            //
            continue;
        }

        //
        //  Assume it is a valid token for Field 1
        //
        if (ptok)
        {
            index = (unsigned long)atol(ptok);

            //
            //  Get the Revent identifier string -- (Field 2)
            //
            ptok = strtok(NULL, whiteseps);
            if (ptok)
            {
                char *plower, *py;

                //
                //  Make case-insensitive
                //
                plower = convert_tolower(ptok);
                strncpy(RvStr, plower, (REVENT_MAX_NAME_LENGTH-1));
                RvStr[(REVENT_MAX_NAME_LENGTH-1)] = 0x00;
                py = strchr(RvStr, 13);
                if (py)
                    *py = 0x00;
                py = strchr(RvStr, 10);
                if (py)
                    *py = 0x00;

                //
                //  Switch to match the string to a TYPE
                //
                if (strcmp(RvStr, Rv_Names[REVENT_QUIT]) == 0)
                {
                    //
                    //  QUIT
                    //
                    type = REVENT_QUIT;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE1]) == 0)
                {
                    //
                    //  F1 key
                    //
                    type = REVENT_CHMUTE1;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE2]) == 0)
                {
                    //
                    //  F2 key
                    //
                    type = REVENT_CHMUTE2;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE3]) == 0)
                {
                    //
                    //  F3 key
                    //
                    type = REVENT_CHMUTE3;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE4]) == 0)
                {
                    //
                    //  F4 key
                    //
                    type = REVENT_CHMUTE4;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE5]) == 0)
                {
                    //
                    //  F5 key
                    //
                    type = REVENT_CHMUTE5;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE6]) == 0)
                {
                    //
                    //  F6 key
                    //
                    type = REVENT_CHMUTE6;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE7]) == 0)
                {
                    //
                    //  F7 key
                    //
                    type = REVENT_CHMUTE7;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHMUTE8]) == 0)
                {
                    //
                    //  F8 key
                    //
                    type = REVENT_CHMUTE8;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SONG_START]) == 0)
                {
                    //
                    //  F9 key
                    //
                    type = REVENT_SONG_START;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SONG_RWSTART]) == 0)
                {
                    //
                    //  F10 key
                    //
                    type = REVENT_SONG_RWSTART;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SONG_STOP]) == 0)
                {
                    //
                    //  F11 key
                    //
                    type = REVENT_SONG_STOP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SONG_REWIND]) == 0)
                {
                    //
                    //  F12 key
                    //
                    type = REVENT_SONG_REWIND;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_KEY_DOWN]) == 0)
                {
                    //
                    //  Down key
                    //
                    type = REVENT_KEY_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_KEY_UP]) == 0)
                {
                    //
                    //  Up  key
                    //
                    type = REVENT_KEY_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_KEY_LEFT]) == 0)
                {
                    //
                    //  Left key
                    //
                    type = REVENT_KEY_LEFT;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_KEY_RIGHT]) == 0)
                {
                    //
                    //  Right key
                    //
                    type = REVENT_KEY_RIGHT;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SHIFT_KEY_DOWN]) == 0)
                {
                    //
                    //  Shift Down key
                    //
                    type = REVENT_SHIFT_KEY_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SHIFT_KEY_UP]) == 0)
                {
                    //
                    //  Shift Up key
                    //
                    type = REVENT_SHIFT_KEY_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SHIFT_KEY_LEFT]) == 0)
                {
                    //
                    //  Shift left key
                    //
                    type = REVENT_SHIFT_KEY_LEFT;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SHIFT_KEY_RIGHT]) == 0)
                {
                    //
                    //  Shift right key
                    //
                    type = REVENT_SHIFT_KEY_LEFT;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_MIDICHANNEL_UP]) == 0)
                {
                    //
                    //  Midi Channel up (Ins key)
                    //
                    type = REVENT_MIDICHANNEL_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_MIDICHANNEL_DOWN]) == 0)
                {
                    //
                    //  Midi Channel down (Del key)
                    //
                    type = REVENT_MIDICHANNEL_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_VOLUME_UP]) == 0)
                {
                    //
                    //  Volume up ('*' key)
                    //
                    type = REVENT_VOLUME_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_VOLUME_DOWN]) == 0)
                {
                    //
                    //  Volume down ('/' key)
                    //
                    type = REVENT_VOLUME_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_VELOCITY_UP]) == 0)
                {
                    //
                    //  Velocity up (Home key)
                    //
                    type = REVENT_VELOCITY_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_VELOCITY_DOWN]) == 0)
                {
                    //
                    //  Velocity down (End key)
                    //
                    type = REVENT_VELOCITY_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_PATTERN_UP]) == 0)
                {
                    //
                    //  Pattern up (']' key)
                    //
                    type = REVENT_PATTERN_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_PATTERN_DOWN]) == 0)
                {
                    //
                    //  Pattern down ('[' key)
                    //
                    type = REVENT_PATTERN_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_OCTAVE_UP]) == 0)
                {
                    //
                    //  Octave up ('+' key)
                    //
                    type = REVENT_OCTAVE_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_OCTAVE_DOWN]) == 0)
                {
                    //
                    //  Octave down ('-' key)
                    //
                    type = REVENT_OCTAVE_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SEMITONE_UP]) == 0)
                {
                    //
                    //  semitone up ( apostraphe key)
                    //
                    type = REVENT_SEMITONE_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SEMITONE_DOWN]) == 0)
                {
                    //
                    //  semitone down ( backslash key)
                    //
                    type = REVENT_SEMITONE_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SONGPOSITION_UP]) == 0)
                {
                    //
                    //  song position up (PgUp key)
                    //
                    type = REVENT_SONGPOSITION_UP;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SONGPOSITION_DOWN]) == 0)
                {
                    //
                    //  song position down (PgDn key)
                    //
                    type = REVENT_SONGPOSITION_DOWN;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_TOGGLE_SLIDE]) == 0)
                {
                    //
                    //  toggle slide (` key)
                    //
                    type = REVENT_TOGGLE_SLIDE;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_EMPTY]) == 0)
                {
                    //
                    //  note empty (spacebar)
                    //
                    type = REVENT_NOTE_EMPTY;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_C__LO]) == 0)
                {
                    //
                    //  note C lower (Z key)
                    //
                    type = REVENT_NOTE_C__LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_CS_LO]) == 0)
                {
                    //
                    //  note C# lower (S key)
                    //
                    type = REVENT_NOTE_CS_LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_D__LO]) == 0)
                {
                    //
                    //  note D lower (X key)
                    //
                    type = REVENT_NOTE_D__LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_DS_LO]) == 0)
                {
                    //
                    //  note D# lower (D key)
                    //
                    type = REVENT_NOTE_DS_LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_E__LO]) == 0)
                {
                    //
                    //  note D lower (C key)
                    //
                    type = REVENT_NOTE_E__LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_F__LO]) == 0)
                {
                    //
                    //  note F lower (V key)
                    //
                    type = REVENT_NOTE_F__LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_FS_LO]) == 0)
                {
                    //
                    //  note F# lower (G key)
                    //
                    type = REVENT_NOTE_FS_LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_G__LO]) == 0)
                {
                    //
                    //  note G lower (B key)
                    //
                    type = REVENT_NOTE_G__LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_GS_LO]) == 0)
                {
                    //
                    //  note G# lower (H key)
                    //
                    type = REVENT_NOTE_GS_LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_A__LO]) == 0)
                {
                    //
                    //  note A lower (N key)
                    //
                    type = REVENT_NOTE_A__LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_AS_LO]) == 0)
                {
                    //
                    //  note A# lower (J key)
                    //
                    type = REVENT_NOTE_AS_LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_B__LO]) == 0)
                {
                    //
                    //  note B lower (M key)
                    //
                    type = REVENT_NOTE_B__LO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_C__HI]) == 0)
                {
                    //
                    //  note C higher (Q key)
                    //
                    type = REVENT_NOTE_C__HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_CS_HI]) == 0)
                {
                    //
                    //  note C# higher (2 key)
                    //
                    type = REVENT_NOTE_CS_HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_D__HI]) == 0)
                {
                    //
                    //  note D higher (W key)
                    //
                    type = REVENT_NOTE_D__HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_DS_HI]) == 0)
                {
                    //
                    //  note D# higher (3 key)
                    //
                    type = REVENT_NOTE_DS_HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_E__HI]) == 0)
                {
                    //
                    //  note D higher (E key)
                    //
                    type = REVENT_NOTE_E__HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_F__HI]) == 0)
                {
                    //
                    //  note F higher (R key)
                    //
                    type = REVENT_NOTE_F__HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_FS_HI]) == 0)
                {
                    //
                    //  note F# higher (5 key)
                    //
                    type = REVENT_NOTE_FS_HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_G__HI]) == 0)
                {
                    //
                    //  note G higher (T key)
                    //
                    type = REVENT_NOTE_G__HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_GS_HI]) == 0)
                {
                    //
                    //  note G# higher (6 key)
                    //
                    type = REVENT_NOTE_GS_HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_A__HI]) == 0)
                {
                    //
                    //  note A higher (Y key)
                    //
                    type = REVENT_NOTE_A__HI;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_AS_HI]) == 0)
                {
                    //
                    //  note A# higher (7 key)
                    //
                    type = REVENT_NOTE_AS_HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_B__HI]) == 0)
                {
                    //
                    //  note B higher (U key)
                    //
                    type = REVENT_NOTE_B__HI;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_C__HX]) == 0)
                {
                    //
                    //  note C highest (I key)
                    //
                    type = REVENT_NOTE_C__HX;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_CS_HX]) == 0)
                {
                    //
                    //  note C# highest (9 key)
                    //
                    type = REVENT_NOTE_CS_HX;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_D__HX]) == 0)
                {
                    //
                    //  note D highest (O key)
                    //
                    type = REVENT_NOTE_D__HX;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_DS_HX]) == 0)
                {
                    //
                    //  note D# highest (0 key)
                    //
                    type = REVENT_NOTE_DS_HX;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_E__HX]) == 0)
                {
                    //
                    //  note D highest (P key)
                    //
                    type = REVENT_NOTE_E__HX;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_TEMPO]) == 0)
                {
                    char *px;
                    //
                    //  Tempo (ctrl-T)
                    //
                    valid_event = 1;
                    type = REVENT_TEMPO;
                    px = strtok(NULL, whiteseps);
                    if (px)
                    {
                        has_iParams = 1;
                        ERROR_RANGE(0,REVENT_MAX_IPARMS,"(tempo from file)");
                        iParams[0] = atoi(px);
                    }
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_LOOP_ON]) == 0)
                {
                    //
                    //  Loop mode on ( ! key )
                    //
                    type = REVENT_LOOP_ON;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_LOOP_POSITION]) == 0)
                {
                    //
                    //  Loop position (ctrl-L)
                    //
                    type = REVENT_LOOP_POSITION;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_LOOP_DESTINATION]) == 0)
                {
                    //
                    //  Loop destination (ctrl-K)
                    //
                    type = REVENT_LOOP_DESTINATION;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_START_POSITION]) == 0)
                {
                    //
                    //  Start position (ctrl-F)
                    //
                    type = REVENT_START_POSITION;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SAVE_SONG]) == 0)
                {
                    //
                    //  Save song (ctrl-S)
                    //
                    type = REVENT_SAVE_SONG;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_LOAD_SONG]) == 0)
                {
                    //
                    //  Load song (ctrl-A)
                    //
                    type = REVENT_LOAD_SONG;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_TOGGLE_SYNC]) == 0)
                {
                    //
                    //  Toggle sync (ctrl-E)
                    //
                    type = REVENT_TOGGLE_SYNC;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SAVE_COPY]) == 0)
                {
                    //
                    //  Save copy (ctrl-C)
                    //
                    type = REVENT_SAVE_COPY;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_CHANNEL_NAME]) == 0)
                {
                    //
                    //  Channel name (ctrl-N)
                    //
                    type = REVENT_CHANNEL_NAME;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_PATTERN_NAME]) == 0)
                {
                    //
                    //  Pattern name (ctrl-P)
                    //
                    type = REVENT_PATTERN_NAME;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_NOTE_OFF]) == 0)
                {
                    //
                    //  Note off (backspace)
                    //
                    type = REVENT_NOTE_OFF;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_TOGGLE_ENDNOTE]) == 0)
                {
                    //
                    //  Toggle end note (ctrl-Z)
                    //
                    type = REVENT_TOGGLE_ENDNOTE;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_MIDI_PORT]) == 0)
                {
                    char *p1;
                    //
                    //  Midi port (ctrl-B)
                    //
                    valid_event = 1;
                    type = REVENT_MIDI_PORT;
                    p1 = strtok(NULL, whiteseps);
                    if (p1)
                    {
                        has_iParams = 1;
                        ERROR_RANGE(0,REVENT_MAX_IPARMS,"(midi port from file)");
                        iParams[0] = atoi(p1);
                    }
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_MIDI_PATCH]) == 0)
                {
                    char *p1, *p2;
                    //
                    //  Midi patch (ctrl-H)
                    //
                    valid_event = 1;
                    type = REVENT_MIDI_PATCH;
                    p1 = strtok(NULL, whiteseps);
                    if (p1)
                    {
                        has_iParams = 2;
                        ERROR_RANGE(0,REVENT_MAX_IPARMS,"(midi bank from file)");
                        iParams[0] = atoi(p1);
                        p2 = strtok(NULL, whiteseps);
                        if (p2)
                        {
                            ERROR_RANGE(1,REVENT_MAX_IPARMS,"(midi patch from file)");
                            iParams[1] = atoi(p2);
                        }
                    }
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SHOW_INFO]) == 0)
                {
                    //
                    //  Show Midi info (ctrl-G)
                    //
                    type = REVENT_SHOW_INFO;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_SWITCH_MODE]) == 0)
                {
                    //
                    //  switch mode (tab)
                    //
                    type = REVENT_SWITCH_MODE;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_EXT_LOOP_POS]) == 0)
                {
                    char *p1;
                    //
                    //  extended session command
                    //  set loop position
                    //
                    type = REVENT_EXT_LOOP_POS;
                    valid_event = 1;
                    p1 = strtok(NULL, whiteseps);
                    if (p1)
                    {
                        has_iParams = 1;
                        ERROR_RANGE(0,REVENT_MAX_IPARMS,"(loop pos from file)");
                        iParams[0] = atoi(p1);
                    }
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_EXT_LOOP_DEST]) == 0)
                {
                    char *p1;
                    //
                    //  extended session command
                    //  set loop destination
                    //
                    type = REVENT_EXT_LOOP_DEST;
                    valid_event = 1;
                    p1 = strtok(NULL, whiteseps);
                    if (p1)
                    {
                        has_iParams = 1;
                        ERROR_RANGE(0,REVENT_MAX_IPARMS,"(loop dest from file)");
                        iParams[0] = atoi(p1);
                    }
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_EXT_MIDI_FLUSH]) == 0)
                {
                    //
                    //  extended session command
                    //  flush the midi buffer
                    //
                    type = REVENT_EXT_MIDI_FLUSH;
                    valid_event = 1;
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_EXT_MIDI_CONTROL]) == 0)
                {
                    char *p1;
                    int k=0;
                    //
                    //  extended session command
                    //  send a midi control message
                    //
                    type = REVENT_EXT_MIDI_CONTROL;
                    valid_event = 1;
                    p1 = strtok(NULL, whiteseps);
                    while (p1)
                    {
                        //  ignore comments
                        if (strchr(p1, 0x23) == p1)
                            break;
                        has_iParams++;
                        iParams[k++] = atoi(p1);
                        p1 = strtok(NULL, whiteseps);
                    }
                    ERROR_RANGE(has_iParams,REVENT_MAX_IPARMS,
                                "(midi control from file)");
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_EXT_VELOCITY_RESET]) == 0)
                {
                    char *p1;
                    int k=0;
                    //
                    //  extended session command
                    //  reset the velocity
                    //
                    type = REVENT_EXT_VELOCITY_RESET;
                    valid_event = 1;
                    p1 = strtok(NULL, whiteseps);
                    while (p1)
                    {
                        //  ignore comments
                        if (strchr(p1, 0x23) == p1)
                            break;
                        has_iParams++;
                        iParams[k++] = atoi(p1);
                        p1 = strtok(NULL, whiteseps);
                    }
                    ERROR_RANGE(has_iParams,REVENT_MAX_IPARMS,
                                "(veloc reset from file)");
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_EXT_VELOCITY_CHANGE]) == 0)
                {
                    char *p1;
                    int k=0;
                    //
                    //  extended session command
                    //  set the velocity
                    //
                    type = REVENT_EXT_VELOCITY_CHANGE;
                    valid_event = 1;
                    p1 = strtok(NULL, whiteseps);
                    while (p1)
                    {
                        //  ignore comments
                        if (strchr(p1, 0x23) == p1)
                            break;
                        has_iParams++;
                        iParams[k++] = atoi(p1);
                        p1 = strtok(NULL, whiteseps);
                    }
                    ERROR_RANGE(has_iParams,REVENT_MAX_IPARMS,
                                "(veloc change from file)");
                }
                else if (strcmp(RvStr, Rv_Names[REVENT_EXT_PATTERN_SET]) == 0)
                {
                    char *p1;
                    int k=0;
                    //
                    //  extended session command
                    //  set a pattern number in song mode
                    //
                    type = REVENT_EXT_PATTERN_SET;
                    valid_event = 1;
                    p1 = strtok(NULL, whiteseps);
                    while (p1)
                    {
                        //  ignore comments
                        if (strchr(p1, 0x23) == p1)
                            break;
                        has_iParams++;
                        iParams[k++] = atoi(p1);
                        p1 = strtok(NULL, whiteseps);
                    }
                    ERROR_RANGE(has_iParams,REVENT_MAX_IPARMS,
                                "(veloc change from file)");
                }
                //
                //  End -- Identifier strings
                //
            }
            //  if -- second tok
        }
        //  if -- first tok


        //
        //  Add another event to the temporary chain
        //
        if (valid_event)
        {
            if (revent == NULL)
            {
                revent = curr_revent = Create_RecordEvent(index, type);
            }
            else
            {
                curr_revent->next = Create_RecordEvent(index, type);
                curr_revent = curr_revent->next;
            }
            if (has_iParams)
            {
                int j;

                curr_revent->has_iParams = has_iParams;
                for (j=0; j<REVENT_MAX_IPARMS; j++)
                    curr_revent->iParams[j] = iParams[j];
            }
            count++;
        }
    }
    //  while

    fclose(file);
    return revent;
}
//  function -- Read_Events_File
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Write_Events_File
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int Write_Events_File(char *filename, struct RecordEvent *revent)
{
    FILE *file=NULL;
    struct RecordEvent *curr_revent;

    if (revent == NULL)
    {
        LogFile_Printf("WARNING: Write_Events_File: empty (null)"
                       " REvent chain\n");
        return 0;
    }

    if (filename == NULL)
    {
        Error_Printf("E404: Write_Events_File: NULL filename\n");
        return -1;
    }

    //
    //  Create the Events file
    //
    file = fopen(filename, "w");
    if (file == NULL)
    {
        LogFile_Printf("E405: Write_Events_File: cannot create events"
                       " file (%s)\n", filename);
        return -1;
    }

    curr_revent = revent;
    while (curr_revent)
    {
        if (curr_revent->type > REVENT_MAX_RV_NAMES)
        {
            LogFile_Printf("E406: Write_Events_File: event type (%d) out-of"
                           "-range at song beat (%ld)\n",
                            curr_revent->type, curr_revent->index);
            return -1;
        }

        if (curr_revent->has_iParams)
        {
            int j;

            fprintf(file, "%ld %s", curr_revent->index,
                                     Rv_Names[curr_revent->type]);
            for (j=0; j < curr_revent->has_iParams; j++)
                fprintf(file, " %d", curr_revent->iParams[j]);
            fprintf(file, "\n");
        }
        else
        {
            fprintf(file, "%ld %s\n", curr_revent->index,
                                     Rv_Names[curr_revent->type]);
        }

        curr_revent = curr_revent->next;
    }

    return 0;
}
//  function -- Write_Events_File
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Write_Combined_Events_File
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int Write_Combined_Events_File(char *filename, struct RecordEvent *revent1,
                               struct RecordEvent *revent2)
{
    FILE *file=NULL;
    struct RecordEvent *curr_revent, *curr_revent1=NULL, *curr_revent2=NULL;
    int use_both_chains = 1;

    if (revent1 == NULL && revent2 == NULL)
    {
        LogFile_Printf("WARNING: Write_Combined_Events_File: empty (null)"
                       " REvent chain\n");
        return 0;
    }

    if (filename == NULL)
    {
        Error_Printf("E407: Write_Combined_Events_File: NULL filename\n");
        return -1;
    }

    //
    //  Create the Events file
    //
    file = fopen(filename, "w");
    if (file == NULL)
    {
        LogFile_Printf("E408: Write_Combined_Events_File: cannot create events"
                       " file (%s)\n", filename);
        return -1;
    }

    //
    //  One of the chains may be NULL, but we will use the valid chain
    //
    if (revent1 == NULL || revent2 == NULL)
    {
        use_both_chains = 0;
        if (revent1)
            curr_revent = curr_revent1 = revent1;
        else
            curr_revent = curr_revent2 = revent2;
    }
    else
    {
        use_both_chains = 1;
        curr_revent1 = revent1;
        curr_revent2 = revent2;
        curr_revent = curr_revent1;
    }

    while (curr_revent)
    {
        //
        //  If both chains still have links to go
        //
        if (use_both_chains)
        {
            //
            //  get the next event -- lowest song beat
            //
            if (curr_revent1->index <= curr_revent2->index)
            {
                curr_revent = curr_revent1;
                curr_revent1 = curr_revent1->next;
            }
            else
            {
                curr_revent = curr_revent2;
                curr_revent2 = curr_revent2->next;
            }
        }


        if (curr_revent->type > REVENT_MAX_RV_NAMES)
        {
            LogFile_Printf("E409: Write_Combined_Events_File: event type (%d) out-of"
                           "-range at song beat (%ld)\n",
                            curr_revent->type, curr_revent->index);
            return -1;
        }

        if (curr_revent->has_iParams)
        {
            int j;

            fprintf(file, "%ld %s", curr_revent->index,
                                     Rv_Names[curr_revent->type]);
            for (j=0; j < curr_revent->has_iParams; j++)
                fprintf(file, " %d", curr_revent->iParams[j]);
            fprintf(file, "\n");
        }
        else
        {
            fprintf(file, "%ld %s\n", curr_revent->index,
                                     Rv_Names[curr_revent->type]);
        }

        //
        //  terminate loop condition
        //
        if (use_both_chains)
        {
            if (curr_revent1 == NULL)
            {
                //
                //  end of chain
                //
                curr_revent = curr_revent2;
                use_both_chains = 0;
            }
            else if (curr_revent2 == NULL)
            {
                //
                //  end of chain
                //
                curr_revent = curr_revent1;
                use_both_chains = 0;
            }
            //
            //  Otherwise, the pointers are already incremented
            //
        }
        else
        {
            //
            //  One chain left
            //
            curr_revent = curr_revent->next;
        }
    }
    // while

    return 0;
}
//  function -- Write_Combined_Events_File
//  -------------------------------------------------------------------------

