/**
 * C-code modifications Copyright (C) 2009 Alfred Steffens Jr.
 * Copyright (C) 2001 Paul Davis, Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>
#ifdef HAVE_LINUX_RTC_H
// This is a 2.4'ism.
#include <linux/rtc.h>
#else
// mc146..rtc.h needs me to include asm/splinlock.
#include <asm/spinlock.h>
#include <linux/mc146818rtc.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <fcntl.h>
#include <math.h>
#include <sys/ioctl.h>
#include <defs.h>
#include <errorprintf.h>
#include <rtctimer.h>


//  -------------------------------------------------------------------------
//  Function    --  Init_RealTimeClock
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int Init_RealTimeClock(struct RealTimeClock *rtc)
{

    ERROR_NULL(rtc,-1);

    rtc->rtc_fd = open( "/dev/rtc", O_RDONLY);
    if (rtc->rtc_fd < 0)
    {
        Error_Printf("E170: RealTimeClock: Cannot open /dev/rtc: %s\n"
                     "You cannot do internal sync without the real "
                     "time clock, so I'm shutting down.\n",
            strerror(errno));
        return -1;
    }

    if (fcntl(rtc->rtc_fd, F_SETOWN, getpid()) < 0 )
    {
        Error_Printf("E171: RealTimeClock: cannot set ownership of "
                     "/dev/rtc: %s\n"
                     "You cannot do internal sync without the real "
                     "time clock, so I'm shutting down.\n",
            strerror(errno));
        return -1;
    }

    rtc->rtc_running = M_FALSE;
    rtc->current_hz = 0;
    return 0;
}
//  function -- Init_RealTimeClock
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_RealTimeClock
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct RealTimeClock *Create_RealTimeClock(void)
{
    struct RealTimeClock *rtc;

    rtc = (struct RealTimeClock *)malloc( sizeof(struct RealTimeClock) );
    if (rtc == NULL)
    {
        Error_Printf("E172: Create_RealTimeClock: mem alloc\n");
        return NULL;
    }

    if (Init_RealTimeClock(rtc))
        return NULL;

    return rtc;
}
//  function -- Create_RealTimeClock
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_RealTimeClock
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void Delete_RealTimeClock(struct RealTimeClock *rtc)
{
    if (rtc == NULL)
        return;

    rtc_stopClock(rtc);
    close(rtc->rtc_fd);

    free(rtc);
}
//  function -- Delete_RealTimeClock
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  rtc_nextTick
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int rtc_nextTick(struct RealTimeClock *rtc)
{
    unsigned long rtc_data;
    struct pollfd pfd;

    ERROR_NULL(rtc,0);

    pfd.fd = rtc->rtc_fd;
    pfd.events = POLLIN | POLLERR;

again:

    if (poll(&pfd, 1, 100000 ) < 0)
    {
        if (errno == EINTR)
        {
            // This happens mostly when run under gdb, or when
            // exiting due to a signal.
            goto again;
        }

        Error_Printf("E173: RealTimeClock: poll call failed: %s\n",
            strerror( errno ) );
        return 0;
    }

    read(rtc->rtc_fd, &rtc_data, sizeof(rtc_data));
    return 1;
}
//  function -- rtc_nextTick
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  rtc_setInterval
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int rtc_setInterval(struct RealTimeClock *rtc, int hz)
{
    int restart;

    ERROR_NULL(rtc,M_FALSE);

    if(hz == rtc->current_hz)
    {
        return M_TRUE;
	}

    restart = rtc_stopClock(rtc);

    if (ioctl(rtc->rtc_fd, RTC_IRQP_SET, hz) < 0 )
    {
        Error_Printf("E174: RealTimeClock: cannot set periodic interval: %s"
                     ": must run as ROOT for this\n",
            strerror( errno ) );
        return M_FALSE;
	}

    rtc->current_hz = hz;

    if (restart)
    {
        rtc_startClock(rtc);
    }

    return M_TRUE;
}
//  function -- rtc_setInterval
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  rtc_startClock
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int rtc_startClock(struct RealTimeClock *rtc)
{
    ERROR_NULL(rtc,M_FALSE);

    if (rtc->rtc_running == M_FALSE)
    {
        if (ioctl(rtc->rtc_fd, RTC_PIE_ON, 0) < 0)
        {
            Error_Printf("E175: RealTimeClock: cannot start periodic "
                "interrupts: %s\n", strerror( errno ) );
            return M_FALSE;
        }
        rtc->rtc_running = M_TRUE;
    }
    return rtc->rtc_running;
}
//  function -- rtc_startClock
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  rtc_stopClock
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int rtc_stopClock(struct RealTimeClock *rtc)
{
	int was_running;

    ERROR_NULL(rtc,M_FALSE);

    was_running = rtc->rtc_running;

    if (rtc->rtc_running)
    {
        if (ioctl(rtc->rtc_fd, RTC_PIE_OFF, 0) < 0)
        {
            Error_Printf("E176: RealTimeClock: cannot stop periodic "
                "interrupts: %s\n", strerror( errno ) );
            return rtc->rtc_running;
        }
        rtc->rtc_running = M_FALSE;
    }

    return was_running;
}
//  function -- rtc_stopClock
//  -------------------------------------------------------------------------



