/**
 * C-code modifications Copyright (C) 2009 Alfred Steffens Jr.
 * based on TekTracker, Copyright (C) 2001 Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif // HAVE_CONFIG_H
#include <stdlib.h>
#include <signal.h>
#include <slang.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>
#include <defs.h>
#include <logprintf.h>
#include <errorprintf.h>
#include <msgprintf.h>
#include <utils.h>
#include <xmltools.h>
#include <record.h>
#include <song.h>
#include <midicontrol.h>
#include <patbar.h>
#include <songbar.h>
#include <screen.h>


static volatile int Screen_Size_Changed = 0;
static const int numbars = NUM_BARS;
static char inputbuf[80];
static volatile int CopyBuffer[MAX_COPYBUFFER];

extern struct MidiControl *p_MidiControl;
extern struct RecordEvent *p_REvents, *p_CurrREvents;
extern struct RecordEvent *p_Rec, *p_CurrRec;

extern int g_Record_Session;
extern int g_Playback_Session;
extern char g_Playback_Filename[MAX_PATH_LENGTH];

int screen_WaitNc(int port, int timeout_seconds);


//  -------------------------------------------------------------------------
//  Function    --  SLang_message
//
//  Description
//
//      General print function to the screen, specifically for messages to
//      the lower, left prompt window.
//
//  Arguments
//
//      msg
//
//          Pointer to string containing the message to be printed to the
//          screen.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void SLang_message(char *msg)
{
    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(msg);
    SLsmg_refresh();
}
//  function -- SLang_message
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  inputBuffer
//
//  Description
//
//      Prompts the user for input and stores the response into the global
//      string buffer, inputbuf.
//
//  Arguments
//
//      prompt
//
//          Pointer to character string containing the prompt message.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
static void inputBuffer(char *prompt)
{
    int i = 0;
    unsigned int c;

    ERROR_NULL_VOID(prompt);

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(prompt);
    SLsmg_refresh();

    for(;;)
    {
        if (SLang_input_pending(0))
        {
            c = SLkp_getkey();
            if (c == '\r')
            {
                inputbuf[i] = '\0';
                    break;
            }
            if (c == SL_KEY_BACKSPACE )
            {
                SLsmg_gotorc(SLtt_Screen_Rows - 1, SLsmg_get_column() - 1);
                SLsmg_write_char( ' ' );
                SLsmg_gotorc(SLtt_Screen_Rows - 1, SLsmg_get_column() - 1);
                SLsmg_refresh();
                i--;
                if (i == -1)
                    i = 0;
                continue;
            }
            inputbuf[i++] = c;
            SLsmg_write_char(c);
            SLsmg_refresh();
        }
        usleep(1);
    }
    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
}
//  function -- inputBuffer
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  sigwinch_handler
//
//  Description
//
//      Signal handler inherited from the original TekTracker.
//
//  Arguments
//
//      sig
//
//          Signal identifier.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
static void sigwinch_handler( int sig )
{
    Screen_Size_Changed = 1;
    SLsignal( SIGWINCH, sigwinch_handler );
}
//  function -- sigwinch_handler
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Init_Screen
//
//  Description
//
//      Initialization routine for the Screen struct.  Called when the
//      Screen struct is allocated.
//
//  Arguments
//
//      screen
//
//          Pointer to the new Screen struct that has been newly created.
//
//      song
//
//          Pointer to the current Song struct instance for the current
//          session.
//
//  Return
//
//      Zero on success, a negative integer on failure.
//
//  -------------------------------------------------------------------------
int Init_Screen(struct Screen *screen, struct Song *song)
{
    int ret, i;

    ERROR_NULL(screen,-1);
    ERROR_NULL(song,-1);

    screen->bars = NULL;
    screen->patbars = NULL;
    screen->songbars = NULL;
    screen->curr_song = song;
    screen->octave = DEFAULT_OCTAVE;
    screen->pos = 0;
    screen->patmode = M_FALSE;
    screen->curcol = 0;
    screen->scrsize = 0;
    screen->mark_chan1 = -1;
    screen->mark_chan2 = -1;
    screen->mark_row1 = -1;
    screen->mark_row2 = -1;
    screen->copybuffer = NULL;
    screen->copyrows = 0;
    screen->copycols = 0;

    // Initialize s-lang stuff.
    SLsignal( SIGWINCH, sigwinch_handler );
    SLtt_get_terminfo();
    SLang_init_tty( 0, 1, 0 );
    ret = SLkp_init();
    assert( ret != -1 );
    ret = SLsmg_init_smg();
    assert( ret != -1 );

    // Define all the extra keys.
    SLkp_define_keysym( "[21~", SL_KEY_F10 );
    SLkp_define_keysym( "[23~", SL_KEY_F11 );
    SLkp_define_keysym( "[24~", SL_KEY_F12 );
    SLkp_define_keysym( "[2D", SL_KEY_SHIFTLEFT );
    SLkp_define_keysym( "[2C", SL_KEY_SHIFTRIGHT );
    SLkp_define_keysym( "[2A", SL_KEY_SHIFTUP );
    SLkp_define_keysym( "[2B", SL_KEY_SHIFTDOWN );

    // Cursor colour.
    SLtt_set_color(CURSOR_COLOR, 0, "white", "blue" );

    // Hilight colour.
    SLtt_set_color(HILIGHT_COLOR, 0, "green", "black" );

    // Hilight + cursor colour.
    SLtt_set_color(HILIGHTCURSOR_COLOR, 0, "green", "blue" );

    // Muted channel colour.
    SLtt_set_color(MUTEDCHANNEL_COLOR, 0, "blue", "black" );

    // Realtime Light colour.
    SLtt_set_color(REALTIMELIGHT_COLOR, 0, "white", "red" );

    // Play position colour.
    SLtt_set_color(PLAYPOSITION_COLOR, 0, "red", "black" );

    // Loop hilight colour.
    SLtt_set_color(LOOPHILIGHT_COLOR, 0, "black", "green" );

    // Loop inactive hilight colour.
    SLtt_set_color(LOOPINACTIVE_COLOR, 0, "black", "red" );

    // Default colour.
    SLtt_set_color( 0, 0, "white", "black" );

    // Remember to stay at the default colour from now on.
    SLsmg_set_color(DEFAULT_COLOR);

    //
    // Create the generic bars.
    //
    screen->bars = (void **)malloc(
                       sizeof(void *) * numbars
                      );
    if (screen->bars == NULL)
    {
        Error_Printf("E200: Init_Screen: mem alloc bars\n");
        goto Error;
    }

    //
    // Create the pattern bars.
    //
    screen->patbars = (struct PatBar **)malloc(
                       sizeof(struct PatBar *) * numbars
                      );
    if (screen->patbars == NULL)
    {
        Error_Printf("E201: Init_Screen: mem alloc parbars\n");
        goto Error;
    }
    for (i = 0; i < numbars; i++)
        screen->patbars[i] = NULL;
    for (i = 0; i < numbars; i++)
    {
        screen->patbars[i] = Create_PatBar(i, 3, 5+(i*9));
        if (screen->patbars[i] == NULL)
            goto Error;
    }

    //
	// Create the channel bars.
    //
    screen->songbars = (struct SongBar **)malloc(
                       sizeof(struct SongBar *) * numbars
                      );
    if (screen->songbars == NULL)
    {
        Error_Printf("E202: Init_Screen: mem alloc songbars\n");
        goto Error;
    }
    for (i = 0; i < numbars; i++)
        screen->songbars[i] = NULL;
    for (i = 0; i < numbars; i++)
    {
        screen->songbars[i] = Create_SongBar(i, 3, 5+(i*9));
        if (screen->songbars[i] == NULL)
            goto Error;
    }

    //  Init the Kmarker array
    for (i = 0; i < 26; i++)
    {
        screen->kmark[i].row = 0;
        screen->kmark[i].col = 0;
    }

    //
    //  Create the "copyPattern"
    //
    screen->copyPattern = Create_SongPattern(0);


    // Clear the virtual screen.
    SLsmg_cls();

	// Print the help.  This is only printed once.
    screen_printHelp(screen);

    // Switch the mode to setup everything initially.
    screen->patmode = 0;
	screen_switchMode(screen);

    // Reset the cursor.
	pbar_setCursor(screen->patbars[0], 0);
	sbar_setCursor(screen->songbars[0], 0);
	screen_moveCursorPosition(screen, 0, 0);

	// Refresh the screen.
	screen_refresh(screen);

    return 0;

Error:
    if (screen->bars)
    {
        free(screen->bars);
    }
    if (screen->patbars)
    {
        for (i = 0; i < numbars; i++)
            if (screen->patbars[i])
                free(screen->patbars[i]);
        free(screen->patbars);
    }
    if (screen->songbars)
    {
        for (i = 0; i < numbars; i++)
            if (screen->songbars[i])
                free(screen->songbars[i]);
        free(screen->songbars);
    }
    return -1;
}
//  function -- Init_Screen
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_Screen
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct Screen *Create_Screen(struct Song *song)
{
    struct Screen *screen;

    ERROR_NULL(song,NULL);

    screen = (struct Screen *)malloc( sizeof(struct Screen) );
    if (screen == NULL)
    {
        Error_Printf("E203: Create_Screen: mem alloc\n");
        return NULL;
    }

    if (Init_Screen(screen, song))
    {
        free(screen);
        return NULL;
    }

    return screen;
}
//  function -- Create_Screen
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_Screen
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void Delete_Screen(struct Screen *screen)
{
    if (screen == NULL)
        return;

    SLsmg_reset_smg();
    SLang_reset_tty();


    if (screen->bars)
    {
        free(screen->bars);
    }
    if (screen->patbars)
    {
        int i;
        for (i = 0; i < numbars; i++)
            if (screen->patbars[i])
                free(screen->patbars[i]);
        free(screen->patbars);
    }
    if (screen->songbars)
    {
        int i;
        for (i = 0; i < numbars; i++)
            if (screen->songbars[i])
                free(screen->songbars[i]);
        free(screen->songbars);
    }
    if (screen->copyPattern)
        Delete_SongPattern(screen->copyPattern);

    if (screen->copybuffer)
        free(screen->copybuffer);

    free(screen);
}
//  function -- Delete_Screen
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_switchMode
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_switchMode(struct Screen *screen)
{
	int i;

    ERROR_NULL_VOID(screen);

    screen->patmode = !screen->patmode;
    if (screen->patmode)
    {
        for (i = 0; i < numbars; ++i)
        {
            screen->bars[i] = (void *)screen->patbars[i];
            pbar_refresh(screen->patbars[i]);
        }
        screen->curstate = screen->patstate;
    }
    else
    {
        for (i = 0; i < numbars; ++i)
        {
            screen->bars[i] = (void *)screen->songbars[i];
            sbar_refresh(screen->songbars[i]);
        }
        screen->curstate = screen->songstate;
    }
    screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol);
}
//  function -- screen_switchMode
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_refresh
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_refresh(struct Screen *screen)
{
    ERROR_NULL_VOID(screen);

    // Need to refresh the header.
    screen_printHeader(screen);

    // Need to refresh the sidebar too.
    screen_printSidebar(screen);

    // Print beat pulse.
    if (mctrl_beatPulse(p_MidiControl))
    {
        SLsmg_gotorc(0, SLtt_Screen_Cols - 2);
        SLsmg_printf("*");
    }
    else
    {
        SLsmg_gotorc(0, SLtt_Screen_Cols - 2);
        SLsmg_printf( " " );
    }

    // Announce if we got leet root access.
    SLsmg_gotorc(SLtt_Screen_Rows - 2, 0);
    if (mctrl_isRealtime(p_MidiControl))
    {
        SLsmg_forward(SLtt_Screen_Cols - 2);
        SLsmg_set_color(REALTIMELIGHT_COLOR);
        SLsmg_printf("RT");
        SLsmg_set_color(DEFAULT_COLOR);
	}

    // Always leave the cursor in the bottom-right corner.
    SLsmg_gotorc(SLtt_Screen_Rows - 1, SLtt_Screen_Cols - 1);
    SLsmg_refresh();
}
//  function -- screen_refresh
//  -------------------------------------------------------------------------


#ifdef OLD_SCREEN
//  -------------------------------------------------------------------------
//  Function    --  screen_pollScreen
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int screen_pollScreen(struct Screen *screen)
{
    unsigned int c;

    ERROR_NULL(screen,-1);
    ERROR_NULL(screen->curr_song,-1);

    //
    // Handle a screen size change (annoying).
    //
    if (Screen_Size_Changed)
    {
        SLtt_get_screen_size();
#ifdef HAVE_SLSMG_REINIT_SMG
        SLsmg_reinit_smg();
#else
        SLsmg_reset_smg();
        SLsmg_init_smg();
#endif // HAVE_SLSMG_REINIT_SMG
        Screen_Size_Changed = 0;
        screen_refresh(screen);
        screen_printHelp(screen);
        if (screen->patmode)
        {
            int i;
            for (i = 0; i < numbars; i++)
            {
                pbar_refresh(screen->patbars[i]);
            }
        }
        else
        {
            int i;
            for (i = 0; i < numbars; i++)
            {
                sbar_refresh(screen->songbars[i]);
            }
        }
    }

    //
    // This input call will throttle the main loop.
    //
    if (SLang_input_pending(0))
    {
        c = SLkp_getkey();
        c = toupper(c);
        switch (c)
        {

        case 17: return M_TRUE; // Ctrl-Q is quit

        //
        // Channel mutes.
        //
        case SL_KEY_F(1):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[0]);
            else
                sbar_toggleChannelMute(screen->songbars[0]);
        break;
        case SL_KEY_F(2):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[1]);
            else
                sbar_toggleChannelMute(screen->songbars[1]);
        break;
        case SL_KEY_F(3):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[2]);
            else
                sbar_toggleChannelMute(screen->songbars[2]);
        break;
        case SL_KEY_F(4):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[3]);
            else
                sbar_toggleChannelMute(screen->songbars[3]);
        break;
        case SL_KEY_F(5):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[4]);
            else
                sbar_toggleChannelMute(screen->songbars[4]);
        break;
        case SL_KEY_F(6):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[5]);
            else
                sbar_toggleChannelMute(screen->songbars[5]);
        break;
        case SL_KEY_F(7):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[6]);
            else
                sbar_toggleChannelMute(screen->songbars[6]);
        break;
        case SL_KEY_F(8):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[7]);
            else
                sbar_toggleChannelMute(screen->songbars[7]);
        break;

        //
        // Song Control.
        //
        case SL_KEY_F(9):
            song_start(screen->curr_song);
        break;
        case SL_KEY_F10:
            song_rewind(screen->curr_song);
            song_start(screen->curr_song);
        break;
        case SL_KEY_F11:
            song_stop(screen->curr_song);
        break;
        case SL_KEY_F12:
            song_rewind(screen->curr_song);
        break;

        case SL_KEY_DOWN:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow+1,
                              screen->curcol);
        break;
        case SL_KEY_UP:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow-1,
                              screen->curcol);
        break;
        case SL_KEY_LEFT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol-1);
        break;
        case SL_KEY_RIGHT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol+1);
        break;

        case SL_KEY_SHIFTDOWN:
        case '':
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow+screen->scrsize,
                              screen->curcol);
        break;
        case SL_KEY_SHIFTUP:
        case '':
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow-screen->scrsize,
                              screen->curcol);
        break;

        case SL_KEY_SHIFTLEFT:
        case '':
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol-numbars);
        break;
        case SL_KEY_SHIFTRIGHT:
        case '':
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol+numbars);
        break;

        case SL_KEY_IC:
            if (screen->patmode)
                pbar_editMidiChannel(screen->patbars[screen->curcol], 1);
            else
                sbar_editMidiChannel(screen->songbars[screen->curcol], 1);
        break;
        case SL_KEY_DELETE:
            if (screen->patmode)
                pbar_editMidiChannel(screen->patbars[screen->curcol], -1);
            else
                sbar_editMidiChannel(screen->songbars[screen->curcol], -1);
        break;

        case '/':
            if (screen->patmode)
                pbar_editChannelVolume(screen->patbars[screen->curcol], -1);
            else
                sbar_editChannelVolume(screen->songbars[screen->curcol], -1);
        break;
        case '*':
            if (screen->patmode)
                pbar_editChannelVolume(screen->patbars[screen->curcol], 1);
            else
                sbar_editChannelVolume(screen->songbars[screen->curcol], 1);
        break;

        case SL_KEY_HOME:
            if (screen->patmode)
            {
                pbar_editCursorVolume(screen->patbars[screen->curcol], 1);
            }
        break;
        case SL_KEY_END:
            if (screen->patmode)
            {
                pbar_editCursorVolume(screen->patbars[screen->curcol], -1);
            }
        break;

        case '[':
            if (screen->patmode)
                pbar_editPattern(screen->patbars[screen->curcol], -1);
            else
                sbar_editPattern(screen->songbars[screen->curcol], -1);
        break;
        case ']':
            if (screen->patmode)
                pbar_editPattern(screen->patbars[screen->curcol], 1);
            else
                sbar_editPattern(screen->songbars[screen->curcol], 1);
        break;

        case '+': ++screen->octave; break;
        case '-': --screen->octave; break;

        case '\'':
            if (screen->patmode)
                pbar_transposeNotes(screen->patbars[screen->curcol], -1);
        break;
        case '\\':
            if (screen->patmode)
                pbar_transposeNotes(screen->patbars[screen->curcol], 1);
        break;

        case SL_KEY_PPAGE:
            if (screen->patmode)
            {
                screen_setPosition(screen, screen->pos-1);
            }
            else
            {
                screen_moveCursorPosition(screen, screen->curstate.cursorrow-8,
                                          screen->curcol );
            }
            break;
        case SL_KEY_NPAGE:
            if (screen->patmode)
            {
                screen_setPosition(screen, screen->pos+1);
            }
            else
            {
                screen_moveCursorPosition(screen, screen->curstate.cursorrow+8,
                                          screen->curcol );
            }
            break;

        case '`':
            if (screen->patmode)
            {
                pbar_toggleCursorSlide(screen->patbars[screen->curcol]);
                screen_moveCursorPosition(screen, screen->patstate.cursorrow+1,
                                          screen->curcol);
            }
            break;

        case ' ':
            if (screen->patmode)
            {
                screen_setNote(screen, SONG_NOTEEMPTY);
            }
            else
            {
                sbar_editPattern(screen->songbars[screen->curcol], 0-(SONGMAXPAT+1));
                screen_moveCursorPosition(screen, screen->curstate.cursorrow+1,
                                          screen->curcol);
            }
            break;

        case 'Z':
            screen_setNote(screen, (screen->octave+1)*12+0);
        break;
        case 'S':
            screen_setNote(screen, (screen->octave+1)*12+1);
        break;
        case 'X':
            screen_setNote(screen, (screen->octave+1)*12+2);
        break;
        case 'D':
            screen_setNote(screen, (screen->octave+1)*12+3);
        break;
        case 'C':
            screen_setNote(screen, (screen->octave+1)*12+4);
        break;
        case 'V':
            screen_setNote(screen, (screen->octave+1)*12+5);
        break;
        case 'G':
            screen_setNote(screen, (screen->octave+1)*12+6);
        break;
        case 'B':
            screen_setNote(screen, (screen->octave+1)*12+7);
        break;
        case 'H':
            screen_setNote(screen, (screen->octave+1)*12+8);
        break;
        case 'N':
            screen_setNote(screen, (screen->octave+1)*12+9);
        break;
        case 'J':
            screen_setNote(screen, (screen->octave+1)*12+10);
        break;
        case 'M':
            screen_setNote(screen, (screen->octave+1)*12+11);
        break;

        case 'Q':
            screen_setNote(screen, (screen->octave+1)*12+12);
        break;
        case '2':
            screen_setNote(screen, (screen->octave+1)*12+13);
        break;
        case 'W':
            screen_setNote(screen, (screen->octave+1)*12+14);
        break;
        case '3':
            screen_setNote(screen, (screen->octave+1)*12+15);
        break;
        case 'E':
            screen_setNote(screen, (screen->octave+1)*12+16);
        break;
        case 'R':
            screen_setNote(screen, (screen->octave+1)*12+17);
        break;
        case '5':
            screen_setNote(screen, (screen->octave+1)*12+18);
        break;
        case 'T':
            screen_setNote(screen, (screen->octave+1)*12+19);
        break;
        case '6':
            screen_setNote(screen, (screen->octave+1)*12+20);
        break;
        case 'Y':
            screen_setNote(screen, (screen->octave+1)*12+21);
        break;
        case '7':
            screen_setNote(screen, (screen->octave+1)*12+22);
        break;
        case 'U':
            screen_setNote(screen, (screen->octave+1)*12+23);
        break;

        case 'I':
            screen_setNote(screen, (screen->octave+1)*12+24);
        break;
        case '9':
            screen_setNote(screen, (screen->octave+1)*12+25);
        break;
        case 'O':
            screen_setNote(screen, (screen->octave+1)*12+26);
        break;
        case '0':
            screen_setNote(screen, (screen->octave+1)*12+27);
        break;
        case 'P':
            screen_setNote(screen, (screen->octave+1)*12+28);
        break;

        case SL_KEY_BACKSPACE:
            screen_setNote(screen, SONG_NOTEOFF);
        break;

        case '':
            screen_inputTempo(screen);
        break;
        case '':
            song_setLoopPosition(screen->curr_song, screen->pos+1);
        break;
        case '':
            song_setLoopDestination(screen->curr_song, screen->pos);
        break;
        case '':
             song_setStartPos(screen->curr_song, screen->pos);
        break;
        case  19 :
            screen_saveSongAscii(screen); // Ctrl-S
        break;
        case '':
            ; // Ctrl-W
        break;
        case '':
             screen_loadSongAscii(screen); // Ctrl-A
        break;
        case '':
             screen_toggleSync(screen);
        break;
        case '':
            screen_saveCopy(screen);
        break;
        case '':
            screen_pasteCopy(screen);
        break;
        case '':
            screen_enterChanName(screen);
        break;
        case '':
            screen_enterPatName(screen);
        break;
        case '':
            screen_toggleEndNote(screen);
        break;
        case '':
            screen_editMidiPort(screen);
        break;
        case '':
            screen_editMidiPatch(screen);
        break;
        case  '!':
            song_toggleLoop(screen->curr_song);
        break;
        case '	':
            screen_switchMode(screen);
        break;
        case '':
            screen_showMIDIinfo(screen);
        break;
        default:
            SLtt_beep();
        break;
        }
    }

    screen_refresh(screen);
    return M_FALSE;
}
//  function -- screen_pollScreen
//  -------------------------------------------------------------------------
#endif // OLD_SCREEN


//  -------------------------------------------------------------------------
//  Function    --  screen_pollScreenRec
//
//  Description
//
//      The function that polls the keyboard input.  This function replaces
//      screen_pollScreen because it can record session events.
//
//  Arguments
//
//      screen
//
//          Pointer to a Screen struct instance.  There is actually only one
//          screen struct at a time, so this is THE screen.
//
//  Return
//
//      Returns 0 on success, nonzero to Quit.
//
//  -------------------------------------------------------------------------
int screen_pollScreenRec(struct Screen *screen)
{
    unsigned int c;

    ERROR_NULL(screen,-1);
    ERROR_NULL(screen->curr_song,-1);

    //
    // Handle a screen size change (annoying).
    //
    if (Screen_Size_Changed)
    {
        SLtt_get_screen_size();
#ifdef HAVE_SLSMG_REINIT_SMG
        SLsmg_reinit_smg();
#else
        SLsmg_reset_smg();
        SLsmg_init_smg();
#endif // HAVE_SLSMG_REINIT_SMG
        Screen_Size_Changed = 0;
        screen_refresh(screen);
        screen_printHelp(screen);
        if (screen->patmode)
        {
            int i;
            for (i = 0; i < numbars; i++)
            {
                pbar_refresh(screen->patbars[i]);
            }
        }
        else
        {
            int i;
            for (i = 0; i < numbars; i++)
            {
                sbar_refresh(screen->songbars[i]);
            }
        }
    }

    //
    // This input call will throttle the main loop.
    //
    if (SLang_input_pending(0))
    {
        c = SLkp_getkey();
        c = toupper(c);
        switch (c)
        {

        case 17:
            if (screen->curr_song->re_record)
                Write_Events_File("session.dat", p_Rec);

            return M_TRUE; // Ctrl-Q is quit

        //
        // Channel mutes.
        //
        case SL_KEY_F(1):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[0]);
            else
                sbar_toggleChannelMute(screen->songbars[0]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE1,
                                          NULL);
        break;
        case SL_KEY_F(2):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[1]);
            else
                sbar_toggleChannelMute(screen->songbars[1]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE2,
                                          NULL);
        break;
        case SL_KEY_F(3):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[2]);
            else
                sbar_toggleChannelMute(screen->songbars[2]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE3,
                                          NULL);
        break;
        case SL_KEY_F(4):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[3]);
            else
                sbar_toggleChannelMute(screen->songbars[3]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE4,
                                          NULL);
        break;
        case SL_KEY_F(5):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[4]);
            else
                sbar_toggleChannelMute(screen->songbars[4]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE5,
                                          NULL);
        break;
        case SL_KEY_F(6):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[5]);
            else
                sbar_toggleChannelMute(screen->songbars[5]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE6,
                                          NULL);
        break;
        case SL_KEY_F(7):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[6]);
            else
                sbar_toggleChannelMute(screen->songbars[6]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE7,
                                          NULL);
        break;
        case SL_KEY_F(8):
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[7]);
            else
                sbar_toggleChannelMute(screen->songbars[7]);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_CHMUTE8,
                                          NULL);
        break;

        //
        // Song Control.
        //
        case SL_KEY_F(9):
            song_start(screen->curr_song);
        break;
        case SL_KEY_F10:
            song_rewind(screen->curr_song);
            song_start(screen->curr_song);
        break;
        case SL_KEY_F11:
            song_stop(screen->curr_song);
        break;
        case SL_KEY_F12:
            song_rewind(screen->curr_song);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SONG_REWIND,
                                          NULL);
        break;

        case SL_KEY_DOWN:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow+1,
                              screen->curcol);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_KEY_DOWN,
                                          NULL);
        break;
        case SL_KEY_UP:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow-1,
                              screen->curcol);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_KEY_UP,
                                          NULL);
        break;
        case SL_KEY_LEFT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol-1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_KEY_LEFT,
                                          NULL);
        break;
        case SL_KEY_RIGHT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol+1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_KEY_RIGHT,
                                          NULL);
        break;

        case SL_KEY_SHIFTDOWN:
        case '':
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow+screen->scrsize,
                              screen->curcol);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SHIFT_KEY_DOWN,
                                          NULL);
        break;
        case SL_KEY_SHIFTUP:
        case '':
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow-screen->scrsize,
                              screen->curcol);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SHIFT_KEY_UP,
                                          NULL);
        break;

        case SL_KEY_SHIFTLEFT:
        case '':
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol-numbars);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SHIFT_KEY_LEFT,
                                          NULL);
        break;
        case SL_KEY_SHIFTRIGHT:
        case '':
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol+numbars);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SHIFT_KEY_RIGHT,
                                          NULL);
        break;

        case SL_KEY_IC:
            if (screen->patmode)
                pbar_editMidiChannel(screen->patbars[screen->curcol], 1);
            else
                sbar_editMidiChannel(screen->songbars[screen->curcol], 1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_MIDICHANNEL_UP,
                                          NULL);
        break;
        case SL_KEY_DELETE:
            if (screen->patmode)
                pbar_editMidiChannel(screen->patbars[screen->curcol], -1);
            else
                sbar_editMidiChannel(screen->songbars[screen->curcol], -1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_MIDICHANNEL_DOWN,
                                          NULL);
        break;

        case '/':
            if (screen->patmode)
                pbar_editChannelVolume(screen->patbars[screen->curcol], -1);
            else
                sbar_editChannelVolume(screen->songbars[screen->curcol], -1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_VOLUME_DOWN,
                                          NULL);
        break;
        case '*':
            if (screen->patmode)
                pbar_editChannelVolume(screen->patbars[screen->curcol], 1);
            else
                sbar_editChannelVolume(screen->songbars[screen->curcol], 1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_VOLUME_UP,
                                          NULL);
        break;

        case SL_KEY_HOME:
            if (screen->patmode)
            {
                pbar_editCursorVolume(screen->patbars[screen->curcol], 1);

                if (screen->curr_song->re_record)
                    p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_VELOCITY_UP,
                                          NULL);
            }
        break;
        case SL_KEY_END:
            if (screen->patmode)
            {
                pbar_editCursorVolume(screen->patbars[screen->curcol], -1);

                if (screen->curr_song->re_record)
                    p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_VELOCITY_DOWN,
                                          NULL);
            }
        break;

        case '[':
            if (screen->patmode)
                pbar_editPattern(screen->patbars[screen->curcol], -1);
            else
                sbar_editPattern(screen->songbars[screen->curcol], -1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_PATTERN_DOWN,
                                          NULL);
        break;
        case ']':
            if (screen->patmode)
                pbar_editPattern(screen->patbars[screen->curcol], 1);
            else
                sbar_editPattern(screen->songbars[screen->curcol], 1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_PATTERN_UP,
                                          NULL);
        break;

        case '+':
            ++screen->octave;

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_OCTAVE_UP,
                                          NULL);
        break;
        case '-':
            --screen->octave;

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_OCTAVE_DOWN,
                                          NULL);
        break;

        case '\'':
            if (screen->patmode)
                pbar_transposeNotes(screen->patbars[screen->curcol], -1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SEMITONE_UP,
                                          NULL);
        break;
        case '\\':
            if (screen->patmode)
                pbar_transposeNotes(screen->patbars[screen->curcol], 1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SEMITONE_DOWN,
                                          NULL);
        break;

        case SL_KEY_PPAGE:
            if (screen->patmode)
                screen_setPosition(screen, screen->pos-1);
            else
                screen_moveCursorPosition(screen, screen->curstate.cursorrow-8,
                                          screen->curcol );

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SONGPOSITION_UP,
                                          NULL);
            break;
        case SL_KEY_NPAGE:
            if (screen->patmode)
                screen_setPosition(screen, screen->pos+1);
            else
                screen_moveCursorPosition(screen, screen->curstate.cursorrow+8,
                                          screen->curcol );

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SONGPOSITION_DOWN,
                                          NULL);
            break;

        case '`':
            if (screen->patmode)
            {
                pbar_toggleCursorSlide(screen->patbars[screen->curcol]);
                screen_moveCursorPosition(screen, screen->patstate.cursorrow+1,
                                          screen->curcol);

                if (screen->curr_song->re_record)
                    p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_TOGGLE_SLIDE,
                                          NULL);
            }
            break;

        case ' ':
            if (screen->patmode)
            {
                screen_setNote(screen, SONG_NOTEEMPTY);
            }
            else
            {
                sbar_editPattern(screen->songbars[screen->curcol], 0-(SONGMAXPAT+1));
                screen_moveCursorPosition(screen, screen->curstate.cursorrow+1,
                                          screen->curcol);
            }

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_EMPTY,
                                          NULL);
            break;

        case 'Z':
            screen_setNote(screen, (screen->octave+1)*12+0);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_C__LO,
                                          NULL);
        break;
        case 'S':
            screen_setNote(screen, (screen->octave+1)*12+1);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_CS_LO,
                                          NULL);
        break;
        case 'X':
            screen_setNote(screen, (screen->octave+1)*12+2);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_D__LO,
                                          NULL);
        break;
        case 'D':
            screen_setNote(screen, (screen->octave+1)*12+3);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_DS_LO,
                                          NULL);
        break;
        case 'C':
            screen_setNote(screen, (screen->octave+1)*12+4);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_E__LO,
                                          NULL);
        break;
        case 'V':
            screen_setNote(screen, (screen->octave+1)*12+5);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_F__LO,
                                          NULL);
        break;
        case 'G':
            screen_setNote(screen, (screen->octave+1)*12+6);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_FS_LO,
                                          NULL);
        break;
        case 'B':
            screen_setNote(screen, (screen->octave+1)*12+7);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_G__LO,
                                          NULL);
        break;
        case 'H':
            screen_setNote(screen, (screen->octave+1)*12+8);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_GS_LO,
                                          NULL);
        break;
        case 'N':
            screen_setNote(screen, (screen->octave+1)*12+9);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_A__LO,
                                          NULL);
        break;
        case 'J':
            screen_setNote(screen, (screen->octave+1)*12+10);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_AS_LO,
                                          NULL);
        break;
        case 'M':
            screen_setNote(screen, (screen->octave+1)*12+11);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_B__LO,
                                          NULL);
        break;

        case 'Q':
            screen_setNote(screen, (screen->octave+1)*12+12);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_C__HI,
                                          NULL);
        break;
        case '2':
            screen_setNote(screen, (screen->octave+1)*12+13);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_CS_HI,
                                          NULL);
        break;
        case 'W':
            screen_setNote(screen, (screen->octave+1)*12+14);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_D__HI,
                                          NULL);
        break;
        case '3':
            screen_setNote(screen, (screen->octave+1)*12+15);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_DS_HI,
                                          NULL);
        break;
        case 'E':
            screen_setNote(screen, (screen->octave+1)*12+16);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_E__HI,
                                          NULL);
        break;
        case 'R':
            screen_setNote(screen, (screen->octave+1)*12+17);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_F__HI,
                                          NULL);
        break;
        case '5':
            screen_setNote(screen, (screen->octave+1)*12+18);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_FS_HI,
                                          NULL);
        break;
        case 'T':
            screen_setNote(screen, (screen->octave+1)*12+19);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_G__HI,
                                          NULL);
        break;
        case '6':
            screen_setNote(screen, (screen->octave+1)*12+20);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_GS_HI,
                                          NULL);
        break;
        case 'Y':
            screen_setNote(screen, (screen->octave+1)*12+21);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_A__HI,
                                          NULL);
        break;
        case '7':
            screen_setNote(screen, (screen->octave+1)*12+22);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_AS_HI,
                                          NULL);
        break;
        case 'U':
            screen_setNote(screen, (screen->octave+1)*12+23);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_B__HI,
                                          NULL);
        break;

        case 'I':
            screen_setNote(screen, (screen->octave+1)*12+24);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_C__HX,
                                          NULL);
        break;
        case '9':
            screen_setNote(screen, (screen->octave+1)*12+25);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_CS_HX,
                                          NULL);
        break;
        case 'O':
            screen_setNote(screen, (screen->octave+1)*12+26);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_D__HX,
                                          NULL);
        break;
        case '0':
            screen_setNote(screen, (screen->octave+1)*12+27);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_DS_HX,
                                          NULL);
        break;
        case 'P':
            screen_setNote(screen, (screen->octave+1)*12+28);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_E__HX,
                                          NULL);
        break;

        case SL_KEY_BACKSPACE:
            screen_setNote(screen, SONG_NOTEOFF);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_NOTE_OFF,
                                          NULL);
        break;

        case '':
            //
            //  Revent record code is embedded inside this function
            //
            screen_inputTempoRec(screen);
        break;
        case '':
            song_setLoopPosition(screen->curr_song, screen->pos+1);

            if (screen->curr_song->re_record)
            {
                static int iParams[REVENT_MAX_IPARMS];

                ERROR_RANGE(0,REVENT_MAX_IPARMS,"loop position index");
                iParams[0] = screen->pos+1;
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_LOOP_POSITION,
                                          (int *)iParams);
            }
        break;
        case '':
            song_setLoopDestination(screen->curr_song, screen->pos);

            if (screen->curr_song->re_record)
            {
                static int iParams[REVENT_MAX_IPARMS];

                ERROR_RANGE(0,REVENT_MAX_IPARMS,"loop position index");
                iParams[0] = screen->pos;
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_LOOP_DESTINATION,
                                          (int *)iParams);
            }
        break;
        case '':
             song_setStartPos(screen->curr_song, screen->pos);

            if (screen->curr_song->re_record)
            {
                static int iParams[REVENT_MAX_IPARMS];

                ERROR_RANGE(0,REVENT_MAX_IPARMS,"loop position index");
                iParams[0] = screen->pos;
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_START_POSITION,
                                          (int *)iParams);
            }
        break;
        case  19 :
            screen_saveSongAscii(screen); // Ctrl-S
        break;
        case '':
            screen_loadSession(screen); // Ctrl-W, load session
        break;
        case '':
             screen_loadSongAscii(screen); // Ctrl-A
        break;
        case '':
             screen_toggleSync(screen);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_TOGGLE_SYNC,
                                          NULL);
        break;
        case '':
            screen_saveCopy(screen);
        break;
        case '':
            screen_pasteCopy(screen);
        break;
        case '':
            screen_enterChanName(screen);
        break;
        case '':
            screen_enterPatName(screen);
        break;
        case '':
            screen_toggleEndNote(screen);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_TOGGLE_ENDNOTE,
                                          NULL);
        break;
        case '':
            //
            //  Revent record code is embedded inside this function
            //
            screen_editMidiPortRec(screen);
        break;
        case '':
            //
            //  Revent record code is embedded inside this function
            //
            screen_editMidiPatchRec(screen);
        break;
        case  '!':
            song_toggleLoop(screen->curr_song);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_LOOP_ON,
                                          NULL);
        break;
        case '	':
            screen_switchMode(screen);

            if (screen->curr_song->re_record)
                p_CurrRec = Attach_Revent(p_CurrRec,
                                          screen->curr_song->beatcount,
                                          REVENT_SWITCH_MODE,
                                          NULL);
        break;
        case '':
            screen_showMIDIinfo(screen);
        break;

        case ':':
            //
            //  Extendex commamnd
            //
            screen_extCommandRec(screen);
        break;

        default:
            SLtt_beep();
        break;
        }
    }

    screen_refresh(screen);
    return M_FALSE;
}
//  function -- screen_pollScreenRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_printHeader
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_printHeader(struct Screen *screen)
{
    static char status[16];
    static char loopstat[16];
    static char tempo[4];

    ERROR_NULL_VOID(screen);

    if (song_isPlaying(screen->curr_song))
    {
        strcpy(status, "Playing");
    }
    else
    {
        strcpy(status, "Stopped");
    }

    if (song_isLoopActive(screen->curr_song))
    {
        strcpy(loopstat, "  Active");
    }
    else
    {
        strcpy(loopstat, "Inactive");
    }

    if (mctrl_usingInternalClock(p_MidiControl))
    {
        sprintf(tempo, "%03d", song_getTempo(screen->curr_song));
	}
    else
    {
        strcpy(tempo, "---");
    }

    SLsmg_gotorc(0, 0);
    SLsmg_printf(" %03x:%02d %s          Speed: %sbpm            Loop: %s, %03x -> %03x",
        song_getCurPat(screen->curr_song), song_getCurBeat(screen->curr_song),
        status, tempo, loopstat,
		song_getLoopPosition(screen->curr_song),
        song_getLoopDestination(screen->curr_song));
    SLsmg_gotorc(1, 0);
    SLsmg_printf( "[%03x]                  Viewing: %03x             Octave: %02d",
        song_getStartPos(screen->curr_song), screen->pos, screen->octave);
}
//  function -- screen_printHeader
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_printSidebar
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_printSidebar(struct Screen *screen)
{
	int i;
    int scrsize;

    ERROR_NULL_VOID(screen);

    SLsmg_gotorc(8, 0);
    SLsmg_printf("====");

    scrsize = SLtt_Screen_Rows - 3 - HEADERSIZE - BOTTOMSIZE;
    screen->scrsize = scrsize;
    if (screen->patmode)
    {
        #ifdef ORIG_PATVIEW
        for (i = screen->patstate.scrolltop; i < SONGPATLEN; i++)
        {
            if (i >= (screen->patstate.scrolltop + scrsize))
                break;
            SLsmg_gotorc(9 + i - screen->patstate.scrolltop, 0);
            if (song_getCurPat(screen->curr_song) == screen->pos &&
                               song_getCurBeat(screen->curr_song) == i)
            {
                SLsmg_set_color(PLAYPOSITION_COLOR);
            }
            else if (i % 4 == 0)
                SLsmg_set_color(HILIGHT_COLOR);
            SLsmg_printf(" %02d: ", i);
            SLsmg_set_color(DEFAULT_COLOR);
        }
        #else
        for (i = screen->patstate.scrolltop; i < SONGPATLEN; i++)
        {
            if (i >= (screen->patstate.scrolltop + scrsize))
                break;
            SLsmg_gotorc(9 + i - screen->patstate.scrolltop, 0);
            if (song_getCurPat(screen->curr_song) == screen->pos &&
                               song_getCurBeat(screen->curr_song) == i)
            {
                SLsmg_set_color(PLAYPOSITION_COLOR);
            }
            else if (i % 4 == 0)
            {
                SLsmg_set_color(HILIGHT_COLOR);
            }
            SLsmg_printf(" %02d: ", i);
            SLsmg_set_color(DEFAULT_COLOR);
        }
        #endif // ORIG_PATVIEW
    }
    else
    {
        for (i = screen->songstate.scrolltop; i < SONGLEN; i++)
        {
            if (i >= (screen->songstate.scrolltop + scrsize))
                break;
            SLsmg_gotorc(9 + i - screen->songstate.scrolltop, 0);
            if (song_getCurPat(screen->curr_song) == i)
            {
                SLsmg_set_color(PLAYPOSITION_COLOR);
            }
            else if (i % 4 == 0)
            {
                SLsmg_set_color(HILIGHT_COLOR);
            }
            SLsmg_printf("%03x:", i);
            if (i < song_getLoopPosition(screen->curr_song) &&
                i >= song_getLoopDestination(screen->curr_song))
            {
                if (song_isLoopActive(screen->curr_song))
                {
                    SLsmg_set_color(LOOPHILIGHT_COLOR);
                }
                else
                {
                    SLsmg_set_color(LOOPINACTIVE_COLOR);
                }
            }
            SLsmg_printf(" ");
            SLsmg_set_color(DEFAULT_COLOR);
        }
    }
}
//  function -- screen_printSidebar
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_printHelp
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_printHelp(struct Screen *screen)
{
	int currow;

    ERROR_NULL_VOID(screen);

    currow = SLtt_Screen_Rows - 8;
    SLsmg_gotorc( currow, 0 );
    SLsmg_printf( "%13s %20s %20s %20s",
        "Quit: ^Q", "SetStart: ^F   ", "View: PgUp/Dn", "ChanMute: F1-F8" );
    SLsmg_gotorc( currow + 1, 0 );
    SLsmg_printf( "%13s %20s %20s %20s",
        "Load: ^A", "LoopStart: ^L   ", "Octave: +/-    ", "Start: F9   " );
    SLsmg_gotorc( currow + 2, 0 );
    SLsmg_printf( "%13s %20s %20s %20s",
        "Save: ^S", "LoopDest: ^K   ", " NoteVel: Hm/End ", "Rew+Start: F10  " );
    SLsmg_gotorc( currow + 3, 0 );
    SLsmg_printf( "%13s %20s %20s %20s",
        "Tempo: ^T", "ExtSync: ^E   ", "MidiCh: Ins/Del", "Stop: F11  " );
    SLsmg_gotorc( currow + 4, 0 );
    SLsmg_printf( "%13s %20s %20s %20s",
        "ChanName: ^N", "Copy: ^C   ", "ChanVol: *//    ", "Rewind: F12  " );

    SLsmg_gotorc( currow + 5, 0 );
    SLsmg_printf( "%13s %20s %20s %20s",
        " PatName: ^P", "Paste: ^V   ", "Pattern: [/]    ", "View: <tab>" );
}
//  function -- screen_printHelp
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_showStatusMessage
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_showStatusMessage(struct Screen *screen, char *newmsg)
{
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(newmsg);

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
    SLsmg_write_string(newmsg);
    SLsmg_gotorc(SLtt_Screen_Rows - 1, SLtt_Screen_Cols - 1);
    SLsmg_refresh();
}
//  function -- screen_showStatusMessage
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_moveCursorPosition
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_moveCursorPosition(struct Screen *screen, int newrow, int newcol)
{
	int i;
	int range;

    ERROR_NULL_VOID(screen);

    // Wrap up-down.
    if (screen->patmode)
    {
        if (newrow < 0)
            newrow = 0;
        if (newrow >= SONGPATLEN)
            newrow = SONGPATLEN - 1;
    }
    else
    {
        if (newrow < 0)
            newrow = 0;
        if (newrow >= SONGLEN)
            newrow = SONGLEN - 1;
    }

    // Wrap left-right.
    if (screen->patmode)
        newcol += pbar_getIndex(screen->patbars[0]);
    else
        newcol += sbar_getIndex(screen->songbars[0]);

    if (newcol < 0)
        newcol = 0;
    if (newcol >= SONGMAXCHAN)
        newcol = SONGMAXCHAN - 1;

    range = (newcol / numbars);
    for (i = 0; i < numbars; i++)
    {
		pbar_setChannel(screen->patbars[i], (numbars * range ) + i);
		sbar_setChannel(screen->songbars[i], (numbars * range ) + i);
    }
    newcol %= numbars;

    //
    //  Clear previous cursor hilighting
    //
    pbar_setCursor(screen->patbars[screen->curcol], -1);
	sbar_setCursor(screen->songbars[screen->curcol], -1);

    //
    //  Set the new cursor hilight
    //  Update the "scrolltop", the row number that should appear at the top
    //
    if (screen->patmode)
        screen->curstate.scrolltop = pbar_setCursor(screen->patbars[newcol], newrow);
    else
        screen->curstate.scrolltop = sbar_setCursor(screen->songbars[newcol], newrow);

    //
    //  Update the visible tracks
    //
    for (i = 0; i < numbars; i++)
    {
        if (screen->patmode)
		    pbar_setScrollTop(screen->patbars[i], screen->curstate.scrolltop);
        else
		    sbar_setScrollTop(screen->songbars[i], screen->curstate.scrolltop);
    }

    //
    //  Update the left-hand side of the screen with the row numbers
    //
	screen_printSidebar(screen);

	screen->curstate.cursorrow = newrow;
	screen->curcol = newcol;
    if (screen->patmode)
    {
        screen->patstate = screen->curstate;
    }
    else
    {
        screen->songstate = screen->curstate;
		screen_setPosition(screen, screen->songstate.cursorrow);
    }
    screen_printSidebar(screen);
}
//  function -- screen_moveCursorPosition
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_setPosition
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_setPosition(struct Screen *screen,  int newpos)
{
    int i;

    ERROR_NULL_VOID(screen);

    if (newpos < 0)
        newpos = 0;
    if (newpos > SONGLEN)
        newpos = SONGLEN;
    screen->pos = newpos;

    for (i = 0; i < numbars; i++)
    {
        pbar_setPosition(screen->patbars[i], screen->pos);
        if (screen->patmode)
            pbar_refresh(screen->patbars[i]);
    }
    screen->songstate.cursorrow = screen->pos;

    screen_refresh(screen);
}
//  function -- screen_setPosition
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_inputTempo
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_inputTempo(struct Screen *screen)
{
    ERROR_NULL_VOID(screen);

    inputBuffer("Enter Tempo: ");
    song_setTempo(screen->curr_song, atoi(inputbuf));
    screen_refresh(screen);
}
//  function -- screen_inputTempo
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_inputTempoRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_inputTempoRec(struct Screen *screen)
{
    int tempo;

    //
    //  "Rec" suffix for session recording
    //
    ERROR_NULL_VOID(screen);

    //
    //  Keystroke for input of Tempo will be recorded
    //

    inputBuffer("Enter Tempo: ");
    tempo = atoi(inputbuf);
    song_setTempo(screen->curr_song, tempo);

    if (screen->curr_song->re_record)
    {
        static int iParams[REVENT_MAX_IPARMS];

        ERROR_RANGE(0,REVENT_MAX_IPARMS,"(tempo in iParams)");
        iParams[0] = tempo;
        p_CurrRec = Attach_Revent(p_CurrRec,
                                  screen->curr_song->beatcount,
                                  REVENT_TEMPO,
                                  (int *)iParams);
        p_CurrRec->has_iParams = 1;
    }
    screen_refresh(screen);
}
//  function -- screen_inputTempoRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_inputTempoRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_inputTempoRE(struct Screen *screen)
{
    int tempo;

    //
    //  "RE" suffix for session playback
    //
    ERROR_NULL_VOID(screen);

    //
    //  REVENT will be used to change tempo
    //

    ERROR_RANGE(0,REVENT_MAX_IPARMS,"(tempo in iParams)");

    //
    //  If this function is called, we assume that the REVENT has already
    //  been check and found to be REVENT_TEMPO, and that re_play is on
    //
    tempo = p_CurrREvents->iParams[0];
    song_setTempo(screen->curr_song, tempo);

    screen_refresh(screen);
}
//  function -- screen_inputTempoRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_saveSong
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_saveSong(struct Screen *screen)
{
    ERROR_NULL_VOID(screen);

    inputBuffer("Enter Save Filename: ");
    if (inputbuf[0] == '\0')
    {
        screen_showStatusMessage(screen, "Save cancelled.");
        return;
    }

    if (!song_saveSong(screen->curr_song, inputbuf))
    {
        screen_showStatusMessage(screen, "Save failed.");
    }
	screen_refresh(screen);
    screen_showStatusMessage(screen, "Saved.");
}
//  function -- screen_saveSong
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_saveSongAscii
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_saveSongAscii(struct Screen *screen)
{
    ERROR_NULL_VOID(screen);

    inputBuffer("Enter Save Filename (Ascii): ");

    if (inputbuf[ 0 ] == '\0')
    {
        screen_showStatusMessage(screen, "Save cancelled.");
        return;
    }

    if (song_saveSongAscii(screen->curr_song, inputbuf) == M_FALSE)
    {
        screen_showStatusMessage(screen, "Save failed.");
    }
    screen_refresh(screen);
    screen_showStatusMessage(screen, "Saved.");
}
//  function -- screen_saveSongAscii
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_saveSongAsciiWrite
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_saveSongAsciiWrite(struct Screen *screen)
{
    struct Song *song;

    ERROR_NULL_VOID(screen);

    song = screen->curr_song;
    ERROR_NULL_VOID(song);

    //
    //  Same as  screen_saveSongAscii(), but do not prompt filename
    //  Just use the existing filename
    //
    if (strlen(song->filename) == 0)
    {
        screen_showStatusMessage(screen, "no filename -- aborted save.");
        return;
    }

    if (song_saveSongAscii(song, song->filename) == M_FALSE)
    {
        screen_showStatusMessage(screen, "Save failed.");
    }
    else
    {
        static char msg[MAX_PATH_LENGTH];
        sprintf(msg, "Saved %s", song->filename);
        screen_showStatusMessage(screen, msg);
    }

    screen_refresh(screen);
}
//  function -- screen_saveSongAsciiWrite
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_saveSongAscoreWrite
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_saveSongAscoreWrite(struct Screen *screen)
{
    struct Song *song;
    static char filename[MAX_FILENAME_LENGTH];
    char *pch;

    ERROR_NULL_VOID(screen);
    song = screen->curr_song;
    ERROR_NULL_VOID(song);

    //
    //  Same as  screen_saveSongAscii(), but do not prompt filename
    //  Just use the existing filename
    //
    if (strlen(song->filename) == 0)
    {
        screen_showStatusMessage(screen, "no filename -- aborted save.");
        return;
    }

    //
    //  Provide the ".asc"
    //
    strcpy(filename, song->filename);
    pch = strchr(filename, '.');
    if (pch)
    {
        *pch = 0x00;
    }
    strcat(filename, ".asc");

    if (song_saveSongAscore(song, filename) == M_FALSE)
    {
        screen_showStatusMessage(screen, "Save failed.");
    }
    else
    {
        static char msg[MAX_PATH_LENGTH];
        sprintf(msg, "Saved %s", filename);
        screen_showStatusMessage(screen, msg);
    }

    screen_refresh(screen);
}
//  function -- screen_saveSongAscoreWrite
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_loadSong
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_loadSong(struct Screen *screen)
{
    int i;
    struct Song *song;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    inputBuffer("Enter Load Filename: ");
    if (inputbuf[ 0 ] == '\0')
    {
        screen_showStatusMessage(screen, "Load cancelled.");
        return;
    }

    //
    //  Read-in the song file
    //
    if (song_loadSong(song, inputbuf))
        screen_showStatusMessage(screen, "Load failed.");

    //
    //  Refresh the screen
    //
    for (i = 0; i < numbars; i++)
    {
        if (screen->patmode)
            pbar_refresh(screen->patbars[i]);
        else
            sbar_refresh(screen->songbars[i]);
        screen_showStatusMessage(screen, "Song loaded.");
    }
    screen_refresh(screen);
}
//  function -- screen_loadSong
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_loadSongAscii
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_loadSongAscii(struct Screen *screen)
{
    int i;
    struct Song *song;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    inputBuffer("Enter Load Filename: ");
    if (inputbuf[ 0 ] == '\0')
    {
        screen_showStatusMessage(screen, "Load cancelled.");
        return;
    }

    //
    //  Read-in the song file
    //
    if (song_loadSongAscii(song, inputbuf))
        screen_showStatusMessage(screen, "Load failed.");

    //
    //  Refresh the screen
    //
    for (i = 0; i < numbars; i++)
    {
        if (screen->patmode)
            pbar_refresh(screen->patbars[i]);
        else
            sbar_refresh(screen->songbars[i]);
        screen_showStatusMessage(screen, "Song loaded.");
    }
    screen_refresh(screen);
}
//  function -- screen_loadSongAscii
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_enterPatName
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_enterPatName(struct Screen *screen)
{
    ERROR_NULL_VOID(screen);

    if (screen->patmode)
    {
        if (!pbar_patternSet(screen->patbars[screen->curcol]))
        {
            screen_showStatusMessage(screen,
                  "No pattern active on current channel.");
            return;
        }
    }
    else
    {
        if (!sbar_patternSet(screen->songbars[screen->curcol]))
        {
            screen_showStatusMessage(screen,
                  "No pattern active on current channel.");
            return;
        }
    }

    inputBuffer("Enter Pattern Name: ");
    // For now, trim all names to 8 characters.
    inputbuf[8] = '\0';
    if (screen->patmode)
        pbar_setPatternName(screen->patbars[screen->curcol], inputbuf);
    else
        sbar_setPatternName(screen->songbars[screen->curcol], inputbuf);
    screen_showStatusMessage(screen, "Name set.");
}
//  function -- screen_enterPatName
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_enterChanName
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_enterChanName(struct Screen *screen)
{
    static char newname[9];
    int j;

    ERROR_NULL_VOID(screen);

    //
    //  Save 8 characters of the input NAME exactly as entered
    //  to be displayed in the Channel Name field
    //
    newname[0] = 0x00;
    inputBuffer("Enter Channel Name: ");
    for (j=0; j<8; j++) newname[j] = inputbuf[j];
    newname[8] = 0x00;

    //
    //  Set the Channel Name field for the screen as typed
    //
    if (screen->patmode)
        pbar_setChannelName(screen->patbars[screen->curcol], newname);
    else
        sbar_setChannelName(screen->songbars[screen->curcol], newname);

    screen_showStatusMessage(screen, "Name set.");

}
//  function -- screen_enterChanName
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_toggleSync
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_toggleSync(struct Screen *screen)
{
    ERROR_NULL_VOID(screen);

    if (!mctrl_usingInternalClock(p_MidiControl))
    {
        mctrl_useInternalClock(p_MidiControl);
        screen_showStatusMessage(screen, "Using internal sync.");
    }
    else
    {
        mctrl_useExternalClock(p_MidiControl);
        screen_showStatusMessage(screen, "Using external sync.");
    }
}
//  function -- screen_toggleSync
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_saveCopy
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_saveCopy(struct Screen *screen)
{
    struct SongPattern *pat;
    struct SongChannel *schan;

    ERROR_NULL_VOID(screen);

    if (!screen->patmode)
        return;

    schan = song_getChannel(screen->curr_song,
                            pbar_getIndex(screen->patbars[screen->curcol]));
    if (schan == NULL)
    {
        Error_Printf("E204: screen_saveCopy: NULL SongChannel\n");
        return;
    }
    pat = schan_getSongPattern(schan, screen->pos);
    
    if (pat == NULL)
        return;

    spat_copyTo(pat, screen->copyPattern);
    screen_showStatusMessage(screen, "Copied.");
}
//  function -- screen_saveCopy
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_pasteCopy
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_pasteCopy(struct Screen *screen)
{
    struct SongChannel *chan;
    struct SongPattern *pat;

    ERROR_NULL_VOID(screen);

    if (!screen->patmode)
        return;

    chan = song_getChannel(screen->curr_song,
        pbar_getIndex(screen->patbars[screen->curcol]));
    if (chan == NULL)
    {
        Error_Printf("E205: screen_pasteCopy: NULL SongChannel\n");
        return;
    }

    pat = schan_getSongPattern(chan, screen->pos);

    if (pat == NULL)
    {
        int i;
        for (i = 0; i < SONGMAXPAT; i++)
        {
            if (spat_isBlank(schan_getChannelPattern(chan, i)))
            {
                schan_setSongPattern(chan, screen->pos, i + 1);
                pat = schan_getSongPattern(chan, screen->pos);
                break;
            }
        }
    }

    pat = schan_getSongPattern(chan, screen->pos);
    if (pat == NULL)
        return;

    spat_copyTo(screen->copyPattern, pat);

    pbar_refresh(screen->patbars[screen->curcol]);
    screen_showStatusMessage(screen, "Pasted.");
}
//  function -- screen_pasteCopy
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_setNote
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_setNote(struct Screen *screen, int newnote)
{
    ERROR_NULL_VOID(screen);

    if (screen->patmode)
    {
        pbar_setNote(screen->patbars[screen->curcol], newnote);
		screen_moveCursorPosition(screen, screen->patstate.cursorrow + 1,
                                  screen->curcol);
    }
}
//  function -- screen_setNote
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_toggleEndNote
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_toggleEndNote(struct Screen *screen)
{
    ERROR_NULL_VOID(screen);

    if (screen->patmode) 
        pbar_toggleEndNote(screen->patbars[screen->curcol]);
}
//  function -- screen_toggleEndNote
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  screen_editMidiPort
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_editMidiPort(struct Screen *screen)
{
    int port;
    struct Song *song;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    inputBuffer("Enter Port (1-8): ");
    if (inputbuf[0] == 0x00)
    {
        return;
    }
    port = atoi(inputbuf);
    if (port < 0)
        port = 0;

    if (screen->patmode)
        pbar_editMidiPort(screen->patbars[screen->curcol], port);
    else
        sbar_editMidiPort(screen->songbars[screen->curcol], port);

    screen_refresh(screen);
}
//  function -- screen_editMidiPort
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_editMidiPortRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_editMidiPortRec(struct Screen *screen)
{
    int port;
    struct Song *song;

    //
    //  "Rec" suffix for session recording
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  Keystroke for input of Midi Port  will be recorded
    //

    inputBuffer("Enter Port (1-8): ");
    if (inputbuf[0] == 0x00)
    {
        return;
    }
    port = atoi(inputbuf);
    if (port < 0)
        port = 0;

    if (screen->patmode)
        pbar_editMidiPort(screen->patbars[screen->curcol], port);
    else
        sbar_editMidiPort(screen->songbars[screen->curcol], port);

    if (screen->curr_song->re_record)
    {
        static int iParams[REVENT_MAX_IPARMS];

        ERROR_RANGE(0,REVENT_MAX_IPARMS,"(midi port in iParams)");
        iParams[0] = port;
        p_CurrRec = Attach_Revent(p_CurrRec,
                                  screen->curr_song->beatcount,
                                  REVENT_MIDI_PORT,
                                  (int *)iParams);
        p_CurrRec->has_iParams = 1;
    }
    screen_refresh(screen);
}
//  function -- screen_editMidiPortRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_editMidiPortRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_editMidiPortRE(struct Screen *screen)
{
    int port;
    struct Song *song;

    //
    //  "RE" suffix for session playback
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  REVENT will be used to change midi port
    //

    ERROR_RANGE(0,REVENT_MAX_IPARMS,"(midi port in iParams)");
    //
    //  If this function is called, we assume that the REVENT has already
    //  been check and found to be REVENT_MIDI_PORT, and that re_play is on
    //
    port = p_CurrREvents->iParams[0];
    if (port < 0)
        port = 0;
    if (screen->patmode)
        pbar_editMidiPort(screen->patbars[screen->curcol], port);
    else
        sbar_editMidiPort(screen->songbars[screen->curcol], port);

    screen_refresh(screen);
}
//  function -- screen_editMidiPortRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_editMidiPatch
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_editMidiPatch(struct Screen *screen)
{
    char *pch;
    int prognum=0, banknum=0;
    struct Song *song;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    inputBuffer("Enter Patch,Bank: ");
    if (inputbuf[0] == 0x00)
    {
        return;
    }

    pch = strtok(inputbuf, ",");
    if (pch)
    {
        prognum = atoi(pch);
        pch = strtok(NULL, ",");
        if (pch)
        {
            banknum = atoi(pch);
        }
    }

    if (screen->patmode)
        pbar_setChannelPatch(screen->patbars[screen->curcol],
                             banknum, prognum);
    else
        sbar_setChannelPatch(screen->songbars[screen->curcol],
                             banknum, prognum);

    screen_refresh(screen);
}
//  function -- screen_editMidiPatch
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_editMidiPatchRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_editMidiPatchRec(struct Screen *screen)
{
    char *pch;
    int prognum=0, banknum=0;
    struct Song *song;

    //
    //  "Rec" suffix for session recording
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    inputBuffer("Enter Patch,Bank: ");
    if (inputbuf[0] == 0x00)
    {
        return;
    }

    pch = strtok(inputbuf, ",");
    if (pch)
    {
        prognum = atoi(pch);
        pch = strtok(NULL, ",");
        if (pch)
        {
            banknum = atoi(pch);
        }
    }

    if (screen->patmode)
        pbar_setChannelPatch(screen->patbars[screen->curcol],
                             banknum, prognum);
    else
        sbar_setChannelPatch(screen->songbars[screen->curcol],
                             banknum, prognum);

    if (screen->curr_song->re_record)
    {
        static int iParams[REVENT_MAX_IPARMS];

        ERROR_RANGE(1,REVENT_MAX_IPARMS,"(midi patch in iParams)");
        iParams[0] = banknum;
        iParams[1] = prognum;
        p_CurrRec = Attach_Revent(p_CurrRec,
                                  screen->curr_song->beatcount,
                                  REVENT_MIDI_PATCH,
                                  (int *)iParams);
        p_CurrRec->has_iParams = 2;
    }

    screen_refresh(screen);
}
//  function -- screen_editMidiPatchRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_editMidiPatchRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_editMidiPatchRE(struct Screen *screen)
{
    int prognum=0, banknum=0;
    struct Song *song;

    //
    //  "RE" suffix for session playback
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  REVENT will be used to change tempo
    //

    ERROR_RANGE(1,REVENT_MAX_IPARMS,"(midi patch in iParams)");

    //
    //  If this function is called, we assume that the REVENT has already
    //  been check and found to be REVENT_MIDI_PATCH, and that re_play is on
    //
    banknum = p_CurrREvents->iParams[0];
    prognum = p_CurrREvents->iParams[1];

    if (screen->patmode)
        pbar_setChannelPatch(screen->patbars[screen->curcol],
                             banknum, prognum);
    else
        sbar_setChannelPatch(screen->songbars[screen->curcol],
                             banknum, prognum);
    screen_refresh(screen);
}
//  function -- screen_editMidiPatchRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_showMIDIinfo
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_showMIDIinfo(struct Screen *screen)
{
    struct Song *song;
    struct SongChannel *schan;
    struct PatBar *pbar;
    struct SongBar *sbar;
    int chan;
    static char msg[128];

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    if (screen->patmode)
    {
        pbar = screen->patbars[screen->curcol];
        chan = pbar->chan;
    }
    else
    {
        sbar = screen->songbars[screen->curcol];
        chan = sbar->chan;
    }

    schan = song_getChannel(song, chan);
    if (schan == NULL)
    {
        Error_Printf("E206: screen_showMIDIinfo: NULL SongChannel\n");
        return;
    }

    sprintf(msg, "patch %03d, bank %02d : channel %02d, port %d",
                 schan->prognum+1, schan->banknum+1,
                 schan->midichannel+1, schan->midiport+1);


    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(msg);
    SLsmg_refresh();
}
//  function -- screen_showMIDIinfo
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_pollRecord
//
//  Description
//
//      Performs playback actions from the session file.
//
//  Arguments
//
//      screen
//
//          The Screen struct.
//
//  Return
//
//      0 on success
//
//  -------------------------------------------------------------------------
int screen_pollRecord(struct Screen *screen)
{
    ERROR_NULL(screen,-1);
    ERROR_NULL(screen->curr_song,-1);

    //
    //  If playback is not turned on, return normally
    //
    if (p_CurrREvents == NULL || screen->curr_song->re_play == 0)
        return 0;

    //
    // This input call will throttle the main loop.
    //
    if (screen->curr_song->beatcount >= p_CurrREvents->index)
    {
        switch (p_CurrREvents->type)
        {

        case REVENT_QUIT:
            //  nop
        break;

        //
        // Channel mutes.
        //
        case REVENT_CHMUTE1:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[0]);
            else
                sbar_toggleChannelMute(screen->songbars[0]);
        break;
        case REVENT_CHMUTE2:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[1]);
            else
                sbar_toggleChannelMute(screen->songbars[1]);
        break;
        case REVENT_CHMUTE3:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[2]);
            else
                sbar_toggleChannelMute(screen->songbars[2]);
        break;
        case REVENT_CHMUTE4:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[3]);
            else
                sbar_toggleChannelMute(screen->songbars[3]);
        break;
        case REVENT_CHMUTE5:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[4]);
            else
                sbar_toggleChannelMute(screen->songbars[4]);
        break;
        case REVENT_CHMUTE6:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[5]);
            else
                sbar_toggleChannelMute(screen->songbars[5]);
        break;
        case REVENT_CHMUTE7:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[6]);
            else
                sbar_toggleChannelMute(screen->songbars[6]);
        break;
        case REVENT_CHMUTE8:
            if (screen->patmode)
                pbar_toggleChannelMute(screen->patbars[7]);
            else
                sbar_toggleChannelMute(screen->songbars[7]);
        break;

        //
        // Song Control.
        //
        case REVENT_SONG_START:
            // nop
        break;
        case REVENT_SONG_RWSTART:
            // nop
        break;
        case REVENT_SONG_STOP:
            // nop
        break;
        case REVENT_SONG_REWIND:
            song_rewind(screen->curr_song);
        break;

        case REVENT_KEY_DOWN:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow+1,
                              screen->curcol);
        break;
        case REVENT_KEY_UP:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow-1,
                              screen->curcol);
        break;
        case REVENT_KEY_LEFT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol-1);
        break;
        case REVENT_KEY_RIGHT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol+1);
        break;

        case REVENT_SHIFT_KEY_DOWN:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow+screen->scrsize,
                              screen->curcol);
        break;
        case REVENT_SHIFT_KEY_UP:
            screen_moveCursorPosition(screen,
                              screen->curstate.cursorrow-screen->scrsize,
                              screen->curcol);
        break;

        case REVENT_SHIFT_KEY_LEFT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol-numbars);
        break;
        case REVENT_SHIFT_KEY_RIGHT:
            screen_moveCursorPosition(screen, screen->curstate.cursorrow,
                              screen->curcol+numbars);
        break;

        case REVENT_MIDICHANNEL_UP:
            if (screen->patmode)
                pbar_editMidiChannel(screen->patbars[screen->curcol], 1);
            else
                sbar_editMidiChannel(screen->songbars[screen->curcol], 1);
        break;
        case REVENT_MIDICHANNEL_DOWN:
            if (screen->patmode)
                pbar_editMidiChannel(screen->patbars[screen->curcol], -1);
            else
                sbar_editMidiChannel(screen->songbars[screen->curcol], -1);
        break;

        case REVENT_VOLUME_DOWN:
            if (screen->patmode)
                pbar_editChannelVolume(screen->patbars[screen->curcol], -1);
            else
                sbar_editChannelVolume(screen->songbars[screen->curcol], -1);
        break;
        case REVENT_VOLUME_UP:
            if (screen->patmode)
                pbar_editChannelVolume(screen->patbars[screen->curcol], 1);
            else
                sbar_editChannelVolume(screen->songbars[screen->curcol], 1);
        break;

        case REVENT_VELOCITY_UP:
            if (screen->patmode)
            {
                pbar_editCursorVolume(screen->patbars[screen->curcol], 1);
            }
        break;
        case REVENT_VELOCITY_DOWN:
            if (screen->patmode)
            {
                pbar_editCursorVolume(screen->patbars[screen->curcol], -1);
            }
        break;

        case REVENT_PATTERN_DOWN:
            if (screen->patmode)
                pbar_editPattern(screen->patbars[screen->curcol], -1);
            else
                sbar_editPattern(screen->songbars[screen->curcol], -1);
        break;
        case REVENT_PATTERN_UP:
            if (screen->patmode)
                pbar_editPattern(screen->patbars[screen->curcol], 1);
            else
                sbar_editPattern(screen->songbars[screen->curcol], 1);
        break;

        case REVENT_OCTAVE_UP: ++screen->octave; break;
        case REVENT_OCTAVE_DOWN: --screen->octave; break;

        case REVENT_SEMITONE_UP:
            if (screen->patmode)
                pbar_transposeNotes(screen->patbars[screen->curcol], -1);
        break;
        case REVENT_SEMITONE_DOWN:
            if (screen->patmode)
                pbar_transposeNotes(screen->patbars[screen->curcol], 1);
        break;

        case REVENT_SONGPOSITION_UP:
            if (screen->patmode)
            {
                screen_setPosition(screen, screen->pos-1);
            }
            else
            {
                screen_moveCursorPosition(screen, screen->curstate.cursorrow-8,
                                          screen->curcol );
            }
            break;
        case REVENT_SONGPOSITION_DOWN:
            if (screen->patmode)
            {
                screen_setPosition(screen, screen->pos+1);
            }
            else
            {
                screen_moveCursorPosition(screen, screen->curstate.cursorrow+8,
                                          screen->curcol );
            }
            break;

        case REVENT_TOGGLE_SLIDE:
            if (screen->patmode)
            {
                pbar_toggleCursorSlide(screen->patbars[screen->curcol]);
                screen_moveCursorPosition(screen, screen->patstate.cursorrow+1,
                                          screen->curcol);
            }
            break;

        case REVENT_NOTE_EMPTY:
            if (screen->patmode)
            {
                screen_setNote(screen, SONG_NOTEEMPTY);
            }
            else
            {
                sbar_editPattern(screen->songbars[screen->curcol], 0-(SONGMAXPAT+1));
                screen_moveCursorPosition(screen, screen->curstate.cursorrow+1,
                                          screen->curcol);
            }
            break;

        case REVENT_NOTE_C__LO:
            screen_setNote(screen, (screen->octave+1)*12+0);
        break;
        case REVENT_NOTE_CS_LO:
            screen_setNote(screen, (screen->octave+1)*12+1);
        break;
        case REVENT_NOTE_D__LO:
            screen_setNote(screen, (screen->octave+1)*12+2);
        break;
        case REVENT_NOTE_DS_LO:
            screen_setNote(screen, (screen->octave+1)*12+3);
        break;
        case REVENT_NOTE_E__LO:
            screen_setNote(screen, (screen->octave+1)*12+4);
        break;
        case REVENT_NOTE_F__LO:
            screen_setNote(screen, (screen->octave+1)*12+5);
        break;
        case REVENT_NOTE_FS_LO:
            screen_setNote(screen, (screen->octave+1)*12+6);
        break;
        case REVENT_NOTE_G__LO:
            screen_setNote(screen, (screen->octave+1)*12+7);
        break;
        case REVENT_NOTE_GS_LO:
            screen_setNote(screen, (screen->octave+1)*12+8);
        break;
        case REVENT_NOTE_A__LO:
            screen_setNote(screen, (screen->octave+1)*12+9);
        break;
        case REVENT_NOTE_AS_LO:
            screen_setNote(screen, (screen->octave+1)*12+10);
        break;
        case REVENT_NOTE_B__LO:
            screen_setNote(screen, (screen->octave+1)*12+11);
        break;

        case REVENT_NOTE_C__HI:
            screen_setNote(screen, (screen->octave+1)*12+12);
        break;
        case REVENT_NOTE_CS_HI:
            screen_setNote(screen, (screen->octave+1)*12+13);
        break;
        case REVENT_NOTE_D__HI:
            screen_setNote(screen, (screen->octave+1)*12+14);
        break;
        case REVENT_NOTE_DS_HI:
            screen_setNote(screen, (screen->octave+1)*12+15);
        break;
        case REVENT_NOTE_E__HI:
            screen_setNote(screen, (screen->octave+1)*12+16);
        break;
        case REVENT_NOTE_F__HI:
            screen_setNote(screen, (screen->octave+1)*12+17);
        break;
        case REVENT_NOTE_FS_HI:
            screen_setNote(screen, (screen->octave+1)*12+18);
        break;
        case REVENT_NOTE_G__HI:
            screen_setNote(screen, (screen->octave+1)*12+19);
        break;
        case REVENT_NOTE_GS_HI:
            screen_setNote(screen, (screen->octave+1)*12+20);
        break;
        case REVENT_NOTE_A__HI:
            screen_setNote(screen, (screen->octave+1)*12+21);
        break;
        case REVENT_NOTE_AS_HI:
            screen_setNote(screen, (screen->octave+1)*12+22);
        break;
        case REVENT_NOTE_B__HI:
            screen_setNote(screen, (screen->octave+1)*12+23);
        break;

        case REVENT_NOTE_C__HX:
            screen_setNote(screen, (screen->octave+1)*12+24);
        break;
        case REVENT_NOTE_CS_HX:
            screen_setNote(screen, (screen->octave+1)*12+25);
        break;
        case REVENT_NOTE_D__HX:
            screen_setNote(screen, (screen->octave+1)*12+26);
        break;
        case REVENT_NOTE_DS_HX:
            screen_setNote(screen, (screen->octave+1)*12+27);
        break;
        case REVENT_NOTE_E__HX:
            screen_setNote(screen, (screen->octave+1)*12+28);
        break;

        case REVENT_NOTE_OFF:
            screen_setNote(screen, SONG_NOTEOFF);
        break;

        case REVENT_TEMPO:
            screen_inputTempoRE(screen);
        break;
        case REVENT_LOOP_POSITION:
            song_setLoopPosition(screen->curr_song, screen->pos+1);
        break;
        case REVENT_LOOP_DESTINATION:
            song_setLoopDestination(screen->curr_song, screen->pos);
        break;
        case REVENT_START_POSITION:
             song_setStartPos(screen->curr_song, screen->pos);
        break;
        case REVENT_SAVE_SONG :
            // nop
        break;
        case REVENT_LOAD_SONG:
            // nop
        break;
        case REVENT_TOGGLE_SYNC:
             screen_toggleSync(screen);
        break;
        case REVENT_SAVE_COPY:
            // nop
        break;
        case REVENT_PASTE_COPY:
            // nop
        break;
        case REVENT_CHANNEL_NAME:
            // nop
        break;
        case REVENT_PATTERN_NAME:
            // nop
        break;
        case REVENT_TOGGLE_ENDNOTE:
            screen_toggleEndNote(screen);
        break;
        case REVENT_MIDI_PORT:
            screen_editMidiPortRE(screen);
        break;
        case REVENT_MIDI_PATCH:
            screen_editMidiPatchRE(screen);
        break;
        case REVENT_LOOP_ON:
            song_toggleLoop(screen->curr_song);
        break;
        case REVENT_SWITCH_MODE:
            screen_switchMode(screen);
        break;
        case REVENT_SHOW_INFO:
            // nop
        break;
        case REVENT_EXT_LOOP_POS:
            //
            //  extended session command
            //  set loop position
            //
            screen_setLoopPositionRE(screen);
        break;
        case REVENT_EXT_LOOP_DEST:
            //
            //  extended session command
            //  set loop position
            //
            screen_setLoopDestinationRE(screen);
        break;
        case REVENT_EXT_MIDI_FLUSH:
            //
            //  extended session command
            //  flush all midi events to the ports
            //
            screen_extCmdMidiFlushRE(screen);
        break;
        case REVENT_EXT_MIDI_CONTROL:
            //
            //  extended session command
            //  send a midi control message to a device
            //
            screen_extCmdMidiControlRE(screen);
        break;
        case REVENT_EXT_VELOCITY_RESET:
            //
            //  extended session command
            //  velocity reset for channel,pattern,beat
            //
            screen_extCmdVelocityRsRE(screen);
        break;
        case REVENT_EXT_VELOCITY_CHANGE:
            //
            //  extended session command
            //  velocity change for velocity,channel,pattern,beat
            //
            screen_extCmdVelocityRE(screen);
        break;
        case REVENT_EXT_PATTERN_SET:
            //
            //  extended session command
            //  set a pattern number in song mode
            //
            screen_extCmdPatternSetRE(screen);
        break;
        default:
            // nop
        break;
        }

        p_CurrREvents = p_CurrREvents->next;
    }

    screen_refresh(screen);
    return 0;   // normal
}
//  function -- screen_pollRecord
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_loadSession
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_loadSession(struct Screen *screen)
{
    struct Song *song;
    struct RecordEvent *prec;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    inputBuffer("Enter Session Filename: ");
    if (inputbuf[ 0 ] == '\0')
    {
        screen_showStatusMessage(screen, "Load cancelled.");
        return;
    }

    //
    //  Read-in the session file
    //
    if (p_REvents)
    {
        //
        //  Delete an existing session
        //
        Delete_RecordEvent(p_REvents);
        p_REvents = p_CurrRec = NULL;
    }

    prec = Read_Events_File(inputbuf);
    if (prec)
    {
        p_REvents = p_CurrREvents = prec;
        song->re_play = 1;
        Msg_Printf("Session Play file %s ON\n", inputbuf);
    }
    else
        Error_Printf("E207: cannot open session (%s)\n", inputbuf);

    screen_refresh(screen);
}
//  function -- screen_loadSession
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_setLoopPositionRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_setLoopPositionRE(struct Screen *screen)
{
    int pos = 0;
    struct Song *song;

    //
    //  "RE" suffix for session playback
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    //
    //  REVENT will be used to change tempo
    //

    ERROR_RANGE(0,REVENT_MAX_IPARMS,"(loop Position in iParams)");

    //
    //  If this function is called, we assume that the REVENT has already
    //  been check and found to be REVENT_MIDI_PATCH, and that re_play is on
    //
    pos = p_CurrREvents->iParams[0];
    song_setLoopPosition(screen->curr_song, pos+1);

    screen_refresh(screen);
}
//  function -- screen_setLoopPositionRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_setLoopDestinationRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_setLoopDestinationRE(struct Screen *screen)
{
    int pos = 0;
    struct Song *song;

    //
    //  "RE" suffix for session playback
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    //
    //  REVENT will be used to change tempo
    //

    ERROR_RANGE(0,REVENT_MAX_IPARMS,"(loop Destination in iParams)");

    //
    //  If this function is called, we assume that the REVENT has already
    //  been check and found to be REVENT_MIDI_PATCH, and that re_play is on
    //
    pos = p_CurrREvents->iParams[0];
    song_setLoopDestination(screen->curr_song, pos);

    screen_refresh(screen);
}
//  function -- screen_setLoopDestinationRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_getPatBarWithChannel
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct PatBar *screen_getPatBarWithChannel(struct Screen *screen, int ichan)
{
    struct PatBar *pbar;
    int i, iichan;

    ERROR_NULL(screen,0);

    for (i=0; i<NUM_BARS; i++)
    {
        pbar = screen->patbars[i];
        iichan = pbar_getIndex(pbar);
        if (iichan == ichan)
            return pbar;    // success
    }

    return NULL;    // failure
}
//  function -- screen_getPatBarWithChannel
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_getSongBarWithChannel
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct SongBar *screen_getSongBarWithChannel(struct Screen *screen, int ichan)
{
    struct SongBar *sbar;
    int i, iichan;

    ERROR_NULL(screen,0);

    for (i=0; i<NUM_BARS; i++)
    {
        sbar = screen->songbars[i];
        iichan = sbar_getIndex(sbar);
        if (iichan == ichan)
            return sbar;    // success
    }

    return NULL;    // failure
}
//  function -- screen_getSongBarWithChannel
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  exitWaitmsg
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_exitWaitmsg(struct Screen *screen, int count)
{
    static char msg[128];
    int i;

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    msg[0] = 0x00;
    for (i=0; i < count; i++)
    {
        strcat(msg, ".");
    }
    SLsmg_write_string(msg);
    SLsmg_refresh();
}
//  function -- screen_exitWaitmsg
//  -------------------------------------------------------------------------



//  -------------------------------------------------------------------------
//  Function    --  screen_extCommandRec
//
//  Description
//
//      The entry point for all extended commands.
//
//  Arguments
//
//      screen
//
//          Pointer to a Screen struct instance.  There is actually only one
//          screen struct at a time, so this is THE screen.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCommandRec(struct Screen *screen)
{
    struct Song *song;
    char *p1, *p2, *p3, *p4, *p5;
    int mindex = 0, newrow = 0, newcol = 0;
    int err, timeout_seconds=30, portno=2999, silence=1;
    static char msg[32];

    //
    //  "Rec" suffix for session recording
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  Keystroke for input of Midi Port  will be recorded
    //

    inputBuffer("Enter command: ");
    if (inputbuf[0] == 0x00)
    {
        return;
    }

    //
    //  Parse the command
    //
    p1 = strtok(inputbuf, " ");
    if (p1 == NULL)
        return;

    switch (*p1)
    {
        case 'm':
            //
            //  midi command
            //
            p2 = strtok(NULL, " ");
            if (p2 == NULL)
            {
                //
                //  no arguments -- just flush all MIDI
                //
                screen_extCmdMidiFlushRec(screen);
                goto Refresh;
            }
            switch (*p2)
            {
                int n1, n2;

                case 'c':
                    //
                    //  control message
                    //
                    p3 = strtok(NULL, " ");
                    if (p3 == NULL)
                        return;
                    n1 = atoi(p3);
                    p4 = strtok(NULL, " ");
                    if (p4 == NULL)
                        return;
                    n2 = atoi(p4);
                    screen_extCmdMidiControlRec(screen, n1, n2);
                    goto Refresh;
                break;

                case 'r':
                    //
                    //  refresh the MIDI setup messages
                    //
                    song_refreshMidiSetup(screen->curr_song);
                    goto Refresh;
                break;
            }
        break;

        case 'a':
            //
            //  attack (velocity) for the current note
            //
            p2 = strtok(NULL, " ");
            if (p2 == NULL)
            {
                //
                //  no arguments -- just reset the velocity
                //
                screen_extCmdVelocityRsRec(screen);
                goto Refresh;
            }
            else
            {
                int velocity;

                velocity = atoi(p2);
                screen_extCmdVelocityRec(screen, velocity);
                goto Refresh;
            }
        break;

        case 'w':
            //
            //  check for arguments
            //
            p2 = strtok(NULL, " ");
            if (p2 == NULL)
            {
                //
                //  save the current song (without having to type the file name)
                //
                screen_extCmdWriteRec(screen);
            }
            else
            {
                if (*p2 == 'a')
                {
                    //
                    //  Save as an ASCORE score file
                    //
                    screen_extCmdWriteARec(screen);
                }
            }
            goto Refresh;
        break;

        case 'e':
            //
            //  "edit"
            //
            p2 = strtok(NULL, " ");
            if (p2)
            {
                switch (*p2)
                {
                    case 'm':
                        //
                        //  mark the upper, left of a song block
                        //
                        screen_extCmdEdMarkMRec(screen);
                        goto Refresh;
                    break;

                    case 'n':
                        //
                        //  mark the lower, right of a song block
                        //
                        screen_extCmdEdMarkNRec(screen);
                        goto Refresh;
                    break;

                    case 'z':
                        //
                        //  clear song block marks
                        //
                        screen_extCmdEdMarkClearRec(screen);
                        goto Refresh;
                    break;

                    case 'c':
                        //
                        //  copy a song block
                        //
                        screen_extCmdEdCopyRec(screen);
                        goto Refresh;
                    break;

                    case 'x':
                        //
                        //  cut a song block
                        //
                        screen_extCmdEdCutRec(screen);
                        goto Refresh;
                    break;

                    case 'v':
                        //
                        //  paste a song block
                        //
                        screen_extCmdEdPasteRec(screen);
                        goto Refresh;
                    break;

                    case 'd':
                        //
                        //  delete a song block
                        //
                        screen_extCmdEdDelRec(screen);
                        goto Refresh;
                    break;

                    case 'y':
                        //
                        //  copy (Yank) a song line
                        //
                        screen_extCmdEdYankRec(screen);
                        goto Refresh;
                    break;

                }
            }
            goto Refresh;
        break;

        case 's':
            //
            //  Session file command
            //
            p2 = strtok(NULL, " ");
            if (p2)
            {
                switch (*p2)
                {
                    case 'r':
                        //
                        //  Record session
                        //
                        p3 = strtok(NULL, " ");
                        if (p3)
                        {
                            if (atoi(p3) > 0)
                            {
                                //
                                //  Turn ON recording
                                //
                                g_Record_Session = 1;
                                song->re_record = 1;
                                Msg_Printf("Session Record ON\n");
                            }
                            else
                            {
                                //
                                //  Turn OFF recording
                                //
                                g_Record_Session = 0;
                                song->re_record = 0;
                                Msg_Printf("Session Record OFF\n");
                            }
                        }
                    break;

                    case 'p':
                        //
                        //  Playback session
                        //  ":s p <filename>"
                        //
                        p3 = strtok(NULL, " ");
                        if (p3)
                        {
                            //
                            //  Turn ON playback
                            //
                            if (p3[0] == 0x31 || p3[0] == 0x30)
                            {
                                screen_extCmdPlayOnOff(screen, p3);
                            }
                            else if (p3[0] == 0x72)
                            {
                                //
                                //  'r' is for reset to the beginning
                                //  Put the "curr" pointer back on the head of
                                //  the chain.
                                //
                                p_CurrREvents = p_REvents;
                                screen->curr_song->beatcount = 0;
                            }
                            else
                            {
                                //
                                //  p3==filename
                                //
                                screen_extCmdLoadSession(screen, p3);
                            }
                        }
                        else
                        {
                            //
                            //  ":s p"
                            //  with no argument after the p, puts the playback
                            //  in standby mode
                            //
                            screen_extCmdTogglePlay(screen);
                        }
                    break;
                }
            }
            goto Refresh;
        break;

        case 'p':
            //
            //  Pattern number entry, in song mode
            //
            p2 = strtok(NULL, " ");
            if (p2)
            {
                //
                //  p2 should be a pattern number
                //
                screen_extCmdPatternSetRec(screen, atoi(p2));
                goto Refresh;
            }
        break;

        case 'k':
            //
            //  set -- song mode cursor marker
            //
            p2 = strtok(NULL, " ");
            if (p2 == NULL)
            {
                //
                //  no arguments -- abort
                //
                goto Refresh;
            }
            mindex = tolower(*p2) -97;
            if (mindex < 0 || mindex > 26)
                goto Refresh;

            //
            //  Save current position in marker array
            //
            screen->kmark[mindex].row = screen->curstate.cursorrow;
            screen->kmark[mindex].col = screen->curcol;
            sprintf(msg, "set mark %c", tolower(*p2));
            screen_showStatusMessage(screen, msg);
            goto Refresh;
        break;

        case 'n':
            //
            //  moveto -- song mode cursor marker
            //
            p2 = strtok(NULL, " ");
            if (p2 == NULL)
            {
                //
                //  no arguments -- abort
                //
                goto Refresh;
            }
            mindex = tolower(*p2) - 97;
            if (mindex < 0 || mindex > 26)
                goto Refresh;

            //
            //  Get the position in marker array
            //
            newrow = screen->kmark[mindex].row;
            newcol = screen->kmark[mindex].col;
            //
            //  Move to
            //
            screen_moveCursorPosition(screen, newrow, newcol);
            goto Refresh;
        break;

        case 'c':
            //
            //  general command
            //
            p2 = strtok(NULL, " ");
            if (p2 == NULL)
            {
                //
                //  no arguments -- abort
                //
                goto Refresh;
            }
            switch (*p2)
            {
                int n1, n2, n3;

                case 'b':
                    //
                    //  Only allow this command when stopped.
                    //
                    if (song->playing == M_TRUE)
                        goto Refresh;
                    //
                    //  "start" while waiting for sync message on netcat
                    //
                    screen_showStatusMessage(screen, "Waiting for start signal (nc)");
                    //
                    //  blocking call that waits for signal from nc
                    //  Shell script waitnc.sh should have the netcat
                    //  command
                    //      nc -l -p 2998 -w 20
                    //  The -w (for timeout) is a safety in case the op
                    //  needs to be aborted.
                    //
                    err = system("sh waitnc.sh > /dev/null 2> /dev/null");
                    if (err == 0)
                    {
                        screen_showStatusMessage(screen, "Song started from signal");
                        song_start(song);
                    }
                    else
                    {
                        screen_showStatusMessage(screen, "Start signal error: waitnc.sh not found");
                    }
                break;

                case 'c':
                    //
                    //  Only allow this command when stopped.
                    //
                    if (song->playing == M_TRUE)
                        goto Refresh;

                    p3 = strtok(NULL, " ");
                    if (p3)
                    {
                        // User may set timeout on command-line: ":c c <port>"
                        // Default is 30 seconds
                        n1 = atoi(p3);
                        if (n1 > 0)
                        {
                            portno = n1;
                        }
                        p4 = strtok(NULL, " ");
                        if (p4)
                        {
                            // User may set timeout on command-line:
                            //       ":c c <port> <seconds>"
                            // Default is 30 seconds
                            n2 = atoi(p4);
                            if (n2 > 0)
                            {
                                timeout_seconds = n2;
                            }
                            p5 = strtok(NULL, " ");
                            if (p5)
                            {
                                // User may set silence (seconds) before start
                                // on command-line:
                                //       ":c c <port> <timeout> <silence>"
                                // Default is 1 seconds (0 is valid)
                                n3 = atoi(p5);
                                if (n3 >= 0)
                                {
                                    silence = n3;
                                }
                            } // p5
                        } // p4
                    } // p3

                    LogFile_Printf("INFO: socket signal: port %d, timeout %d,"
                                   " silence %d\n", portno, timeout_seconds,
                                   silence);

                    //
                    //  "start" while waiting for sync message on netcat
                    //
                    screen_showStatusMessage(screen, "Waiting for start signal (sock)");
                    //
                    //  blocking call that waits for signal from nc
                    //
                    err = screen_WaitNc(portno, timeout_seconds);
                    if (err == 0)
                    {
                        screen_showStatusMessage(screen, "Song started from signal");
                        if (silence > 0)
                            sleep(silence);
                        song_start(song);
                    }
                    else
                    {
                        strcpy(msg, "Start signal error: ");
                        switch (err)
                        {
                            case -1:
                                strcat(msg, "socket failed: see log file");
                            break;

                            case -2:
                                strcat(msg, "bind failed: see log file");
                            break;

                            case -3:
                                strcat(msg, "accept failed: see log file");
                            break;

                            default:
                                strcat(msg, "unknown error: see log file");
                            break;
                        }
                        screen_showStatusMessage(screen, msg);
                    }
                break;

                default:
                    goto Refresh;
            }
        break;

        default:
            ;
    }
    // switch

Refresh:
    screen_refresh(screen);
}
//  function -- screen_extCommandRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdMidiFlushRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdMidiFlushRec(struct Screen *screen)
{
    int ichan;
    struct Song *song;

    //
    //  Helper for the screen_extCommandRec function
    //  Flush all MIDI messages to the ports
    //  Record to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    //
    //  Get which channel we are on
    //
    if (screen->patmode)
        ichan = screen->patbars[screen->curcol]->chan;
    else
        ichan = screen->songbars[screen->curcol]->chan;

    //
    //  no arguments -- just flush all MIDI
    //
    mctrl_doCurrentMessages(p_MidiControl);
    //
    //  record event
    //
    if (song->re_record)
    {
        p_CurrRec = Attach_Revent(p_CurrRec, song->beatcount,
                                  REVENT_EXT_MIDI_FLUSH, NULL);
        p_CurrRec->has_iParams = 0;
    }

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdMidiFlushRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdMidiFlushRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdMidiFlushRE(struct Screen *screen)
{
    //
    //  Flush all MIDI messages to the ports
    //  Playback session
    //
    ERROR_NULL_VOID(screen);

    //
    //  no arguments -- just flush all MIDI
    //
    mctrl_doCurrentMessages(p_MidiControl);

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdMidiFlushRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdMidiControlRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdMidiControlRec(struct Screen *screen, int n1, int n2)
{
    int ichan;
    struct Song *song;
    static int iParams[REVENT_MAX_IPARMS];

    //
    //  Helper for the screen_extCommandRec function
    //  Send an arbitrary MIDI Control message
    //  Record to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    //
    //  Get which channel we are on
    //
    if (screen->patmode)
        ichan = screen->patbars[screen->curcol]->chan;
    else
        ichan = screen->songbars[screen->curcol]->chan;

    //
    //  If this is equivalent to a volume change, just use the ordinary
    //  volume change control--that way we stay in sync on the display
    //
    if (n1 == 7)
    {
        struct SongChannel *schan;

        //
        //  Get pointer to the referenced channel
        //
        schan = song_getChannel(song, ichan);
        if (schan == NULL)
        {
            Error_Printf("E208: screen_extCmdMidiControlRec: NULL channel\n");
            return;
        }
        //
        //  Actually set the volume
        //
        schan_setVolume(schan, n2);

        //
        //  update the screen
        //
        if (screen->patmode)
            pbar_refresh(screen->patbars[screen->curcol]);
        else
            sbar_refresh(screen->songbars[screen->curcol]);
    }
    else
    {
        //
        //  Perform the MIDI action
        //
        mctrl_doControlChange(p_MidiControl, ichan, n1, n2);
    }

    //
    //  record the action in a session tag
    //
    if (song->re_record)
    {
        ERROR_RANGE(2,REVENT_MAX_IPARMS, "(midi ctrl in iParams)");
        iParams[0] = n1;
        iParams[1] = n2;
        iParams[2] = ichan;
        p_CurrRec = Attach_Revent(p_CurrRec, song->beatcount,
                                  REVENT_EXT_MIDI_CONTROL, iParams);
        p_CurrRec->has_iParams = 3;
    }


    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdMidiControlRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdMidiControlRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdMidiControlRE(struct Screen *screen)
{
    int n1, n2, ichan;
    struct Song *song;

    //
    //  Helper for the screen_extCommandRE function
    //  Send an arbitrary MIDI Control message
    //  Record to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;


    ERROR_RANGE(2,REVENT_MAX_IPARMS, "(midi ctrl in iParams)");
    n1 = p_CurrREvents->iParams[0];
    n2 = p_CurrREvents->iParams[1];
    ichan = p_CurrREvents->iParams[2];

    //
    //  If this is equivalent to a volume change, just use the ordinary
    //  volume change control--that way we stay in sync on the display
    //
    if (n1 == 7)
    {
        struct SongChannel *schan;

        //
        //  Get pointer to the referenced channel
        //
        schan = song_getChannel(song, ichan);
        if (schan == NULL)
        {
            Error_Printf("E209: screen_extCmdMidiControlRE: NULL channel\n");
            return;
        }

        //
        //  Actually set the volume with the value read from the file
        //  n2 is expected to be a number in range 1--128
        //
        schan_setVolume(schan, n2);

        //
        //  update the screen
        //  The current column may not correspond to the given channel
        //
        if (screen->patmode)
        {
            struct PatBar *pbar;

            pbar = screen_getPatBarWithChannel(screen, ichan);
            if (pbar)
                pbar_refresh(pbar);
        }
        else
        {
            struct SongBar *sbar;

            sbar = screen_getSongBarWithChannel(screen, ichan);
            if (sbar)
                sbar_refresh(sbar);
        }
    }
    else
    {
        //
        //  Perform the MIDI action
        //
        mctrl_doControlChange(p_MidiControl, ichan, n1, n2);
    }

    screen_refresh(screen);
}
//  function -- screen_extCmdMidiControlRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdVelocityRsRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdVelocityRsRec(struct Screen *screen)
{
    int ichan=0, ipat=0, ibeat=0;
    struct Song *song;
    struct SongChannel *schan=NULL;
    struct PatBar *pbar=NULL;
    struct SongPattern *pat=NULL;
    static int iParams[REVENT_MAX_IPARMS];

    //
    //  Helper for the screen_extCommandRec function
    //  Send a velocity change on the current note
    //  Record to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    //
    //  This can only be done in pattern mode (from the keyboard)
    //  because the vel change corresponds to the note under the
    //  cursor.
    //
    if (!screen->patmode)
        return;

    //
    //  Get which channel we are on
    //
    if (screen->patmode)
        ichan = screen->patbars[screen->curcol]->chan;
    else
        ichan = screen->songbars[screen->curcol]->chan;

    schan = song_getChannel(song, ichan);
    if (schan == NULL)
    {
        Error_Printf("E210: screen_extCmdVelocityRsRec: NULL SongChannel\n");
        return;
    }

    //
    //  This can only be done in pattern mode (from the keyboard)
    //  because the vel change corresponds to the note under the
    //  cursor.
    //
    if (screen->patmode)
    {
        //
        //  Execute the command
        //
        pbar = screen->patbars[screen->curcol];
        pbar_setCursorVolume(pbar, DEFAULT_VELOCITY);

        //
        //  Retreive info for the session file
        //
        pat = schan_getSongPattern(schan, pbar->pos);
        ipat = spat_getIndex(pat);
        ibeat = pbar->cursor;
    }

    //
    //  record the action in a session tag
    //
    if (song->re_record)
    {
        ERROR_RANGE(2,REVENT_MAX_IPARMS, "(vel reset in iParams)");
        iParams[0] = ichan;
        iParams[1] = ipat;
        iParams[2] = ibeat;
        p_CurrRec = Attach_Revent(p_CurrRec, song->beatcount,
                                  REVENT_EXT_VELOCITY_RESET, iParams);
        p_CurrRec->has_iParams = 3;
    }

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdVelocityRsRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdVelocityRsRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdVelocityRsRE(struct Screen *screen)
{
    int ichan=0, ipat=0, ibeat=0;
    struct Song *song;
    struct SongChannel *schan=NULL;
    struct SongPattern *pat=NULL;

    //
    //  Send a velocity reset for chan,pat,beat
    //  Playback to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    ERROR_RANGE(2,REVENT_MAX_IPARMS, "(vel reset in iParams)");
    ichan = p_CurrREvents->iParams[0];
    ipat = p_CurrREvents->iParams[1];
    ibeat = p_CurrREvents->iParams[2];

    schan = song_getChannel(song, ichan);
    if (schan == NULL)
    {
        Error_Printf("E211: screen_extCmdVelocityRsRE: NULL SongChannel\n");
        return;
    }

    pat = schan->patterns[ipat-1];
    if (pat)
        spat_setVolume(pat, ibeat, DEFAULT_VELOCITY);

}
//  function -- screen_extCmdVelocityRsRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdVelocityRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdVelocityRec(struct Screen *screen, int velocity)
{
    int ichan=0, ipat=0, ibeat=0;
    struct Song *song;
    struct SongChannel *schan=NULL;
    struct PatBar *pbar=NULL;
    struct SongPattern *pat=NULL;
    static int iParams[REVENT_MAX_IPARMS];

    //
    //  Helper for the screen_extCommandRec function
    //  Send a velocity change on the current note
    //  Record to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    //
    //  This can only be done in pattern mode (from the keyboard)
    //  because the vel change corresponds to the note under the
    //  cursor.
    //
    if (!screen->patmode)
        return;

    //
    //  Get which channel we are on
    //
    if (screen->patmode)
        ichan = screen->patbars[screen->curcol]->chan;
    else
        ichan = screen->songbars[screen->curcol]->chan;

    schan = song_getChannel(song, ichan);
    if (schan == NULL)
    {
        Error_Printf("E212: screen_extCmdVelocityRec: NULL SongChannel\n");
        return;
    }

    //
    //  This can only be done in pattern mode (from the keyboard)
    //  because the vel change corresponds to the note under the
    //  cursor.
    //
    if (screen->patmode)
    {
        //
        //  Execute the command
        //
        pbar = screen->patbars[screen->curcol];
        pbar_setCursorVolume(pbar, velocity);

        //
        //  Retreive info for the session file
        //
        pat = schan_getSongPattern(schan, pbar->pos);
        ipat = spat_getIndex(pat);
        ibeat = pbar->cursor;
    }

    //
    //  record the action in a session tag
    //
    if (song->re_record)
    {
        ERROR_RANGE(3,REVENT_MAX_IPARMS, "(vel change in iParams)");
        iParams[0] = velocity;
        iParams[1] = ichan;
        iParams[2] = ipat;
        iParams[3] = ibeat;
        p_CurrRec = Attach_Revent(p_CurrRec, song->beatcount,
                                  REVENT_EXT_VELOCITY_CHANGE, iParams);
        p_CurrRec->has_iParams = 4;
    }

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdVelocityRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdVelocityRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdVelocityRE(struct Screen *screen)
{
    int vel=0, ichan=0, ipat=0, ibeat=0;
    struct Song *song;
    struct SongChannel *schan=NULL;
    struct SongPattern *pat=NULL;

    //
    //  Send a velocity change for vel,chan,pat,beat
    //  Playback to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    ERROR_RANGE(3,REVENT_MAX_IPARMS, "(vel chg in iParams)");
    vel = p_CurrREvents->iParams[0];
    ichan = p_CurrREvents->iParams[1];
    ipat = p_CurrREvents->iParams[2];
    ibeat = p_CurrREvents->iParams[3];

    schan = song_getChannel(song, ichan);
    if (schan == NULL)
    {
        Error_Printf("E213: screen_extCmdVelocityRE: NULL SongChannel\n");
        return;
    }

    pat = schan->patterns[ipat-1];
    if (pat)
        spat_setVolume(pat, ibeat, vel);
}
//  function -- screen_extCmdVelocityRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdPatternSetRec
//
//  Description
//
//      Extended command for setting a pattern number in song mode.
//
//  Arguments
//
//      screen
//
//          Pointer to the Screen struct instance.
//
//      ipat
//
//          The pattern number.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdPatternSetRec(struct Screen *screen, int ipat)
{
    int j, ichan=0;
    struct Song *song;
    struct SongChannel *schan=NULL;
    struct SongBar *sbar=NULL;
    static int iParams[REVENT_MAX_IPARMS];

    //
    //  Helper for the screen_extCommandRec function
    //  Send a velocity change on the current note
    //  Record to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    //
    //  This can only be done in song mode (from the keyboard)
    //
    if (screen->patmode)
        return;

    //
    //  Get current song bar  
    //
    sbar = screen->songbars[screen->curcol];

    //
    //  Get which channel we are on
    //
    ichan = sbar->chan;

    schan = song_getChannel(song, ichan);
    if (schan == NULL)
    {
        Error_Printf("E212: screen_extCmdPatternSetRec: NULL SongChannel\n");
        return;
    }

    //
    //  set thy pattern
    //
    schan_setSongPattern(schan, sbar->cursor, ipat);

    //
    //  record the action in a session tag
    //
    if (song->re_record)
    {
        ERROR_RANGE(2,REVENT_MAX_IPARMS, "(pattern set in iParams)");
        iParams[0] = ipat;
        iParams[1] = ichan;
        iParams[2] = sbar->cursor;
        p_CurrRec = Attach_Revent(p_CurrRec, song->beatcount,
                                  REVENT_EXT_PATTERN_SET, iParams);
        p_CurrRec->has_iParams = 3;
    }

    //
    //  identify the song mode cursor column for this channel
    //
    for (j=0; j < NUM_BARS; j++)
    {
        if (screen->songbars[j]->chan == ichan)
        {
            sbar_refresh(screen->songbars[j]);
            return;
        }
    }
}
//  function -- screen_extCmdPatternSetRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdPatternSetRE
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdPatternSetRE(struct Screen *screen)
{
    int ichan=0, ipat=0, ipos=0;
    int j;
    struct Song *song;
    struct SongChannel *schan=NULL;

    //
    //  Set a pattern number in song mode, given   pattern,channel,pos
    //  Playback to a session
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    ERROR_RANGE(2,REVENT_MAX_IPARMS, "(pattern set in iParams)");
    ipat = p_CurrREvents->iParams[0];
    ichan = p_CurrREvents->iParams[1];
    ipos = p_CurrREvents->iParams[2];

    schan = song_getChannel(song, ichan);
    if (schan == NULL)
    {
        Error_Printf("E213: screen_extCmdPatternSetRE: NULL SongChannel\n");
        return;
    }

    //
    // Set the pattern
    // pattern id (ipat) starts counting from 1 (0==no pattern)
    // channel id starts from 0
    // songdata index (ipos) starts from 0
    //
    schan_setSongPattern(schan, ipos, ipat);

    //
    //  Update the screen
    //
    if (screen->patmode)
        return;

    //
    //  identify the song mode cursor column for this channel
    //
    for (j=0; j < NUM_BARS; j++)
    {
        if (screen->songbars[j]->chan == ichan)
        {
            sbar_refresh(screen->songbars[j]);
            return;
        }
    }
}
//  function -- screen_extCmdPatternSetRE
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdWriteRec
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void screen_extCmdWriteRec(struct Screen *screen)
{
    struct Song *song;

    //
    //  Helper for the screen_extCommandRec function
    //  Save the file without having to enter the filename
    //  Syntax:
    //      :w <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    ERROR_NULL_VOID(p_CurrREvents);
    song = screen->curr_song;

    screen_saveSongAsciiWrite(screen);

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdWriteRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdWriteARec
//
//  Description
//
//      Save the current song in ASCORE format.  Use the current name as
//      the basename.
//
//  Arguments
//
//      screen
//
//          Pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdWriteARec(struct Screen *screen)
{

    //
    //  Helper for the screen_extCommandRec function
    //  Save the file without having to enter the filename
    //  Syntax:
    //      :w a<enter>
    //
    //
    ERROR_NULL_VOID(screen);

    screen_saveSongAscoreWrite(screen);

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdWriteARec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdMarkMRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "m" means mark the current
//      position.   This is used only in Song mode for marking blocks of
//      pattern entries.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdMarkMRec(struct Screen *screen)
{
    struct Song *song;

    //
    //  Mark the upper, left corner of a song mode block
    //  Syntax:
    //      :e  m <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    //
    //  Mark the top, left
    //
    screen->mark_chan1 = sbar_getIndex(screen->songbars[screen->curcol]);
    screen->mark_row1 = screen->curstate.cursorrow;

    if (screen->mark_chan2 >= 0 && screen->mark_row2 >= 0)
    {
        static char msg[128];

        sprintf(msg, "MARKED (chan %d, pos %d  --->  chan %d, pos %d)\n",
                screen->mark_chan1, screen->mark_row1,
                screen->mark_chan2, screen->mark_row2);
        SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
        SLsmg_write_string(msg);
        SLsmg_refresh();
    }
    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdEdMarkMRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdMarkNRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "n" means mark the current
//      position.   This is used only in Song mode for marking blocks of
//      pattern entries.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdMarkNRec(struct Screen *screen)
{
    struct Song *song;

    //
    //  Mark the upper, left corner of a song mode block
    //  Syntax:
    //      :e  n <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    //
    //  Mark the top, left
    //
    screen->mark_chan2 = sbar_getIndex(screen->songbars[screen->curcol]);
    screen->mark_row2 = screen->curstate.cursorrow;

    if (screen->mark_chan1 >= 0 && screen->mark_row1 >= 0)
    {
        static char msg[128];

        sprintf(msg, "MARKED (chan %d, pos %d  --->  chan %d, pos %d)\n",
                screen->mark_chan1, screen->mark_row1,
                screen->mark_chan2, screen->mark_row2);
        SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
        SLsmg_write_string(msg);
        SLsmg_refresh();
    }

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdEdMarkNRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdMarkClearRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "z" means clear marks.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdMarkClearRec(struct Screen *screen)
{
    struct Song *song;

    //
    //  Clear the song block marks
    //  Syntax:
    //      :e  z <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    //
    //  Mark the top, left
    //
    screen->mark_chan1 = -1;
    screen->mark_row1 = -1;
    screen->mark_chan2 = -1;
    screen->mark_row2 = -1;
    if (screen->copybuffer)
    {
        free(screen->copybuffer);
        screen->copybuffer = NULL;
    }

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string("block marks cleared");
    SLsmg_refresh();

    //
    //  screen_refresh will be performed by the calling function
    //
}
//  function -- screen_extCmdEdMarkClearRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdCopyRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "c" means copy the block.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdCopyRec(struct Screen *screen)
{
    struct Song *song;
    struct SongChannel *schan;
    static char msg[128];
    int bindex;

    //
    //  Copy the song block
    //  Syntax:
    //      :e  c <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    //
    //  Check for complete block
    //
    if (screen->mark_chan1 >= 0 && screen->mark_row1 >= 0 &&
        screen->mark_chan2 >= screen->mark_chan1 &&
        screen->mark_row2 >= screen->mark_row1)
    {
        int rows, cols;
        int irow, icol, krow, kcol;

        //
        //  Calculate the size
        //
        rows = screen->mark_row2 - screen->mark_row1 + 1;
        screen->copyrows = rows;
        cols = screen->mark_chan2 - screen->mark_chan1 + 1;
        screen->copycols = cols;

        //
        //  Allocate the buffer
        //
        screen->copybuffer = (int *)malloc( sizeof(int) * rows * cols );
        if (screen->copybuffer == NULL)
        {
            Error_Printf("E214: mam alloc of copybuffer\n");
            goto Error;
        }

        //
        //  Copy the songdata
        //
        kcol = 0;
        for (icol = screen->mark_chan1; icol <= screen->mark_chan2; icol++)
        {
            schan = song_getChannel(screen->curr_song, icol);
            if (schan == NULL)
            {
                Error_Printf("E215: NULL SongChannel in Copy Block\n");
                goto Error;
            }
            krow = 0;
            for (irow = screen->mark_row1; irow <= screen->mark_row2; irow++)
            {
                if (irow >= SONGLEN)
                {
                    Error_Printf("E216: row exceeds SONGLEN in Copy Block\n");
                    goto Error;
                }
                bindex = (krow*cols) + kcol;
                screen->copybuffer[bindex] = schan->songdata[irow];
                krow++;
            }

            kcol++;
        }

        //
        //  Clear the marks
        //
        screen->mark_chan1 = -1;
        screen->mark_row1 = -1;
        screen->mark_chan2 = -1;
        screen->mark_row2 = -1;

        sprintf(msg, "copied block: %d rows, %d cols", screen->copyrows,
                screen->copycols);
    }
    else
    {
        sprintf(msg, "invalid block");
        SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
        SLsmg_write_string(msg);
        SLsmg_refresh();
        goto Error;
    }

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(msg);
    SLsmg_refresh();

    //
    //  screen_refresh will be performed by the calling function
    //

    return; // success

Error:
    screen->mark_chan1 = -1;
    screen->mark_row1 = -1;
    screen->mark_chan2 = -1;
    screen->mark_row2 = -1;
    if (screen->copybuffer)
        free(screen->copybuffer);
    screen->copybuffer = NULL;
    screen->copyrows = 0;
    screen->copycols = 0;
}
//  function -- screen_extCmdEdCopyRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdCutRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "x" means cut the block.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdCutRec(struct Screen *screen)
{
    struct Song *song;
    struct SongChannel *schan;
    static char msg[128];
    int ibar;
    int bindex;

    //
    //  Cut the song block
    //  Syntax:
    //      :e  x <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    //
    //  Check for complete block
    //
    if (screen->mark_chan1 >= 0 && screen->mark_row1 >= 0 &&
        screen->mark_chan2 >= screen->mark_chan1 &&
        screen->mark_row2 >= screen->mark_row1)
    {
        int rows, cols;
        int irow, icol, krow, kcol;

        //
        //  Calculate the size
        //
        rows = screen->mark_row2 - screen->mark_row1 + 1;
        screen->copyrows = rows;
        cols = screen->mark_chan2 - screen->mark_chan1 + 1;
        screen->copycols = cols;

        //
        //  Allocate the buffer
        //
        screen->copybuffer = (int *)malloc( sizeof(int) * rows * cols );
        if (screen->copybuffer == NULL)
        {
            Error_Printf("E217: mam alloc of copybuffer\n");
            goto Error;
        }

        //
        //  Cut the songdata
        //
        kcol = 0;
        for (icol = screen->mark_chan1; icol <= screen->mark_chan2; icol++)
        {
            schan = song_getChannel(screen->curr_song, icol);
            if (schan == NULL)
            {
                Error_Printf("E218: NULL SongChannel in Cut Block\n");
                goto Error;
            }
            krow = 0;
            for (irow = screen->mark_row1; irow <= screen->mark_row2; irow++)
            {
                if (irow >= SONGLEN)
                {
                    Error_Printf("E219: row exceeds SONGLEN in Cut Block\n");
                    goto Error;
                }
                bindex = (krow*cols) + kcol;
                //
                //  Copy the data into the buffer
                //
                screen->copybuffer[bindex] = schan->songdata[irow];
                //
                //  Set the original data to 0 (the "cut")
                //
                schan->songdata[irow] = 0;
                krow++;
            }

            kcol++;
        }

        //
        //  Clear the marks
        //
        screen->mark_chan1 = -1;
        screen->mark_row1 = -1;
        screen->mark_chan2 = -1;
        screen->mark_row2 = -1;

        sprintf(msg, "cut block: %d rows, %d cols", screen->copyrows,
                screen->copycols);
    }
    else
    {
        sprintf(msg, "invalid block");
        SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
        SLsmg_write_string(msg);
        SLsmg_refresh();
        goto Error;
    }

    //
    //  Refresh song mode channel bars
    //
    for (ibar = 0; ibar < numbars; ++ibar)
    {
        sbar_refresh(screen->songbars[ibar]);
    }

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(msg);
    SLsmg_refresh();

    //
    //  screen_refresh will be performed by the calling function
    //

    return; // success

Error:
    screen->mark_chan1 = -1;
    screen->mark_row1 = -1;
    screen->mark_chan2 = -1;
    screen->mark_row2 = -1;
    if (screen->copybuffer)
        free(screen->copybuffer);
    screen->copybuffer = NULL;
    screen->copyrows = 0;
    screen->copycols = 0;
}
//  function -- screen_extCmdEdCutRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdPasteRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "v" means paste the block.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdPasteRec(struct Screen *screen)
{
    struct Song *song;
    struct SongChannel *schan;
    static char msg[128];
    int ibar;
    int bindex;

    //
    //  Paste the song block
    //  Syntax:
    //      :e  v <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    if (screen->copybuffer == NULL)
    {
        SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
        SLsmg_write_string("empty buffer");
        SLsmg_refresh();
        return;
    }

    //
    //  Check for complete block
    //
    if (screen->copybuffer && screen->copyrows > 0 && screen->copycols > 0)
    {
        int rows, cols;
        int irow, icol, krow, kcol;
        int curr_chan, curr_pos;

        //
        //  Calculate the size
        //
        rows = screen->copyrows;
        cols = screen->copycols;

        //
        //  get the current position
        //
        curr_chan = sbar_getIndex(screen->songbars[screen->curcol]);
        curr_pos = screen->curstate.cursorrow;

        //
        //  Paste the songdata
        //
        kcol = 0;
        for (icol = curr_chan; icol < (curr_chan+cols); icol++)
        {
            schan = song_getChannel(screen->curr_song, icol);
            if (schan == NULL)
            {
                Error_Printf("E220: NULL SongChannel\n");
                goto Error;
            }
            krow = 0;
            for (irow = curr_pos; irow < (curr_pos+rows); irow++)
            {
                if (irow >= SONGLEN)
                {
                    Error_Printf("E221: row exceeds SONGLEN in Paste Block\n");
                    goto Error;
                }
                bindex = (krow*cols) + kcol;
                schan->songdata[irow] = screen->copybuffer[bindex];
                krow++;
            }

            kcol++;
        }

        //
        //  Clear the buffer
        //
        free(screen->copybuffer);
        screen->copybuffer = NULL;
        screen->copyrows = 0;
        screen->copycols = 0;

        sprintf(msg, "pasted block");
    }
    else
    {
        sprintf(msg, "invalid block");
        SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
        SLsmg_write_string(msg);
        SLsmg_refresh();
        goto Error;
    }

    //
    //  Refresh song mode channel bars
    //
    for (ibar = 0; ibar < numbars; ++ibar)
    {
        sbar_refresh(screen->songbars[ibar]);
    }

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(msg);
    SLsmg_refresh();

    //
    //  screen_refresh will be performed by the calling function
    //
    
    return; // success

Error:
    free(screen->copybuffer);
    screen->copybuffer = NULL;
    screen->copyrows = 0;
    screen->copycols = 0;
}
//  function -- screen_extCmdEdPasteRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdDelRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "d" means delete the block.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdDelRec(struct Screen *screen)
{
    struct Song *song;
    struct SongChannel *schan;
    static char msg[128];
    int ibar;

    //
    //  Cut the song block
    //  Syntax:
    //      :e  d <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    //
    //  Check for complete block
    //
    if (screen->mark_chan1 >= 0 && screen->mark_row1 >= 0 &&
        screen->mark_chan2 >= screen->mark_chan1 &&
        screen->mark_row2 >= screen->mark_row1)
    {
        int rows, cols;
        int irow, icol, krow, kcol;

        //
        //  Calculate the size
        //
        rows = screen->mark_row2 - screen->mark_row1 + 1;
        screen->copyrows = rows;
        cols = screen->mark_chan2 - screen->mark_chan1 + 1;
        screen->copycols = cols;

        //
        //  Delete the songdata
        //
        kcol = 0;
        for (icol = screen->mark_chan1; icol <= screen->mark_chan2; icol++)
        {
            schan = song_getChannel(screen->curr_song, icol);
            if (schan == NULL)
            {
                Error_Printf("E222: NULL SongChannel in Delete Block\n");
                goto Error;
            }
            krow = 0;
            for (irow = screen->mark_row1; irow <= screen->mark_row2; irow++)
            {
                if (irow >= SONGLEN)
                {
                    Error_Printf("E223: row exceeds SONGLEN in Delete Block\n");
                    goto Error;
                }
                //
                //  Set the original data to 0 (the "delete")
                //
                schan->songdata[irow] = 0;
                krow++;
            }

            kcol++;
        }

        //
        //  Clear the marks
        //
        screen->mark_chan1 = -1;
        screen->mark_row1 = -1;
        screen->mark_chan2 = -1;
        screen->mark_row2 = -1;
        screen->copyrows = 0;
        screen->copycols = 0;

        sprintf(msg, "delete block: %d rows, %d cols", screen->copyrows,
                screen->copycols);
    }
    else
    {
        sprintf(msg, "invalid block");
        SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
        SLsmg_erase_eol();
        SLsmg_write_string(msg);
        SLsmg_refresh();
        goto Error;
    }

    //
    //  Refresh song mode channel bars
    //
    for (ibar = 0; ibar < numbars; ++ibar)
    {
        sbar_refresh(screen->songbars[ibar]);
    }

    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(msg);
    SLsmg_refresh();

    //
    //  screen_refresh will be performed by the calling function
    //

    return; // success

Error:
    screen->mark_chan1 = -1;
    screen->mark_row1 = -1;
    screen->mark_chan2 = -1;
    screen->mark_row2 = -1;
    screen->copyrows = 0;
    screen->copycols = 0;
}
//  function -- screen_extCmdEdDelRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdEdYankRec
//
//  Description
//
//      Extendex command.  The "e" means edit.  The "y" means copy the current
//      song line.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdEdYankRec(struct Screen *screen)
{
    struct Song *song;
    struct SongChannel *schan;
    static char msg[128];
    int mark_row, last_col, icol;

    //
    //  Copy a song line
    //  Syntax:
    //      :e  y <enter>
    //
    //
    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  --- NOT RECORDED to the session ---
    //

    //
    //  If in pattern mode, return.  We want song mode.
    //
    if (screen->patmode)
        return;

    //
    //  Mark the row
    //
    mark_row = screen->curstate.cursorrow;
    screen->copyrows = 1;

    //
    //  Find the valid columns to copy
    //
    last_col = -1;
    for (icol = 0; icol < SONGMAXCHAN; icol++)
    {
        schan = song_getChannel(screen->curr_song, icol);
        if (schan->songdata[mark_row] > 0)
        {
            if (icol > last_col)
                last_col = icol;
        }
    }
    screen->copycols = last_col + 1;

    //
    //  If there is valid data on the line
    //
    if (last_col >= 0)
    {
        int kcol, cols;

        cols = screen->copycols;

        //
        //  Allocate the buffer
        //
        screen->copybuffer = (int *)malloc( sizeof(int) * (last_col+1) );
        if (screen->copybuffer == NULL)
        {
            Error_Printf("E224: mam alloc of copybuffer\n");
            goto Error;
        }

        kcol = 0;
        for (icol = 0; icol < cols; icol++)
        {
            schan = song_getChannel(screen->curr_song, icol);
            screen->copybuffer[kcol] = schan->songdata[mark_row];
            kcol++;
        }
        sprintf(msg, "YANKED (chan 0, pos %d  --->  chan %d, pos %d)\n",
                mark_row, last_col, mark_row);
    }
    else
    {
        sprintf(msg, "empty line");
        goto Error;
    }


    SLsmg_gotorc(SLtt_Screen_Rows - 1, 0);
    SLsmg_erase_eol();
    SLsmg_write_string(msg);
    SLsmg_refresh();

    //
    //  screen_refresh will be performed by the calling function
    //

    return; // success

Error:
    if (screen->copybuffer)
        free(screen->copybuffer);
    screen->copybuffer = NULL;
    screen->copyrows = 0;
    screen->copycols = 0;
}
//  function -- screen_extCmdEdYankRec
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdLoadSession
//
//  Description
//
//      Loads a session file taken from an extended command prompt.  For
//      example,
//          :s p <filename>
//      and turns on the session playback mode.
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//      fname
//
//          The file name for the session to be played.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdLoadSession(struct Screen *screen, const char *fname)
{
    struct Song *song;
    struct RecordEvent *prec;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(fname);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  Read-in the session file
    //
    if (p_REvents)
    {
        //
        //  Delete an existing session
        //
        Delete_RecordEvent(p_REvents);
        p_REvents = p_CurrRec = NULL;
    }

    prec = Read_Events_File(fname);
    if (prec)
    {
        p_REvents = p_CurrREvents = prec;
        strcpy(g_Playback_Filename, fname);
        g_Playback_Session = 1;
        song->re_play = 1;
        Msg_Printf("Session Play file %s ON\n", fname);
    }
    else
        Error_Printf("E225: cannot open session (%s)\n", fname);

    //
    //  Screen refresh performed by calling function
    //
}
//  function -- screen_extCmdLoadSession
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdTogglePlay
//
//  Description
//
//      Toggles play session ON or OFF.
//      example,
//          :s p
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdTogglePlay(struct Screen *screen)
{
    struct Song *song;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  Read-in the session file
    //
    if (p_REvents)
    {
        song->re_play = 1 - song->re_play;
        if (song->re_play)
            Msg_Printf("Session Play toggled ON\n");
        else
            Msg_Printf("Session Play toggled OFF\n");
    }

    //
    //  Screen refresh performed by calling function
    //
}
//  function -- screen_extCmdTogglePlay
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_extCmdPlayOnOff
//
//  Description
//
//      Play session ON of OFF
//      example,
//          :s p 1
//          :s p 0
//
//  Arguments
//
//      screen
//
//          The pointer to the Screen struct instance.
//
//      arg
//
//          Character string whose first letter is "0" or "1".
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
void screen_extCmdPlayOnOff(struct Screen *screen, const char *arg)
{
    struct Song *song;

    ERROR_NULL_VOID(screen);
    ERROR_NULL_VOID(arg);
    ERROR_NULL_VOID(screen->curr_song);
    song = screen->curr_song;

    //
    //  Read-in the session file
    //
    if (p_REvents)
    {
        if (arg[0] == 0x31)
            song->re_play = 1;
        else if (arg[0] == 0x30)
            song->re_play = 0;
        else
            return;

        if (song->re_play)
            Msg_Printf("Session Play ON\n");
        else
            Msg_Printf("Session Play OFF\n");
    }

    //
    //  Screen refresh performed by calling function
    //
}
//  function -- screen_extCmdPlayOnOff
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  screen_WaitNc
//
//  Description
//
//      Mimic the netcat "nc" tool to listen on a port and return.
//
//      The purpose of this call is to block (the calling function) while
//      listening for a socket connection request on the specified port.
//      This is used as a triggering method.  Another process on this or
//      a remote computer can send a trigger to Attrk simply by requesting
//      a TCP port connection.
//
//  Arguments
//
//      port
//         The port number to listen on
//
//      timeout_seconds
//          The time in seconds to wait for a connect before timing out.
//
//  Return
//
//      Void
//
//  -------------------------------------------------------------------------
int screen_WaitNc(int port, int timeout_seconds)
{
    int server_sockfd, client_sockfd;
    int server_len, client_len;
    struct sockaddr_in server_address;
    struct sockaddr_in client_address;
    unsigned long loopCount = 0, max_loopCount = 30000;
    int err, flags;
    char *pserr=NULL;

    if (timeout_seconds > 0)
        max_loopCount = timeout_seconds * 1000;

    //  Open a socket (as server)
    server_sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (server_sockfd == -1)
    {
        pserr = strerror(errno);
        if (pserr)
            LogFile_Printf("ERROR: screen_WaitNc: socket: %d (%s)\n",
                           errno, pserr);
        else
            LogFile_Printf("ERROR: screen_WaitNc: socket: errno=%d\n", errno);
        return -1;
    }

    //  Set up server params
    server_address.sin_family = AF_INET;
    server_address.sin_addr.s_addr = INADDR_ANY;
    server_address.sin_port = htons(port);
    server_len = sizeof(server_address);
    client_len = sizeof(client_address);

    //  Set to non-blocking
    flags = fcntl(server_sockfd, F_GETFL, 0);
    fcntl(server_sockfd, F_SETFL, O_NONBLOCK|flags);

    //  Server-mode: bind
    err = bind(server_sockfd, (struct sockaddr *)&server_address, server_len);
    if (err == -1)
    {
        pserr = strerror(errno);
        if (pserr)
            LogFile_Printf("ERROR: screen_WaitNc: bind: %d (%s)\n",
                           errno, pserr);
        else
            LogFile_Printf("ERROR: screen_WaitNc: bind: errno=%d\n", errno);
        close(server_sockfd);
        return -2;
    }

    //  Listen for connections
    listen(server_sockfd, 5);
    do
    {
        //  Call accept in non-blocking mode (see fcntl call above)
        client_sockfd = accept(server_sockfd,
                               (struct sockaddr *)&client_address,
                               &client_len);
        if (client_sockfd == -1)
        {
            int errval;

            errval = errno;
            if (errval == EAGAIN || errval == EWOULDBLOCK)
            {
                //  ignore becaue we are in non-blocking mode
            }
            else
            {
                //  Genuine error
                pserr = strerror(errno);
                if (pserr)
                    LogFile_Printf("ERROR: screen_WaitNc: accept: %d (%s)\n",
                                   errno, pserr);
                else
                    LogFile_Printf("ERROR: screen_WaitNc: accept: errno=%d\n",
                                   errno);
                close(server_sockfd);
                return -3;
            }
        }

        if (client_sockfd >= 0)
        {
            close(client_sockfd);
            break;
        }

        //  Sleep for a millisecond
        usleep(1000);

    } while (loopCount++ < max_loopCount);

    close(server_sockfd);

    return 0;
}
//  function -- screen_WaitNc

