/**
 * C-code modifications Copyright (C) 2009 Alfred Steffens Jr.
 * based on TekTracker, Copyright (C) 2001 Billy Biggs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef SONG_H_INCLUDED
#define SONG_H_INCLUDED

struct SongPattern {
	int id;
	int *note;
	int *slide;
	int *vol;
	char name[MAX_NAME_LENGTH];
	int endnote;	                // last note in pattern
    int asc_noteoff;                // if true, explicit note off for Ascore
    void *curr_song;
    struct SongPattern *next;
};

int Init_SongPattern(struct SongPattern *spat, int id);
struct SongPattern *Create_SongPattern(int id);
void Delete_SongPattern(struct SongPattern *spat);
void spat_clear(struct SongPattern *spat);
int spat_isBlank(struct SongPattern *spat);
int spat_isBlankorOff(struct SongPattern *spat);
void spat_copyTo(struct SongPattern *spat,  struct SongPattern *pat);
void spat_setVolume(struct SongPattern *spat,  int beat, int value);
void spat_setName(struct SongPattern *spat, const char *newname);
void spat_transposeNotes(struct SongPattern *spat, int trans);
int spat_getIndex(struct SongPattern *spat);
int spat_getNote(struct SongPattern *spat, int beat);
void spat_setNote(struct SongPattern *spat, int beat, int value);
int spat_getSlide(struct SongPattern *spat, int beat);
void spat_toggleSlide(struct SongPattern *spat, int beat);
int spat_getVolume(struct SongPattern *spat, int beat);
const char *spat_getName(struct SongPattern *spat);
int spat_getEndNote(struct SongPattern *spat);
void spat_setEndNote(struct SongPattern *spat, int enote);


//
//  TBD -- Linked list of MIDI setup control messages
//
struct SChanMidiCtrl {
    int chg_ready;
    int control_number;
    int control_value;
    struct SChanMidiCtrl *next;
};
struct SChanMidiCtrl *Create_SChanMidiCtrl(void);
void Delete_SChanMidiCtrl(struct SChanMidiCtrl *sctrl);

struct SongChannel {
    int prognum;
    int banknum;
    int prchg_ready;
    int bnchg_ready;
    int midichannel;
    int midiport;
    int volume;
    int vlchg_ready;
    int devnum;
    int lastnote;
    int mute;
    int *songdata;
    char name[MAX_NAME_LENGTH];
    struct SongPattern **patterns;
    int midisetup_ready;    // OR of all the other "change ready" flags
    struct SChanMidiCtrl *ctrl_msg;
    void *curr_song;
    struct SongChannel *next;
};

int Init_SongChannel(struct SongChannel *schan);
struct SongChannel *Create_SongChannel(void);
void Delete_SongChannel(struct SongChannel *schan);
void schan_SetSongPointer(struct SongChannel *schan, void *song);
void schan_clear(struct SongChannel *schan);
void schan_setVolume(struct SongChannel *schan, int newvol);
void schan_setMidiChannel(struct SongChannel *schan, int newchan);
void schan_setMidiPort(struct SongChannel *schan, int newport);
void schan_setSongPattern(struct SongChannel *schan, int index, int patid);
void schan_setName(struct SongChannel *schan, const char *newname);
int schan_getProgramNumber(struct SongChannel *schan);
void schan_setProgramNumber(struct SongChannel *schan, int newprg);
int schan_getBankNumber(struct SongChannel *schan);
void schan_setBankNumber(struct SongChannel *schan, int newbnk);
int schan_getMidiChannel(struct SongChannel *schan);
int schan_getMidiPort(struct SongChannel *schan);
int schan_getDevice(struct SongChannel *schan);
int schan_getLastNote(struct SongChannel *schan);
const char *schan_getName(struct SongChannel *schan);
int schan_isMuted(struct SongChannel *schan);
void schan_toggleMute(struct SongChannel *schan);
struct SongPattern *schan_getSongPattern(struct SongChannel *schan, int index);
struct SongPattern *schan_getChannelPattern(struct SongChannel *schan, int index);
int schan_getVolume(struct SongChannel *schan);
void schan_setLastNote(struct SongChannel *schan, int last);

struct Song {
    int playing;
    int curpat;
    int curbeat;
    unsigned long beatcount;
    int startpat;
    int loopactive;
    int tempo;
    int usecpb;
    int looppos;
    int loopdest;
    int re_record;
    int re_play;
    char filename[MAX_FILENAME_LENGTH];
    struct SongChannel **channels;
    struct Song *next;
};

int Init_Song(struct Song *song);
struct Song *Create_Song(void);
void Delete_Song(struct Song *song);
void song_SetSongPointer(struct Song *song);
void song_clear(struct Song *song);
void song_setTempo(struct Song *song, int newtempo);
int song_getEndNote(struct Song *song);
void song_incrBeat(struct Song *song);
int song_saveSong(struct Song *song, const char *filename);
int song_saveSongAscii(struct Song *song, const char *filename);
int song_saveSongAscore(struct Song *song, const char *filename);
int song_loadSong(struct Song *song, const char *filename);
int song_loadControlSetupAscii(struct Song *song, struct SongChannel *schan,
                               struct XmlToken *xmltok);
int song_loadChannelAscii(struct Song *song, struct SongChannel *schan,
                          struct XmlToken *xmltok, int chan_index);
int song_loadPatternAscii(struct Song *song, struct SongPattern *spat,
                          struct XmlToken *xmltok, int pat_index);
int song_loadSongAscii(struct Song *song, const char *filename);
void song_setLoopPosition(struct Song *song, int newpos);
void song_setLoopDestination(struct Song *song, int newdest);
void song_start(struct Song *song);
void song_stop(struct Song *song);
void song_rewind(struct Song *song);
int song_isPlaying(struct Song *song);
int song_getCurPat(struct Song *song);
int song_getCurBeat(struct Song *song);
int song_getStartPos(struct Song *song);
void song_setStartPos(struct Song *song, int newstartpat);
struct SongChannel *song_getChannel(struct Song *song, int chan);
int song_getTempo(struct Song *song);
int song_getUsecpb(struct Song *song);
int song_isLoopActive(struct Song *song);
void song_toggleLoop(struct Song *song);
int song_getLoopPosition(struct Song *song);
int song_getLoopDestination(struct Song *song);
void song_armMidiSetupMessages(struct Song *song);
void song_doMidiSetup(struct Song *song);
void song_refreshMidiSetup(struct Song *song);

#endif // SONG_H_INCLUDED
