/**
 * Copyright (C) 2009 Alfred Steffens Jr.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <defs.h>
#include <logprintf.h>
#include <errorprintf.h>
#include <xmltools.h>


static char g_seps1[3] = {0x3C, 0x3E, 0x00};
static char g_seps2[4] = {'=', 0x22, 0x27, 0x00};
extern int numlogtabs;


//  -------------------------------------------------------------------------
//  Function    --  convert_tolower
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
char *convert_tolower(char *str)
{
    static char tmpstr[1024];
    long j, len;

    if (str == NULL)
        return NULL;

    memset(tmpstr, 0, 1024);

    len = strlen(str);
    for (j=0; j < len; j++)
    {
        if (j >= 1024)
            break;
        tmpstr[j] = tolower(str[j]);
    }
    tmpstr[1023] = 0x00;
    return (char *)tmpstr;
}
//  function -- convert_tolower
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_StrList
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct StrList *Create_StrList(char *field)
{
    struct StrList *slist;

    slist = (struct StrList *)malloc( sizeof(struct StrList) );
    if (slist == NULL)
    {
        Error_Printf("ERROR: Create_StrList: mem alloc\n");
        return NULL;
    }

    slist->type = 0;
    if (field)
        slist->field = strdup(field);
    else
        slist->field = NULL;
    slist->next = NULL;

    return slist;
}
//  function -- Create_StrList
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_StrList
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void Delete_StrList(struct StrList *slist)
{
    struct StrList *pN, *pX;

    if (slist == NULL)
        return;

    pX = slist;
    while (pX)
    {
        pN = pX;
        pX = pX->next;
        if (pN->field)
            free(pN->field);
        free(pN);
    }
}
//  function -- Delete_StrList
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Split_String
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct StrList *Split_String(char *str, char *seps)
{
    struct StrList *slist = NULL, *curr_slist = NULL;
    char *ptok;
    char *tmpstr=NULL;
    static char tmpbuf[1024];
    int isTmpstrAlloc = 0;

    if (str == NULL || seps == NULL)
        return NULL;

    memset(tmpbuf, 0, 1024);

    if (strlen(str) >= 1024)
    {
        tmpstr = strdup(str);
        if (tmpstr == NULL)
        {
            Error_Printf("ERROR: Split_String: mem alloc on strdup\n");
            return NULL;
        }
        isTmpstrAlloc = 1;
    }
    else
    {
        strcpy(tmpbuf, str);
        tmpstr = (char *)tmpbuf;
    }

    //
    //  Split the string into tokens separated by seps
    //
    ptok = strtok(tmpstr, seps);
    while (ptok)
    {

        if (slist == NULL)
        {
            slist = curr_slist = Create_StrList(ptok);
        }
        else
        {
            curr_slist->next = Create_StrList(ptok);
            curr_slist = curr_slist->next;
        }

        //
        //  get the next token
        //
        ptok = strtok(NULL, seps);
    }

    if (isTmpstrAlloc && tmpstr)
        free(tmpstr);

    return slist;
}
//  function -- Split_String
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Find_Matching_Tag
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct StrList *Find_Matching_Tag(struct StrList *slist)
{
    struct StrList *curr_slist;
    char *ptag;

    if (slist == NULL)
        return NULL;

    //
    //  Get first field in a whitespace-separated list
    //
    ptag = Get_Leading_Token(slist->field, " ");
    if (ptag == NULL)
        return NULL;

    //
    //  Search the remaining list until the last field
    //  matches the above tag
    //
    curr_slist = slist->next;
    while (curr_slist)
    {
        char *ptok;

        ptok = Get_Final_Token(curr_slist->field, " ");
        if (ptok)
        {
            char *pend;

            pend = strchr(ptok, 0x2F);
            if (pend)
            {
                *pend = 0x00;
                if (strcmp(ptag, pend+1)==0)
                {
                    return curr_slist;
                }
            }
        }
        curr_slist = curr_slist->next;
    }
    return NULL;
}
//  function -- Find_Matching_Tag
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_XmlAttr
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct XmlAttr *Create_XmlAttr(void)
{
    struct XmlAttr *attr=NULL;

    attr = (struct XmlAttr *)malloc( sizeof(struct XmlAttr) );
    if (attr == NULL)
    {
        Error_Printf("ERROR: Create_XmlAttr: mem alloc\n");
        return NULL;
    }
    attr->type = 0;
    attr->name[0] = 0x00;
    attr->value[0] = 0x00;
    attr->next = NULL;
    return attr;
}
//  function -- Create_XmlAttr
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_XmlAttr
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void Delete_XmlAttr(struct XmlAttr *attr)
{
    struct XmlAttr *pN, *pX;

    if (attr == NULL)
        return;

    pX = attr;
    while (pX)
    {
        pN = pX;
        pX = pX->next;
        free(pN);
    }
}
//  function -- Delete_XmlAttr
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Create_XmlToken
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct XmlToken *Create_XmlToken(void)
{
    struct XmlToken *token=NULL;

    token = (struct XmlToken *)malloc( sizeof(struct XmlToken) );
    if (token == NULL)
    {
        Error_Printf("ERROR: Create_XmlToken: mem alloc\n");
        return NULL;
    }
    token->type = 0;
    token->name[0] = 0x00;
    token->attr = NULL;
    token->next = NULL;
    token->child = NULL;
    return token;
}
//  function -- Create_XmlToken
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Delete_XmlToken
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void Delete_XmlToken(struct XmlToken *token)
{
    struct XmlToken *pN, *pX;

    if (token == NULL)
        return;

    pX = token;
    while (pX)
    {
        pN = pX;
        pX = pX->next;
        Delete_XmlAttr(pN->attr);
        Delete_XmlToken(pN->child);
        free(pN);
    }
}
//  function -- Delete_XmlToken
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Get_Leading_Token
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
char *Get_Leading_Token(char *str, char *seps)
{
    static char tok[MAX_XMLTOKEN_LENGTH];
    char *ptok = NULL;

    if (str == NULL || seps == NULL)
        return NULL;

    memset(tok, 0, MAX_XMLTOKEN_LENGTH);

    strncpy(tok, str, (MAX_XMLTOKEN_LENGTH-1));
    tok[(MAX_XMLTOKEN_LENGTH-1)] = 0x00;
    ptok = strtok(tok, seps);
    return ptok;
}
//  function -- Get_Leading_Token
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Get_Final_Token
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
char *Get_Final_Token(char *str, char *seps)
{
    static char tok[MAX_XMLTOKEN_LENGTH];
    char *ptok = NULL, *final_ptok = NULL;

    if (str == NULL || seps == NULL)
        return NULL;

    memset(tok, 0, MAX_XMLTOKEN_LENGTH);

    strcpy(tok, str);
    ptok = strtok(tok, seps);
    while (ptok)
    {
        final_ptok = ptok;
        ptok = strtok(NULL, seps);
    }
    return final_ptok;
}
//  function -- Get_Final_Token
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Get_Attributes
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct XmlAttr *Get_Attributes(char *str)
{
    struct XmlAttr *xmlattr = NULL, *curr_xmlattr = NULL;
    struct StrList *slist = NULL, *curr_slist;

    if (str == NULL)
        return NULL;

    //
    //  Split the whole "tag" string on whitespace
    //
    slist = Split_String(str, " ");
    if (slist == NULL)
        return NULL;

    //
    //  Find the Attributes
    //  Each link in this StrList chain should have a string of the form
    //      name=value
    //  or
    //      name="value"
    //  or
    //      name='value'
    //
    curr_slist = slist;
    while (curr_slist)
    {
        static char name[MAX_XMLATTR_LENGTH];
        static char value[MAX_XMLATTR_LENGTH];
        static char tmpstr[MAX_XMLTOKEN_LENGTH];
        char *pch, *psol;

        //
        //  Skip NULL fields (this should not happen)
        //
        if (curr_slist->field == NULL || strlen(curr_slist->field)==0)
        {
            curr_slist = curr_slist->next;
            continue;
        }

        //
        //  Skip the leading token (the tag), or any that lack an "="
        //
        pch = strchr(curr_slist->field, '=');
        if (pch == NULL)
        {
            curr_slist = curr_slist->next;
            continue;
        }

        //
        //  Split on Attribute separators
        //
        strcpy(tmpstr, curr_slist->field);
        pch = Get_Leading_Token(tmpstr, g_seps2);
        if (pch == NULL)
            break;
        //
        //  save the name until we get a value
        //
        strcpy(name, pch);

        //
        //  Split on Attribute separators
        //
        strcpy(tmpstr, curr_slist->field);
        pch = Get_Final_Token(tmpstr, g_seps2);
        if (pch == NULL)
        {
            curr_slist = curr_slist->next;
            continue;
        }
        //
        //  save the value until we get a value
        //
        strcpy(value, pch);

        //
        //  Succes--now create the Atrribute
        //
        if (xmlattr == NULL)
        {
            xmlattr = curr_xmlattr = Create_XmlAttr();
        }
        else
        {
            curr_xmlattr->next = Create_XmlAttr();
            curr_xmlattr = curr_xmlattr->next;
        }

        //
        //  Save the values into the Attribute
        //
        strcpy(curr_xmlattr->name, name);

        //
        //  Save the value
        //
        psol = strchr(value, 0x2F);
        if (psol && (psol > value))
            *psol = 0x00;
        strcpy(curr_xmlattr->value, value);
        
        curr_slist = curr_slist->next;
    }

    Delete_StrList(slist);
    return xmlattr;
}
//  function -- Get_Attributes
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Get_XmlAttr_Value
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
char *Get_XmlAttr_Value(struct XmlAttr *xmlattr, char *name)
{
    struct XmlAttr *curr_xmlattr;

    if (xmlattr == NULL || name == NULL)
        return NULL;

    curr_xmlattr = xmlattr;
    while (curr_xmlattr)
    {
        char *plower;

        plower = convert_tolower(curr_xmlattr->name);
        if (strcmp(plower, name)==0)
        {
            return curr_xmlattr->value;
        }
        curr_xmlattr = curr_xmlattr->next;
    }
    return NULL;
}
//  function -- Get_XmlAttr_Value
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Get_Integer_XmlAttr
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int Get_Integer_XmlAttr(struct XmlAttr *xmlattr, char *attrname, int *result)
{
    if (xmlattr==NULL || attrname==NULL || result==NULL)
        return -1;

    //
    //  "attrname" should be something like "index" or "number" or "position",
    //  etc.
    //
    //
    if (xmlattr)
    {
        char *pval;

        pval = Get_XmlAttr_Value(xmlattr, attrname);
        if (pval)
        {
            *result =  atoi(pval);
            return 0;
        }
        else
            return -1;
    }

    return -1;
}
//  function -- Get_Integer_XmlAttr
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Get_String_XmlAttr
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
int Get_String_XmlAttr(struct XmlAttr *xmlattr, char *attrname, char *result)
{
    if (xmlattr==NULL || attrname==NULL || result==NULL)
        return -1;

    //
    //  "attrname" should be something like "index" or "number" or "position",
    //  etc.
    //
    //
    if (xmlattr)
    {
        char *pval;

        pval = Get_XmlAttr_Value(xmlattr, attrname);
        if (pval)
        {
            strncpy(result, pval, (MAX_NAME_LENGTH-1));
            result[MAX_NAME_LENGTH-1] = 0x00;
            return 0;
        }
        else
            return -1;
    }

    return -1;
}
//  function -- Get_String_XmlAttr
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  ParseXml
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct XmlToken *ParseXml(struct StrList *slist)
{
    struct XmlToken *xmltok = NULL, *curr_xmltok = NULL;
    struct StrList *curr_slist, *last_slist;

    if (slist == NULL)
        return NULL;

    //  ---------------------------------------------------
    //  We assume that slist is split on '<' '>' boundaries
    //  ---------------------------------------------------

    //
    //  Search each string token in the list
    //
    curr_slist = slist;
    while (curr_slist && curr_slist->type == 0)
    {
        char *pch;

        if (curr_slist->field == NULL || strlen(curr_slist->field)==0)
        {
            curr_slist = curr_slist->next;
            continue;
        }

        //
        //  allocate an Xml tag
        //
        if (xmltok == NULL)
        {
            xmltok = curr_xmltok = Create_XmlToken();
            if (xmltok == NULL)
                return NULL;
        }
        else
        {
            curr_xmltok->next = Create_XmlToken();
            curr_xmltok = curr_xmltok->next;
        }

        //
        //  Save the leading field as the tag identifier
        //
        pch = Get_Leading_Token(curr_slist->field, " ");
        if (pch)
        {
            char *pend;

            //
            //  remove a trailing '/' character
            //
            pend = strchr(pch, 0x2F);
            if (pend)
                *pend = 0x00;            
            strcpy(curr_xmltok->name, pch);
        }
        else
        {
            curr_slist = curr_slist->next;
            continue;
        }

        //
        //  Parse the remaining whitespace-separated fields in this tag
        //  as attributes
        //
        curr_xmltok->attr = Get_Attributes(curr_slist->field);

        //
        //  get the matching tag for this tag
        //
        last_slist = Find_Matching_Tag(curr_slist);
        if (last_slist)
        {
            //
            //  There was a closing tag
            //
            curr_xmltok->type = TAG_TYPE_OPEN;
            last_slist->type = TAG_TYPE_CLOSE;
            //
            //  The StrList chain between this and the Last link should
            //  be the child tags--recursively call this function.
            //
            curr_xmltok->child = ParseXml(curr_slist->next);

            //
            //  Try to go to the following tag at this level
            //
            curr_slist = last_slist->next;
        }
        else
        {
            curr_xmltok->type = TAG_TYPE_SINGLE;
            //
            //  Try to go to the following tag at this level
            //
            curr_slist = curr_slist->next;
        }
    }

    return xmltok;
}
//  function -- Parse_Xml
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  LogXml
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
void LogXml(struct XmlToken *xmltok)
{
    struct XmlToken *curr_xmltok = NULL;

    if (xmltok == NULL)
        return;

    LogFile_Printf("LogXml\n");

    curr_xmltok = xmltok;
    while (curr_xmltok)
    {

        LogFile_Printf("%s\n", curr_xmltok->name);
        if (curr_xmltok->child)
        {
            numlogtabs++;
            LogXml(curr_xmltok->child);
            numlogtabs--;
        }

        curr_xmltok = curr_xmltok->next;
    }

    LogFile_Printf("end  LogXml\n");
}
//  function -- LogXml
//  -------------------------------------------------------------------------


//  -------------------------------------------------------------------------
//  Function    --  Read_FileBuffer
//
//  Description
//
//  Arguments
//
//  Return
//
//  -------------------------------------------------------------------------
struct XmlToken *Read_FileBuffer(char *filebuffer)
{
    struct StrList *slist = NULL;
    struct XmlToken *xmltok = NULL;

    //
    //  Spilt the Entire file buffer
    //
    slist = Split_String(filebuffer, g_seps1);
    if (slist == NULL)
    {
        Error_Printf("ERROR: Read_FileBuffer: cannot split file buffer\n");
        return NULL;
    }


    //
    //  Parse the file text into a linked list of Xml tags
    //
    xmltok = ParseXml(slist);
    if (xmltok == NULL)
    {
        Error_Printf("ERROR: Read_FileBuffer: cannot parse file buffer\n");
        goto Error;
    }

    //
    //  Success
    //
    Delete_StrList(slist);
    return xmltok;

Error:
    Delete_StrList(slist);
    return NULL;
}
//  function -- Read_FileBuffer
//  -------------------------------------------------------------------------

